/*
	File:			Camera.h
	
	Function:		Defines a camera. When sent to a Renderer, this defines
					the current view (and perspective) transformations.
					A camera can be embedded in a scene: see SceneObjects.cc

	Author(s):		Andrew Willmott

	Copyright:		(c) 1995-2000, Andrew Willmott
 */
 
#ifndef __Camera__
#define __Camera__

#include "gcl/Geometry.h"
#include "cl/String.h"
#include <iostream.h>

class Camera
// Maps view into viewport [-1,1] x [-1,1]
{
public:	
	Camera();
	
				// Set standard camera parameters. (shared by all.)
	Void		SetClipping(GCLReal nearPlane, GCLReal farPlane);
	Void		SetFOV(GCLReal fieldOfView);
	Void		SetAspect(GCLReal aspect);
	Void		SetScale(GCLReal scale);

				// useful routines for directly setting the view
	Void		SetView(const Point &eyePoint, const Quaternion &orientation);
	Void		SetViewDir(const Point &pos, const Vector &dir);
	Void		SetLookAt(const Point &pos, const Point &at);
	Void		SetModelCentric(const Coord &rot, const Vector &trans);
	Void		SetViewUp(Vector up);

				// set up a particular model from params (low-level)
	Void		SetupViewDir();
	Void		SetupLookAt();
	Void		SetupModelCentric();

				// set params for a particular model from the generic model.
	Void		GetViewDirParams();
	Void		GetLookAtParams();
	Void		GetModelCentricParams();

						// avar handling stuff
	virtual Void		PrintAvars(Bool animFormat = false);
	virtual Bool		CheckAvar(StrConst avarName, GCLReal avarVal);
	virtual Void		SetupFromParams();

	virtual Transform	ProjMatrix() const;
						// Return the projection matrix from Camera space to
						// Display Space ([-1,1] ^ 2)
	virtual Transform	ModelMatrix() const;
						// Return the transformation from World space to Camera
						// space
	

	Int					model;

// generic camera model. all models convert to this.
	Point		position;			///< location of camera
	Quaternion	orient;				///< orientation of camera
	GCLReal		scale;				///< scale of camera c.f. scene
	GCLReal		clipNear, clipFar;	///< near and far clipping planes
	GCLReal		fov;				///< field of view in degrees
	GCLReal		aspect;				///< aspect ratio
	Int			lhand;				///< left handed coordinate system?

// the standard view-up/direction camera model params
	Point		up;					///< up vector, normalised
	Vector		viewDir;			///< viewing direction
	Point		lookAt;				///< look-at point

// model-centric model params
	Point		modelCentre;
	Point		trans;
	Coord		rot;
};

/**
	FindWorldRay maps from display coords ([-1, 1], [-1, 1]) to
	a ray through the world. We assume our world-to-screen matrix
	maps from some form of view pyramid to the box from -1 to 1 in x and y
	and 0 to 1 in z. 
		
	Note that this works regardless of whether a perspective or orthogonal
	projection is used; the only thing you need is the inverse WS matrix:

		inverseW2S = inv(camera.ProjMatrix() * camera.ModelMatrix);

*/

Void FindWorldRay(
			const Coord		&c,
			const Transform &inverseW2S,
			Point			&from,
			Vector			&dir
		);


#endif
