/*
	File:		MeshJoin.cc
	
	Function:	Connect up adjacent triangles in a shared-vertex mesh.
				Sets the 'adjFace' and 'adjEdge' fields correctly for
				groups of triangles belonging to the same surface.
	
	Notes:		Requires triangles to share a single vertex array
				
				Calling conventions: 
					Call Init() with the number of vertices in the model
					Call StartGroup() for each surface group
					Call AddTriangle for all faces in that group
					Call EndGroup() when done.
					Repeat the last 3 calls as necessary.

				The cost of this algorithm is O(f + v), where f is the
				number of faces and v the number of vertices. For a manifold
				surface, this is O(v).
				
	Author:		Andrew Willmott, 1997
*/

/*
	NOTE:

	This code only handles manifold surfaces -- if more than two faces
	share an edge, only the first two adjoining faces will be recognized.
	Generally this will give good results, as the faces that are "supposed"
	to be connected will come from nearby parts of the model.
*/

#include "MeshJoin.h"

#define MJ_DBG if (0) cerr

MeshJoin::MeshJoin() : vertexFaces(0), vertexEdges(0)
{
}

MeshJoin::~MeshJoin()
{
}

Void MeshJoin::Init(Int numVertices)
{
	vertexFaces = new FacePtr[numVertices];
	vertexEdges = new Edge[numVertices];
	SELF.numVertices = numVertices;
	surfaceVertices = numVertices;
}

Void MeshJoin::StartGroup()
{
	Int		i;

	for (i = 0; i < numVertices; i++)
		vertexFaces[i] = 0;
	for (i = 0; i < numVertices; i++)
		vertexEdges[i] = 0;
}

Bool MeshJoin::EdgesMatch(
		FacePtr face1,
		Edge	edge1,
		FacePtr face2,
		Edge	edge2
	)
{
	return(
		// shared vertex?
		(face1->index[edge1] == face2->index[edge2])
	);
}

inline Int NextTriEdge(Int i)
{
	return(++i == 3 ? 0 : i);
}

Void MeshJoin::AddTriangle(Face *addFace)
{
	Int		i, j, k;
	FacePtr	curFace, *curFacePtr;
	Edge	curEdge, *curEdgePtr;
	Int		vIndex;
	Bool	match[4];
	
	/*
		Each vertex has a list of unmatched incident edges (where
		unmatched means that we haven't found a neighbouring triangle
		along that edge.)

		When we add a triangle, we first check to see whether an edge is
		present in this list. If it is, we remove it, and connect the
		two faces along that edge. If it is not, we add it to the list.

	*/
	
	// for all vertices...

	for (i = 0; i < 3; i++)
	{
		j = NextTriEdge(i);	// next vertex
		k = NextTriEdge(j);

#ifdef MJ_SURF_PT
		addFace->clrIdx[i] = addFace->index[i];
#endif
		
		MJ_DBG << "vertex " << j << " = " << addFace->index[j] << endl;

		// Check outgoing edge against the vertex's 
		// edge list.
		
		// we will be looking at vertex j
		// incident edge is (i)->(j), outgoing edge is (j)->(k)

		vIndex = addFace->index[j];
		
		// we first check the edge list at vertex (j) and try
		// to locate an edge which starts from vertex (i).
		
		curFacePtr = &vertexFaces[vIndex];
		curEdgePtr = &vertexEdges[vIndex];
		curFace = *curFacePtr;
		curEdge = *curEdgePtr;
				
		match[j] = false;

		while (curFace)
		{
			// Do we have an edge match?
			MJ_DBG << "checking edge to vertex " << curFace->index[curEdge]
					<< endl;
			
			if (EdgesMatch(curFace, curEdge, addFace, k))
			{
				MJ_DBG << "matched." << endl;

				// Yes! remove curEdge from the list
				*curFacePtr = curFace->adjFace[curEdge];
				*curEdgePtr = curFace->adjEdge[curEdge];
				
				// connect the two triangles that share this edge
				addFace->adjFace[j] = curFace;
				addFace->adjEdge[j] = curEdge;
				curFace->adjFace[curEdge] = addFace;
				curFace->adjEdge[curEdge] = j;
				
				Assert(addFace->index[j] == curFace->index[NextTriEdge(curEdge)], 
					   "1st vertices");
				Assert(addFace->index[k] == curFace->index[curEdge], 
					   "2nd vertices");
				
				// and quit.
				match[j] = true;
				break;
			}

			MJ_DBG << "didn't match." << endl;

			// If not, move on to next edge...
			curFacePtr = &curFace->adjFace[curEdge];
			curEdgePtr = &curFace->adjEdge[curEdge];
			curFace = *curFacePtr;
			curEdge = *curEdgePtr;
		}
		
		// If there was no match, we add the incoming edge to the vertex's
		// edge list.
	}
	
	for (i = 0; i < 3; i++)
	{
		j = NextTriEdge(i);	// next vertex

		if (!match[i])
		{
			vIndex = addFace->index[j];

			MJ_DBG << "adding unmatched edge " << i << " to list." << endl;
			// set this edge to point to first member of the list...
			addFace->adjFace[i] = vertexFaces[vIndex];
			addFace->adjEdge[i] = vertexEdges[vIndex];

			// then set the start of the list to point to this edge
			vertexFaces[vIndex] = addFace;
			vertexEdges[vIndex] = i;
		}
	}
}

Void MeshJoin::SetFaceChainVertices(Face *f, Int edge, Int i)
// Sets the colour index to i of all faces connected to f around
// the point at the end of 'edge'.
{
	if (!f)
		return;
		
	Int nextEdge;

	// find the next face around the point
	nextEdge = NextTriEdge(edge);
#ifdef MJ_SURF_PT
	f->clrIdx[nextEdge] = i;
#endif
	
	SetFaceChainVertices(f->adjFace[nextEdge], f->adjEdge[nextEdge], i);
}

Void MeshJoin::EndGroup()
{
	Int		i, n;
	FacePtr	curFace, nextFace;
	Edge	curEdge, nextEdge;
	
	// Finally, we pass through the vertex face lists again, and
	// set the 'adjacent' fields of all unmatched edges to 0,
	// to mark that they're not connected to anything.
	// (We have to do this because we've been using these fields
	// to help store the unconnected edge list.)

	// For vertices that share several disconnected faces (i.e., there
	// is a surface discontinuity adjacent to the edge), we must make
	// sure that each disconnected segment has its own colour/normal
	// corresponding to that vertex. Otherwise we end up using the
	// same colour for two faces that share a vertex but are not
	// continuous.

	for (i = 0; i < numVertices; i++)
	// Cycle through the remaining unmatched edges.
	{
		// head of the edge list for this vertex
		curFace = vertexFaces[i];
		curEdge = vertexEdges[i];
				
		while (curFace)
		{
			// Store next edge...
			nextFace = curFace->adjFace[curEdge];
			nextEdge = curFace->adjEdge[curEdge];

			// clear adjFace pointer
			curFace->adjFace[curEdge] = 0;

			// move on to next edge in the list
			curFace = nextFace;
			curEdge = nextEdge;

			// Fill in a new colour index for all faces connected to this edge
			// around this vertex.
			// (Need to do this for all unmatched edges but the first one,
			// which retains its original setting.)
#ifdef MJ_SURF_PT
			if (curFace)
				SetFaceChainVertices(curFace, curEdge, surfaceVertices++);
#endif
		}
	}
}

Int MeshJoin::Finished()
{
	delete[] vertexFaces;
	delete[] vertexEdges;
	
	return(surfaceVertices);
}

