/*
	File:		FCElem.cc

	Function:	Elements that implement face cluster radiosity

	Author:		Andrew Willmott

	Notes:		
				
*/

#include "FCElem.h"

#include <stdio.h>
#include "Samples.h"
#include "HierRad.h"
#include "VecLink.h"

#include "gcl/MRModel.h"
#include "gcl/VecUtil.h"
#include "gcl/Draw.h"

#define DBG_COUT if (0) cerr

// XXX add to VL?

inline TMReal dot(TMVec3 &a, const TVec3 &b)
{
	return(a[0] * b[0] + a[1] * b[1] + a[2] * b[2]);
}

inline Double dot(Vec3d &a, const Vec3f &b)
{
	return(a[0] * b[0] + a[1] * b[1] + a[2] * b[2]);
}

Vec3f &operator += (Vec3f &a, const Vec3d &b)
{
	a[0] += b[0];
	a[1] += b[1];
	a[2] += b[2];
}


// --- FCElem methods -----------------------------------------------------


FCElem::FCElem(Int myFCI, MRInfo *myInfo) : 
	fci(myFCI), 
	info(myInfo), 
	fc(myInfo->model->Cluster(myFCI)),
	B(vl_0),
	R(vl_0),
	HRElem()
{
	
	Float	totSampArea;
	GCLReal temp;
	Point	diff;
	
	// find stuff in global coords...

	flags.Set(hrFaceClus);
	child[0] = 0;
	child[1] = 0;

	DBG_COUT << "hydrating face cluster " << fci << endl;
	
	fc.GetAxes(axes);
	centre = ToWorld(0.5 * (fc.min + fc.max));

	areaNormal = (info->orient * fc.areaNormal) * info->areaScale;
	projArea = len(areaNormal);

	totArea = info->areaScale * fc.totArea;

	diff = fc.max - fc.min;
	temp = info->areaScale / 255.0;
	sideArea[0] = (diff[1] * diff[2]) * fc.sideArea[0] * temp;
	sideArea[1] = (diff[2] * diff[0]) * fc.sideArea[1] * temp;
	sideArea[2] = (diff[0] * diff[1]) * fc.sideArea[2] * temp;
	sideArea[3] = (diff[1] * diff[2]) * fc.sideArea[3] * temp;
	sideArea[4] = (diff[2] * diff[0]) * fc.sideArea[4] * temp;
	sideArea[5] = (diff[0] * diff[1]) * fc.sideArea[5] * temp;

	totSampArea = sideArea[0] + sideArea[1] + sideArea[2]
		 + sideArea[3] + sideArea[4] + sideArea[5];
	maxVisArea = Min(totSampArea, totArea);
	
	DBG_COUT << "areaNormal = " << areaNormal << endl;
	DBG_COUT << "projArea = " << projArea << endl;
	DBG_COUT << "totArea = " << projArea << endl;
	DBG_COUT << "totSampArea = " << totSampArea << endl;
	DBG_COUT << "areaScale = " << info->areaScale << endl;

	Assert(projArea <= totArea, "bad proj/tot areas");
}

Void FCElem::Print(ostream &s)
{
	HRLinkIter i;
	
//	s << "level " << mrv->level << " vertex: " << endl;
	s << "links: ";
	for (i.Begin(links); !i.AtEnd(); i.Inc())
		s << i.Data();
		
	s << endl;
}

Void FCElem::DrawNodeElem(Renderer &r)
{
	if (flags.IsSet(hrMark))
	{
		flags.UnSet(hrMark);
		return;
	}

	if (IsLeaf()
#ifdef RAD_VIS
		|| eltHighlight > 0
#endif
		)
		DrawLeafElem(r);
	else
	{
		child[0]->DrawElem(r);
		child[1]->DrawElem(r);
	}
}

Void FCElem::EltUpdateBounds(Point &min, Point &max)
{
	Point	corners[8];
	Int		i;

	MakeBoxCorners(fc.min, fc.max, corners);

	for (i = 0; i < 8; i++)
		::UpdateBounds(ToWorld(corners[i]), min, max);
}

// --- Radiosity transfer -----------------------------------------------------


Colour *FCElem::B_Coeffs()
{
	return(&B);
}

Void FCElem::Push()
{
	if (child[0]->IsPatch())
		child[0]->ClearB();
	child[0]->AddChanIrradiance(R[0], 0);
	child[0]->AddChanIrradiance(R[1], 1);
	child[0]->AddChanIrradiance(R[2], 2);

	if (child[1]->IsPatch())
		child[1]->ClearB();
	child[1]->AddChanIrradiance(R[0], 0);
	child[1]->AddChanIrradiance(R[1], 1);
	child[1]->AddChanIrradiance(R[2], 2);
}

Void FCElem::Pull()
{
	B =  child[0]->EltBA();
	B += child[1]->EltBA();
	B /= totArea;
}

Colour FCElem::lastB;

Void FCElem::Add()
// We use R for irradiance accumulation, so nothing for the add stage.
{
}

Void FCElem::ClearB()
{
	lastB = B;
	B = vl_0;
}

Void FCElem::ClearR()
{
	R = vl_0;
}

Colour FCElem::EltRho()
{
	return(info->refl);
}

Void FCElem::CalcLeafRadiosity()
{
	// transform vector irradiance to radiosity

	for (Int i = 0; i < 3; i++)
		B[i] = info->refl[i] * Max(0.0, dot(areaNormal, R[i])) 
			+ totArea * info->emit[i];
	B /= maxVisArea;	// correct for internal reflectance...
}

GCLReal FCElem::Error()
{
	Colour		errB = B - lastB;
	GCLReal		result;
	
	result = dot(kRadRGBToLum, errB);

 	return(result);
}

// --- Link handling ----------------------------------------------------------

Void FCElem::MakeChildLinks(HRElem *other, HRLink *link,
							  Int which, Int levels)
// Splits a link into sublinks that point to the children of
// either 'from' or 'to'.
{
	Int i;

	if (IsLeaf())
	{
		Subdivide();
		Push();
		child[0]->CalcLeafRadiosity();
		child[1]->CalcLeafRadiosity();
	}

	if (which == kSubTo)
	{
		// add to edge children
		child[0]->RefineLink(
			link->CreateSubLink(other, child[0]),
			levels);
		child[1]->RefineLink(
			link->CreateSubLink(other, child[1]),
			levels);
	}
	else if (which == kSubFrom)
	{
		// add to edge children
		other->RefineLink(
			link->CreateSubLink(child[0], other),
			levels);
		other->RefineLink(
			link->CreateSubLink(child[1], other),
			levels);
	}
	else
	{
		other->MakeChildLinks(child[0], link, kSubFrom, levels);
		other->MakeChildLinks(child[1], link, kSubFrom, levels);		
	}
}

// --- children abstraction ---------------------------------------------------


Void FCElem::ApplyToChildren(Void (HRElem::*method)(Void*), Void *ref)
{
	if (IsLeaf())
		return;
	(child[0]->*method)(ref);
	(child[1]->*method)(ref);
}


// --- sampling stuff ---------------------------------------------------


GCLReal FCElem::EltMaxProjArea(const Vector &m)
{
	Vector		mLoc;
	GCLReal		result;
	
	mLoc = axes * (m * info->orient);
	result = 0.0;

	if (mLoc[0] > 0.0)
		result += mLoc[0] * sideArea[0];
	else
		result -= mLoc[0] * sideArea[3];
	if (mLoc[1] > 0.0)
		result += mLoc[1] * sideArea[1];
	else
		result -= mLoc[1] * sideArea[4];
	if (mLoc[2] > 0.0)
		result += mLoc[2] * sideArea[2];
	else
		result -= mLoc[2] * sideArea[5];

	{
		GCLReal lb = EltProjArea(m);

		if (0.95 * lb > result)
		{
			cerr << "BAD BOUND: " << result << " " << lb << endl;
			cerr << "areas: " << len(areaNormal) << ' ' << maxVisArea << ' ' << totArea << endl;
			cerr << sideArea[0] << ' '
				<< sideArea[1] << ' '
				<< sideArea[2] << ' '
				<< sideArea[3] << ' '
				<< sideArea[4] << ' '
				<< sideArea[5] << ' '
				<< endl;
				
			cerr << "m: " << m << endl;
			cerr << "mLen: " << len(m) << endl;
			cerr << "mLoc: " << mLoc << endl;
			cerr << "mLocLen: " << len(mLoc) << endl;
			cerr << "mLocNorm: " << safe_norm(mLoc) << endl;
			cerr << "S: " << areaNormal << endl;
			cerr << "SLoc: " << axes * (areaNormal * info->orient) << endl;
			cerr << "SLocNorm: " << safe_norm(axes * (areaNormal * info->orient)) << endl;
			Assert(false, "stopping");
		}
//	cout << "lb = " << EltProjArea(m) << endl;
//	cout << "ub = " << result << endl;
	}
	
	return(result);
}

Void FCElem::EltSampleTransport(
					Int 		numSamples, 
					Point 		p[],
					Vector 		n[],
				   	Matd 		&coeffs
				)
{
	Int		i;
	GCLReal	result, tempLen;
	Vector	temp;

	// A is in src coeffs already...
	// calc  (n[is].^r) r / pi.r^2

	for (i = 0; i < numSamples; i++)
	{
		temp = centre;
		temp -= p[i];

		result = dot(n[i], temp);
		if (result <= 0.0)
		{
			coeffs[i][0] = 0.0;
			continue;
		}
		tempLen = len(temp);
		if (tempLen > 0.0)
			result /= (sqr(tempLen) * tempLen * vl_pi);
	
		coeffs[i][0] = result;
	}
}

GCLReal FCElem::EltCalcTransport(HRElem *from, Matd &coeffs)
{	
	if (from == this)
	{
		coeffs[0][0] = 0.0;
		return(0.0);
	}

	_Error("Shouldn't be called?");

	return(-1);
}

Void FCElem::AddIrradiance(const Colour &E, const Vector &m)
{
	GCLReal		temp;

	// add irradiance vector E * m into R, clipping to the
	// plane of the face cluster.

	temp = dot(m, areaNormal);
	if (temp > 0.0)
		for (Int i = 0; i < 3; i++)
			R[i] += m * E[i];
}

Void FCElem::AddChanIrradiance(const Colour &E, Int chan)
{
	if (dot(areaNormal, E) > 0.0)
		R[chan] += E;
}

Colour FCElem::GetPower(const Vector &m)
{
	return(B * EltProjArea(m));
}

Colour FCElem::EltBA()
{
	Colour	result;

	result = B * totArea;

	return(result);
}


// --- Face Cluster methods ---------------------------------------------------



Void FCElem::DrawLeafElem(Renderer &r)
{
#ifdef RAD_VIS
	r.Push();

	r.SetTransform(info->trans);
	r.N(vl_z);

	if (eltHighlight > 0)
	{
		r.C(cGreen);
		fc.DrawBox(r, false);
	
		if (eltHighlight == 1)
			r.C(Colour4(cRed, 0.5));
		else if (eltHighlight == 2)
			r.C(Colour4(cYellow, 0.5));
		else if (eltHighlight == 3)
			r.C(Colour4(cGreen, 0.5));
		else
			r.C(Colour4(cPurple, 0.5));

		fc.DrawBox(r, true);
	}
	else
	{
		r.C(B);
		fc.DrawBox(r, !gRadControl->wire);
	}
	fc.DrawBox(r, !gRadControl->wire
		 || eltHighlight > 0
		 );

	r.Pop();
#endif
}

Void FCElem::Subdivide()
{
	if (info->model->IsFace(fc.child[0]))
	{
		DBG_COUT << "SWAPPING TO HRELEM [0]" << endl;
		child[0] = dynamic_cast<HierRad*>(gRadControl->radObject)->AddFCLeaf(
			info, fc.child[0]);
	}
	else
		child[0] = new FCElem(fc.child[0], info);

	if (info->model->IsFace(fc.child[1]))
	{
		DBG_COUT << "SWAPPING TO HRELEM [1]" << endl;
		child[1] = dynamic_cast<HierRad*>(gRadControl->radObject)->AddFCLeaf(
			info, fc.child[1]);
	}
	else
		child[1] = new FCElem(fc.child[1], info);

#ifdef RAD_VIS
	info->elemPtrs[fc.child[0]] = child[0];
	info->elemPtrs[fc.child[1]] = child[1];
	child[0]->eltParent = this;
	child[1]->eltParent = this;
#endif
}

// --- visibility estimation --------------------------------------------------

static Int gRotPick = 0;

Void FCElem::EltSetVisPoints(HRElem *to, Point *pts)
{
	Int 		i;
	Point		centre;
	Vector		r;
	Coord 		*d1, delta;
	
	delta = Coord(fc.min[0], fc.min[1]) - Coord(fc.max[0], fc.max[1]);

 	d1 = (Coord *) tSquareSamples16[gRotPick]; 

	for (i = 0; i < 16; i++)
		pts[i] = ToWorld(fc.max + Vector(d1[i] * delta, 0.0));
	
	if (gRadControl->jitterRot)
		if (++gRotPick == kNumSampArrays)
			gRotPick = 0;
}

Void FCElem::EltGetSamples(Int numSamples, Point p[])
{
	Vector		d, *offsets = (Vector*) tCubeSamples;
	Int			i;
	
	d = fc.max - fc.min;

	for (i = 0; i < numSamples; i++)
		p[i] = ToWorld(fc.min + offsets[i] * d);
}


// --- Colouring final model after simulation ----------------------------------


Void FCElem::ColourFaces(MRModel *model, Int fci)
// utility routine to colour all the faces belonging to a (solution)
// leaf cluster.
{
	if (!model->IsFace(fci))
	{
		FaceCluster &fc = model->Cluster(fci);

		ColourFaces(model, fc.child[0]);
		ColourFaces(model, fc.child[1]);

		return;
	}

	Vector		result;
	Vector		normal;
	Int			i;
	FaceIdx		&tri = model->faces[fci];
	GCLReal		area;
	Colour		clr, clrNorm;

	// find face's normal & area in WCS
	normal = (info->orient * model->FaceAreaNormal(fci)) * info->areaScale;
	area = len(normal);
	normal /= area;
	ConvertVec(normal, clrNorm);
	
	// leaf irradiance is in R
	// transform irradiance vector to radiosity of face.
	clr = info->refl * (R * clrNorm) + info->emit;
	ClipColourZero(clr);

	// paint corresponding face vertices
	for (i = 0; i < 3; i++)
	{
		info->colours[tri.v[i]] += clr;
		info->weights[tri.v[i]] += 1;
	}
}

struct FCShadeInfo
{
	Void		Init();
	
	IrradVec	R[4];
};

Void FCShadeInfo::Init()
{
	R[0] = vl_0;
	R[1] = vl_0;
	R[2] = vl_0;
	R[3] = vl_0;
}

const Double kVIInset = 0.3;
const Point kFC1(kVIInset, kVIInset, 0.0);
const Point kFC2(1.0 - 2.0 * kVIInset, 1.0 - 2.0 * kVIInset, 1.0);

Void FCElem::ColourFacesBest(MRModel *model, Int fci, const FCShadeInfo &si)
{
	if (!model->IsFace(fci))
	{
		FaceCluster &fc = model->Cluster(fci);

		ColourFacesBest(model, fc.child[0], si);
		ColourFacesBest(model, fc.child[1], si);

		return;
	}

	Vector		result;
	Vector		normal;
	Int			i;
	FaceIdx		&tri = model->faces[fci];
	GCLReal		area, temp;
	Colour		clr, clrNorm;
	IrradVec	sR, sR0, sR1;
	Point		fcp, centre;
	
	// find face's normal & area in WCS
	normal = (info->orient * model->FaceAreaNormal(fci)) * info->areaScale;
	area = len(normal);
	normal /= area;
	ConvertVec(normal, clrNorm);

	centre = model->points->Item(tri.v[0]);
	centre += model->points->Item(tri.v[1]);
	centre += model->points->Item(tri.v[2]);
	centre /= 3.0;
	
	// find interpolated irradiance vector

	// transform to bbox coords
	fcp = axes * (centre * info->orient);
	fcp -= fc.min;
	temp = fc.max[0] - fc.min[0];
	if (temp > 0.0)
		fcp[0] /= temp;
	temp = fc.max[1] - fc.min[1];
	if (temp > 0.0)
		fcp[1] /= temp;
	
	fcp = (fcp - kFC1) / kFC2;
	
	sR0 = Mix(si.R[0], si.R[1], fcp[0]);
	sR1 = Mix(si.R[3], si.R[2], fcp[0]);
	sR = Mix(sR0, sR1, fcp[1]);
	
	// transform irradiance vector to radiosity of face.
	clr = info->refl * (sR * clrNorm) + info->emit;
	ClipColourZero(clr);

	// paint corresponding face vertices
	for (i = 0; i < 3; i++)
	{
		info->colours[tri.v[i]] += clr;
		info->weights[tri.v[i]] += 1;
	}
}

Void FCElem::DistributeColours()
{
	if (!gRadControl->finalPass || !gRadControl->pushPass)
		return;

	// Here we apply the leaf element irradiances to the leaf face clusters

	if (IsLeaf())
		ColourFaces(info->model, fci);
	else
	{
		child[0]->DistributeColours();
		child[1]->DistributeColours();
	}
}

Void FCElem::CalcShadeInfo(FCShadeInfo &si, const HRLinkList &links)
/// calculate corner irradiance vectors from the given list of links
{
	Point		corners[8], p[4];
	Int			i, j;
	GCLReal		temp;
	HRLinkIter	li;
	Vector		m;
	Colour		E;
	
	MakeBoxCorners(fc.min, fc.max, corners);
#ifdef FCR_MID
	p[0] = ToWorld(0.5 * (corners[0] + corners[4]));
	p[1] = ToWorld(0.5 * (corners[1] + corners[5]));
	p[2] = ToWorld(0.5 * (corners[2] + corners[6]));
	p[3] = ToWorld(0.5 * (corners[3] + corners[7]));
#else
	p[0] = ToWorld(Mix(corners[4], corners[5], kVIInset));
	p[1] = ToWorld(Mix(corners[5], corners[6], kVIInset));
	p[2] = ToWorld(Mix(corners[6], corners[7], kVIInset));
	p[3] = ToWorld(Mix(corners[7], corners[4], kVIInset));
#endif

	for (j = 0; j < 4; j++)
		for (li.Begin(links); !li.AtEnd(); li.Inc())
		{
			VecLink	*link = (VecLink*) &li.Data();
			// add irradiance vector E * m into R, clipping to the
			// plane of the face cluster.

			m = link->PointIrradiance(p[j], areaNormal, E);
			temp = dot(m, areaNormal);
			if (temp > 0.0)
				for (Int i = 0; i < 3; i++)
					si.R[j][i] += m * E[i];
		}
}

struct FCSIRec
{
	FCShadeInfo		si;
	FCElem			*elem;
};

typedef NArray<FCSIRec> FCSIStack;

Void FCElem::CalcShadeInfo(FCShadeInfo &si, const FCSIRec &parIrr)
/// calculate the corner irradiance vectors from the parent's irradiance vectors
{
	Point		corners[8], p[4];
	Int			i, j;
	Coord		delta;
	IrradVec	sR, sR0, sR1;
	Point		fcp;
	
	MakeBoxCorners(fc.min, fc.max, corners);
#ifdef FCR_MID
	p[0] = ToWorld(0.5 * (corners[0] + corners[4]));
	p[1] = ToWorld(0.5 * (corners[1] + corners[5]));
	p[2] = ToWorld(0.5 * (corners[2] + corners[6]));
	p[3] = ToWorld(0.5 * (corners[3] + corners[7]));
#else
	p[0] = ToWorld(Mix(corners[4], corners[5], kVIInset));
	p[1] = ToWorld(Mix(corners[5], corners[6], kVIInset));
	p[2] = ToWorld(Mix(corners[6], corners[7], kVIInset));
	p[3] = ToWorld(Mix(corners[7], corners[4], kVIInset));
#endif

	delta[0] = parIrr.elem->fc.max[0] - parIrr.elem->fc.min[0];
	delta[1] = parIrr.elem->fc.max[1] - parIrr.elem->fc.min[1];

	for (j = 0; j < 4; j++)
	{
		// find interpolated parent irradiance at corner j
		// transform to bbox coords
		fcp = parIrr.elem->axes * (p[j] * info->orient);
		fcp -= parIrr.elem->fc.min;
		if (delta[0] > 0.0)
			fcp[0] /= delta[0];
		if (delta[1] > 0.0)
			fcp[1] /= delta[1];
		
		fcp[0] = Clip(fcp[0], 0.0, 1.0);
		fcp[1] = Clip(fcp[1], 0.0, 1.0);
		fcp = (fcp - kFC1) / kFC2;

		sR0 = Mix(parIrr.si.R[0], parIrr.si.R[1], fcp[0]);
		sR1 = Mix(parIrr.si.R[3], parIrr.si.R[2], fcp[0]);
		sR = Mix(sR0, sR1, fcp[1]);

		for (i = 0; i < 3; i++)
			if (dot(areaNormal, sR[i]) > 0.0)
				si.R[j][i] += sR[i];
	}	
}

Void FCElem::DistributeColoursBest(ShadeInfo &shadeInfo)
{
	if (!gRadControl->pushPass)
		return;

	FCSIStack	*fcsiStack = (FCSIStack*) shadeInfo.data;

	if (!fcsiStack && gRadControl->bestLevels < 0)
	{
		fcsiStack = new FCSIStack;
		shadeInfo.data = fcsiStack;
	}

	// Here we apply the leaf element irradiances to the leaf face clusters

	if (!links.IsEmpty())
	{
		shadeInfo.linkStack.Push(&links);

		if (fcsiStack)
		{
			FCSIRec	frec;
			
			frec.si.Init();
			if (fcsiStack->NumItems() > 0)
				CalcShadeInfo(frec.si, fcsiStack->Last());
			CalcShadeInfo(frec.si, links);
			frec.elem = this;
			
			fcsiStack->Append(frec);
		}	
	}
	
	if (IsLeaf() && shadeInfo.level >= gRadControl->bestLevels * 2)
	{
		Int i, nlinks = 0;
		HRLinkIter	li;
		
#ifdef DEBUG
		for (i = 0; i < shadeInfo.linkStack.NumItems(); i++)
		{
			HRLinkIter j;
			nlinks += shadeInfo.linkStack[i]->NumItems();
		}
		DBG_COUT << "colouring face cluster " << id << ": " << shadeInfo.linkStack.NumItems() 
			<< " levels of links, " << nlinks << " links." << endl;
#endif

		if (false)	
		// reevaluate constant irr. vector at leaves
		{
			// reevaluate irradiance vector using all links...
			// for each link...
			R = vl_0;
			for (i = 0; i < shadeInfo.linkStack.NumItems(); i++)
				for (li.Begin(*shadeInfo.linkStack[i]); !li.AtEnd(); li.Inc())
				{
					li.Data().to = this;	// is this safe?
					// could call vis here to reevaluate?
					li.Data().CalcTransport();
					li.Data().Gather();
				}

			ColourFaces(info->model, fci);
		}
		else if (!fcsiStack)
		// irradiance interpolation at leaves
		{
			FCShadeInfo fcsi;
		
			fcsi.Init();

			for (i = 0; i < shadeInfo.linkStack.NumItems(); i++)
				CalcShadeInfo(fcsi, *shadeInfo.linkStack[i]);

			ColourFacesBest(info->model, fci, fcsi);
		}
		else
		// irradiance interpolation through the hierarchy
		{
			Assert(fcsiStack->Last().elem == this, "leaf has no links");
			ColourFacesBest(info->model, fci, fcsiStack->Last().si);
		}
	}
	else
	{
		Bool amSubdividing = IsLeaf();
		
		if (amSubdividing)
		{
			shadeInfo.level++;
			Subdivide();
		}

		child[0]->DistributeColoursBest(shadeInfo);
		child[1]->DistributeColoursBest(shadeInfo);

		if (amSubdividing)
			shadeInfo.level--;
	}

	if (!links.IsEmpty())
	{
		shadeInfo.linkStack.Pop();
		if (fcsiStack)
			fcsiStack->Pop();
	}
}

Void FCElem::Reset()
{
	HRElem::Reset();

	if (!IsLeaf())
	{
		child[0]->Reset();
		child[1]->Reset();

		delete child[0];
		delete child[1];
		child[0] = 0;
	}
}

Void FCElem::DebugInfo()
{
	cerr << "R = " << R << endl;
	cerr << "S = " << areaNormal << endl;
	cerr << "areas: " << len(areaNormal) << ' ' << maxVisArea << ' ' << totArea << endl;
	HRElem::DebugInfo();
}

// --- Extracting MR models from a scene ---------------------------------------


Void ModelsExtractor::Start()
{
	Int		i;

	MRModelsFinder::Start();
	
	info.FreeAll();
	info.Clear();
}

Void ModelsExtractor::Primitive(scPrimitive *sp)
{
	MRModelsFinder::Primitive(sp);

	// build info as well.
	if (sp->PrimID() == pMRModel)
	{
		Colour		*reflectancePtr = SC_GET(Colour);
		Colour		*emittancePtr = SC_GET(Emittance);
		MRInfo		*newMRInfo = new MRInfo;
		GCLReal		checkScale;
		
		newMRInfo->trans = transAcc;
		GCLMat(3, 3, newMRInfo->orient.Ref()) = sub(transAcc, 3, 3);
		newMRInfo->areaScale = sqrlen(col(transAcc, 0));
		checkScale = sqrlen(col(transAcc, 1));
		Assert(abs(checkScale - newMRInfo->areaScale) < 1e-6, "non uniform scaling for model.");
		newMRInfo->orient /= sqrt(newMRInfo->areaScale);

		newMRInfo->model = models.Last();

		DBG_COUT << "new model:" << endl; 
		DBG_COUT << "trans " << endl << newMRInfo->trans; 
		DBG_COUT << "orient" << endl << newMRInfo->orient;
		DBG_COUT << "areaScale " << newMRInfo->areaScale << endl;
		DBG_COUT << "areaScaleCheck " << checkScale << endl;

		if (reflectancePtr)
			newMRInfo->refl = *reflectancePtr;
		else
			newMRInfo->refl = cOrange;		
		if (emittancePtr)
			newMRInfo->emit = *emittancePtr;
		else
			newMRInfo->emit = cBlack;		

#ifdef RAD_VIS
		newMRInfo->colours.SetSize(models.Last()->points->NumItems());
		for (Int i = 0; i < newMRInfo->colours.NumItems(); i++)
			newMRInfo->colours[i] = Colour(1, i % 2, 0);
		((scMRModel*) sp)->model.colours = &newMRInfo->colours;

		newMRInfo->elemPtrs.SetSize(models.Last()->faces.NumItems()
			+ models.Last()->clusters.NumItems());
//		newMRInfo->elemPtrs.ClearTo(0);
		for (Int i = 0; i < newMRInfo->elemPtrs.NumItems(); i++)
			newMRInfo->elemPtrs[i] = 0;
#endif
		info.Append(newMRInfo);
	}
}


MRInfo::~MRInfo()
{
	rtObj.Free();
}

#ifdef RAD_VIS
HRElem *MRInfo::FindElem(Int fci)
{
	Int		i, idx;

	if (elemPtrs[fci])
		return(elemPtrs[fci]);

	idx = fci;

	for (i = model->faces.NumItems(); i < model->faces.NumItems() + model->clusters.NumItems(); i++)
	{
		FaceCluster		&fc = model->Cluster(i);

		if (fc.child[0] == idx || fc.child[1] == idx)
		{
			if (elemPtrs[i])
				return(elemPtrs[i]);
			idx = i;
		}
	}
	
	return(0);
}


#endif
