/*
	File:		HRElem.cc

	Function:	Defines the element type used by the various hierarchical
				radiosity methods.

	Author:		Andrew Willmott

	Notes:		
*/

#include "HRElem.h"
#include "RadMethod.h"
#include "gcl/SceneLang.h"
#include "gcl/GLRenderer.h"

#ifdef RAD_VIS
#include "gcl/ScenePane.h"
#endif

// --- HRElem methods ---------------------------------------------------------

UInt32 HRElem::gID = 0;

HRElem::HRElem() :
#ifdef RAD_VIS
	eltParent(0),
	eltHighlight(0),
#endif
	id(gID++),
	flags()
{
}	

HRElem::~HRElem()
{
	Reset();
}

Void HRElem::Reset()
{
	links.Free();
}

Void HRElem::DrawElem(Renderer &r)
{
	DrawNodeElem(r);
}

// XXX shift?

Void DrawArrow(Renderer &r, Point start, Point end)
{
	static scPrimitive	*sCone = 0;
	static scPrimitive	*sShaft = 0;
	const Double		ac = 0.1;
	const Double		as = ac / 5.0;

	GCLReal				al = len(end - start);
	GLRenderer			&glr = (GLRenderer &) r;
	Transform 			trans;
	
	if (al < 1e-15)
		return;

	if (!sCone)
	{
		sCone = slGetLibrary()->Member("cone");
		sShaft = slGetLibrary()->Member("cylinder");
	}
	
	glr.SetHeadlight(true);

#ifdef OFF
	r.C(cYellow).Begin(renLines);
	r.P(start).P(end);
	r.End();
#endif

	r.C(cWhite);
	r.Push();
	trans = Shift(end) * AlignToDir(end - start);
	r.SetTransform(trans);

	r.Push();
	trans = Shift(-ac * Vector(vl_y)) * Scale(Vector(as, al - ac, as)) * Shift(-0.5 * Vector(vl_y));
	r.SetTransform(trans);
	sShaft->Draw(r);
	r.Pop();

	r.Push();
	trans = Scalef(ac) * Shift(-Vector(vl_y));
	r.SetTransform(trans);
	sCone->Draw(r);
	r.Pop();
	
	r.Pop();
	glr.SetHeadlight(false);
}

Void HRElem::DrawContributors(Renderer &r)
{
	HRLinkIter iter;

	for (iter.Begin(links); !iter.AtEnd(); iter.Inc())
		DrawArrow(r, iter.Data().from->EltCentre(), iter.Data().to->EltCentre());

	for (iter.Begin(links); !iter.AtEnd(); iter.Inc())
	{
		iter.Data().from->DrawLeafElem(r);
		iter.Data().from->flags.Set(hrMark);
	}
}

Void HRElem::DrawContributorsRec(Void *r)
{
	Int i;
	
	DrawContributors(*(Renderer*) r);
	ApplyToChildren(&HRElem::DrawContributorsRec, r);
}

Void HRElem::DrawMatrix(Renderer &r, Int baseNum)
{
	DMInfo	info;

	info.r = &r;
	info.baseNum = baseNum;
	DrawMatrixRec(&info);
}

Void HRElem::DrawMatrixRec(Void *dmi)
{
}

Void HRElem::DistributeColours()
{
}

Void HRElem::DistributeColoursBest(ShadeInfo &shadeInfo)
{
}

// --- Link Refinement --------------------------------------------------------

//#define RAD_VIS_LINKS
// doesn't seem to have turned out to be all that useful?

Bool HRElem::RefineLink(HRLink *link, Int levels)
/** Recursively refines a particular link, according to
	RefineOracle. */
{
	Int			i, j, which;
	GCLReal		factor, visibility;

	// CreateSubLink() returns null if the link carries no energy, so
	// return if that's the case.
	if (link == 0)
		return(false);

	if (levels-- == 0)
	{
		links.Prepend(link);
		return(false);
	}
	
	which = link->RefineOracle();
	
#ifdef RAD_VIS_LINKS
	if (gRadControl->showLinks)
	{
		RM_DISPLAY_START;
		gRadControl->radObject->display->Begin(renLines);
	}
#endif
		
	if (which == kSubNone)			// Form link here.
	{
		links.Prepend(link);
		
#ifdef RAD_VIS_LINKS
		if (gRadControl->showLinks)
		{
			gRadControl->radObject->display->
				C(cRed).P(EltCentre()).P(link->from->EltCentre()).End();
			RM_DISPLAY_END;
			RM_OUT1("Forming link here.");
			if (RM_PAUSE) return(false);
		}
#endif
		return(false);
	}
	else if (which == kSubTo)		// Subdivide 'to' patch (this one!)
	{
#ifdef RAD_VIS_LINKS
		if (gRadControl->showLinks)
		{
			gRadControl->radObject->display->
				C(cGreen).P(EltCentre()).P(link->from->EltCentre());
			RM_DISPLAY_END;
			RM_OUT1("Pushing link down to children of receiver");
			if (RM_PAUSE) return(false);
		}
#endif		
		MakeChildLinks(link->from, link, which, levels);
	}
	else if (which == kSubFrom)		// Subdivide 'from' patch
	{
#ifdef RAD_VIS_LINKS
		if (gRadControl->showLinks)
		{
			gRadControl->radObject->display->
				C(cBlue).P(EltCentre()).P(link->from->EltCentre()).End();
			RM_DISPLAY_END;
			RM_OUT1("Pushing link down to children of source");
			if (RM_PAUSE) return(false);
		}
#endif
		link->from->MakeChildLinks(this, link, which, levels);
	}		
	else if (which == kSubBoth)		// Subdivide both patches
	{
#ifdef RAD_VIS_LINKS
		if (gRadControl->showLinks)
		{
			gRadControl->radObject->display->
				C(cYellow).P(EltCentre()).P(link->from->EltCentre()).End();
			RM_DISPLAY_END;
			RM_OUT1("Pushing link down to children of source & dest");
			if (RM_PAUSE) return(false);
		}
#endif
		MakeChildLinks(link->from, link, which, levels);
	}		

	link->FreeLink(link);

	return(true);
}

Bool HRElem::Refine()
/**	Refines any links whose error is too high, both in this element and
	its children. Returns true if any links were refined. */
{
	Bool result = false;
	RefineFurther(&result);
	return(result);
}

Void HRElem::RefineFurther(Void *result)
{
	HRLinkIter	i;
	Int 		j, which;
	Bool		split;
	HRLinkList	oldLinks;
	HRLink		*tempLink, *rlink;
	
	// bottom-up traversal to avoid refining a link more than once.
	ApplyToChildren(&HRElem::RefineFurther, result);
	
	rlink = (HRLink*) links.first;
 	links.first = 0;

	// Step through links, see whether any should be pushed down the hierarchy.
	while (rlink)
	{
		tempLink = rlink;
		rlink = (HRLink*) rlink->next;
		if (RefineLink(tempLink, gRadControl->refineLevels))
			*(Bool*) result = true;
	}
}

Void HRStats::Init()
{
	numLinks = 0;
	numPatches = 0;
	numVolClusters = 0;
	numFaceClusters = 0;
	numAliens = 0;
}

Void HRElem::CalcStats(Void *statsIn)
{
	HRLinkIter 	i;
	HRStats		*stats = (HRStats*) statsIn;

	for (i.Begin(links); !i.AtEnd(); i.Inc())
		stats->numLinks++;
	
	if (IsPatch())
		stats->numPatches++; 
	else if (IsCluster())
		stats->numVolClusters++;
	else if (IsFaceClus())
		stats->numFaceClusters++;
	else
		stats->numAliens++;
		
	ApplyToChildren(&HRElem::CalcStats, statsIn);
}

ostream *HRElem::out = &cout;

Void HRElem::DumpHierarchy(Void *v)
// dump the element hierarchy 
{
	if (v)
	{
		// just printing a list of child IDs
		*out << id << ' ';
		return;
	}

	// dump from leaves to root
	ApplyToChildren(&HRElem::DumpHierarchy);	

	HRLinkIter i;
	// elt. format: e ID [childID1 childID2...]
	*out << "e " << id << ' ';
	ApplyToChildren(&HRElem::DumpHierarchy, this);
	*out << endl;
	// link format: L toEltID fromEltID transferCoeff
	for (i.Begin(links); !i.AtEnd(); i.Inc())
		*out << "L " << id << " " << i.Data().from->id 
			<< ' ' << i.Data().Strength() << endl;
}

Void HRElem::Gather()
{
	HRLinkIter i;
	
	ClearR();

	for (i.Begin(links); !i.AtEnd(); i.Inc())
	{
		i.Data().Gather();
		
#ifdef RAD_VIS_LINKS
		// YYY replace with arrow sometime?
		if (gRadControl->showLinks)
			gRadControl->radObject->display->Begin(renLines)
				.C(cRed).P(EltCentre()).P(i.Data().from->EltCentre())
				.End();
#endif
	}
}

Void HRElem::GatherAll(Void *)
{
	ApplyToChildren(&HRElem::GatherAll);
	Gather();    // Gather radiosity across all links to this patch into R
}

#ifdef RAD_VIS

Void HRElem::PushPull(Void *)
{
	Int		i;
				
	if (gRadControl->animate)
	{
		SetHighlight(0);
		if (RM_PAUSE) return;
	}
	
	Add();

	if (gRadControl->animate)
		RM_DISPLAY_SCENE;

	if (!IsLeaf())
	{
		if (gRadControl->animate)
		{
			if (RM_PAUSE) return;
		}
				
		Push();
		
		ApplyToChildren(&HRElem::PushPull);
		
		if (gRadControl->animate)
		{
	        SetHighlight(1);
    	    SetHighlight(0);

			if (RM_PAUSE) return;
		}

		Pull();

		if (gRadControl->animate)
			RM_DISPLAY_SCENE;
	}
	else
		CalcLeafRadiosity();
}

#else

Void HRElem::PushPull(Void *)
{
	Int		i;
			
	// Add R in to B: accumulated radiosity from this and higher levels
	Add();       
	
	if (!IsLeaf())
	{				
		// Push B down to children's B's
	    Push();  
		// recurse
		ApplyToChildren(&HRElem::PushPull);
		// Pull radiosity back up
		Pull();
	}
	else
		CalcLeafRadiosity();
}
#endif

GCLReal	HRElem::EltVisToPoint(Point &p)
{
	Int			i, misses;
	Point		srcPoint[16];
	Bool		hit;

	if (gRadControl->visibility == vis_none)
		return(1.0);
			
	misses = 0;
	EltSetVisPoints(0, srcPoint);
	
#ifdef RAD_VIS
	if (gRadControl->showRays)
		RM_DISPLAY_START;
#endif

	gRadControl->rays += 16;
	
	for (i = 0; i < 16; i++)
	{
		hit = gRadControl->radObject->IntersectsWithRay(srcPoint[i], p);
		
		if (hit)
			misses++;

#ifdef RAD_VIS
		if (gRadControl->showRays)
		{
			if (hit)
				gRadControl->radObject->display->Begin(renLines).C(cRed).P(srcPoint[i]).
					P(p).End();
			else
				gRadControl->radObject->display->Begin(renLines).C(cGreen).P(srcPoint[i]).
					P(p).End();
		}
#endif
	}
		
#ifdef RAD_VIS
	if (gRadControl->showRays)
	{
		RM_DISPLAY_END;
		RM_OUT1("Visibility: " << GCLReal(16 - misses) * 1.0 / 16.0);
		if (RM_PAUSE) return(0);
	}
#endif
		
	return((16 - misses) * (1.0 / 16.0));
}

GCLReal	HRElem::EltVis(HRElem *to)
{
	Int			i, misses;
	Point		srcPoint[16], dstPoint[16];
	Bool		hit;

	if (gRadControl->visibility == vis_none || to == this)
		return(1.0);
			
	misses = 0;
	EltSetVisPoints(to, srcPoint);
	to->EltSetVisPoints(this, dstPoint);
	
#ifdef RAD_VIS
	if (gRadControl->showRays)
		RM_DISPLAY_START;
#endif

	gRadControl->rays += 16;
	
	for (i = 0; i < 16; i++)
	{
		hit = gRadControl->radObject->IntersectsWithRay(srcPoint[i], dstPoint[i]);
		
		if (hit)
			misses++;

#ifdef RAD_VIS
		if (gRadControl->showRays)
		{
			if (hit)
				gRadControl->radObject->display->Begin(renLines).C(cRed).P(srcPoint[i]).
					P(dstPoint[i]).End();
			else
				gRadControl->radObject->display->Begin(renLines).C(cGreen).P(srcPoint[i]).
					P(dstPoint[i]).End();
		}
#endif
	}
		
#ifdef RAD_VIS
	if (gRadControl->showRays)
	{
		RM_DISPLAY_END;
		RM_OUT1("Visibility: " << GCLReal(16 - misses) * 1.0 / 16.0);
		if (RM_PAUSE) return(0);
	}
#endif
		
	return((16 - misses) * (1.0 / 16.0));
}

Colour HRElem::EltBA()
/*! Defaults to returning EltArea() times first coefficient. */
{
	return(B_Coeffs()[0] * EltArea());
}

Void HRElem::InitRad()
{
	ClearB();
	CalcLeafRadiosity();
}

Void HRElem::AddChanIrradiance(const Colour &E, Int chan)
{
	Colour  c(vl_0);
	Vector	m;

	c[chan] = 1.0;
	ConvertVec(E, m);

	AddIrradiance(c, m);
}

GCLReal	HRElem::EltMaxProjArea(const Vector &v)
{
	return(EltProjArea(v));
}

Void HRElem::DebugInfo()
{
	HRLinkIter i;

	cout << "radiosity = " << B_Coeffs()[0] << endl;
	cout << "rho = " << EltRho() << endl;
	cout << "flags = " << flags << endl;
	for (i.Begin(links); !i.AtEnd(); i.Inc())
		i.Data().DebugInfo();
}


GCLReal VecError(const SubVecd &v)
{
	GCLReal		min, max;
	Int			i;

	if (v[0] < v[1])
	{
		min = v[0];
		max = v[1];
	}
	else
	{
		min = v[1];
		max = v[0];
	}

	for (i = 2; i < v.Elts(); i++)
		if (v[i] < min)
			min = v[i];
		else if (v[i] > max)
			max = v[i];

	return(max - min);
}


ShadeInfo::ShadeInfo() : linkStack(), level(0), data(0)
{
}

