/*
    File:       HRElem.h

    Function:	Defines the element type used by the various hierarchical
				radiosity methods.

    Author:     Andrew Willmott

    Copyright:  (c) 2000, Andrew Willmott
*/

#ifndef __HRElem__
#define __HRElem__

#include "HRLink.h"
#include "cl/PtrArray.h"


// --- Hierarchical Radiosity Element -----------------------------------------


struct HRStats
{
	Void	Init();

	Int		numLinks;
	Int		numPatches;
	Int		numVolClusters;
	Int		numFaceClusters;
	Int		numAliens;
};

enum HRFlags
{
	hrPatch				= 0x0001,		///< elem is a patch (polygon)
	hrCluster			= 0x0002,		///< elem is a cluster
	hrFaceClus			= 0x0004,		///< elem is a face cluster
	hrIntNode			= 0x0010,		///< internal node in the simulation
	hrHasLights			= 0x0020,		///< some children are light sources
	hrHasNonLights		= 0x0040,		///< some children aren't
	hrMixed				= 0x0060,		///< mixed volume cluster
	hrMark				= 0x0080,		///< used for patch view
	hrAll				= 0xFFFF
};

typedef PtrArray<HRLinkList> LinkStack;

class ShadeInfo
{
public:
	ShadeInfo();
	
	LinkStack		linkStack;	///< links to current patch
	Int				level;		///< levels past solution leaves	
	Void			*data;		///< extra data
};


class HRElem
/// Class representing a hierarchical radiosity element.
{
public:
					HRElem();
	virtual			~HRElem();

	virtual Void	Reset();
					// clear all data structures associated with this elt.

	//	Link refinement
	Bool			Refine();
					/**< Refines any links whose error is too high, both
						 in this element and its children. Returns true
						 if any links were refined. */

	Bool			RefineLink(HRLink *link, Int levels);
					///< Refines a particular link.
	Void			RefineFurther(Void*);
					///< Internal routine for Refine().

	//	Solving
	Void			PushPull(Void *v = 0);
					///< Transform irradiances into radiosity at all levels
	Void			GatherAll(Void *v = 0);
					///< Gather radiosity to self and all children
	Void			Gather();
					///< Gather radiosity to self

	//	Stub routines for wavelet radiosity 
	
	virtual Bool	IsLeaf() = 0;
					///< Is this a leaf element in the solution?
	virtual Void	ApplyToChildren(Void (HRElem::*method)(Void*), 
									Void *ref = 0) = 0;
					///< Call given method on all children.
	virtual Void	MakeChildLinks(HRElem *other, HRLink *link,
								   Int which, Int levels) = 0;
					///< Create links from children to 'other'
	virtual Void	DistributeColours();
					/**< Use basis functions to colour underlying mesh/model
						 for display or saving mesh. */
 	virtual Void	DistributeColoursBest(ShadeInfo &shadeInfo);
					/**< Use basis functions to colour underlying mesh/model
						 for display or saving mesh. */
					
	//	Standard hierarchy routines.

	virtual GCLReal	Error() = 0;
	///< Returns error of a push/pull traversal
	virtual Void	Add() = 0;
	///< Add received (gathered) energy R to current irradiance B
	virtual Void	Push() = 0;
	///< Push accumulated irradiance to children.
	virtual Void	Pull() = 0;
	///< Pull up radiosity from children.
	virtual Void	ClearB() = 0;
	///< Prepare for a push/pull: clear B
	virtual Void	ClearR() = 0;
	///< Get ready for a gather: clear R.
	virtual Void	CalcLeafRadiosity() = 0;
	///< transforms irradiance in B to radiosity.
	virtual Void	InitRad();
	///< initialise radiosity.
		
	//	Visualisation stuff

	Void 			DrawMatrix(Renderer &r, Int baseNum);
	virtual Void 	DrawMatrixRec(Void *dmi);
	Void			DrawContributors(Renderer &r);
					///< Draw contributors to patch
	Void			DrawContributorsRec(Void *r);
	Void			DrawElem(Renderer &r);
					///< Draw the element (top-level routine)
	virtual Void	DrawNodeElem(Renderer &r) = 0;
					///< Draw the node at this level
	virtual Void	DrawLeafElem(Renderer &r) = 0;
					///< Draw the node by drawing corresponding leaves

	// misc.
	
	Void			CalcStats(Void *statsIn);
					///< fill in HRStats record
	Void			DumpHierarchy(Void *v = 0);

   	Bool			IsPatch()
   					{ return(flags.IsSet(hrPatch)); };
	Bool			IsCluster()
					{ return(flags.IsSet(hrCluster)); };
	Bool			IsFaceClus()
					{ return(flags.IsSet(hrFaceClus)); };

	// Fields: just the links to other elements & flags.
	
	Flags16			flags;
	HRLinkList		links;
	UInt32			id;
	static UInt32	gID;
#ifdef RAD_VIS
	virtual Void	SetHighlight(Int n)
					{ eltHighlight = n; };
	HRElem			*eltParent;	
	Byte			eltHighlight;
#endif

// --- Generic wavelet radiosity routines -------------------------------------

	// these are more flexible than the hard-coded routines in the
	// HaarLink && Flatlet/Multiwavelet links. 
	
	GCLReal			EltVis(HRElem *toElt); 
	GCLReal			EltVisToPoint(Point &p);
					///< These both use EltSetVisPoints to estimate visibility.

	virtual Void	EltSetVisPoints(HRElem *to, Point p[]) = 0;
					///< return 16 points suitable for vis. sampling
	virtual Void	EltUpdateBounds(Point &min, Point &max) = 0;
					///< update min/max to include element

	virtual GCLReal	EltProjArea(const Vector &v) = 0;
					///< Projected area in direction v. Used by HRLink
					///< and cluster stuff.
	virtual GCLReal	EltArea() = 0;
					///< Total area of the element.
	virtual Point	EltCentre() = 0;
					///< Centre of element, used mainly for vis. stuff
	virtual Colour	EltRho() = 0;
					///< Average reflectance of element

	virtual Void	EltSampleTransport(Int numSamples, Point p[], Vector n[],
						Matd &coeffs) = 0;
					/**< Find coeffs such that coeffs * src_radiosity =
						 irradiance at p[], n[]. If n = 0, calculate
						 isotropic point irradiance at each point. */
	virtual GCLReal	EltCalcTransport(HRElem *from, Matd &coeffs) = 0;
					/**< Use EltSampleTransport to find coeffs s.t. 
						 to->R_Coeffs += from->B_Coeffs * coeffs.
						 Returns an error estimate. */

	virtual Int		NumCoeffs() { return 1; };
	virtual Colour	*R_Coeffs() = 0;
					///< radiosity coefficients
	virtual Colour	*B_Coeffs() = 0;
					///< radiosity coefficients

	// cluster/vector radiosity
	virtual Colour	EltE() = 0;
	virtual Colour	EltBA();
					///< Return average power of elem. Used by cluster radiosity.
	virtual Void	AddIrradiance(const Colour &E, const Vector &m) = 0;
	virtual Void	AddChanIrradiance(const Colour &E, Int chan);
	virtual Colour	GetPower(const Vector &m) = 0;
	virtual Void	EltGetSamples(Int numSamples, Point pts[]) = 0;
	virtual GCLReal	EltMaxProjArea(const Vector &v);
					///< Maximum projected area in direction v. Used by
					///< vector radiosity stuff.

	virtual	Void	DebugInfo();

	static ostream	*out;
};

typedef HRElem				*HRElemPtr;
typedef PtrArray<HRElem>	HRElemList;

struct DMInfo
/// Used in drawing matrix
{
	Renderer	*r;
	Int			baseNum;
};

GCLReal VecError(const SubVecd &v);
///< Utility error routine. Should really go somewhere else.

#endif
