/*
	File:			HRLink.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

*/

#include "HRLink.h"
#include "HRElem.h"
#include "HierRad.h"
#include "gcl/Draw.h"

#ifdef RAD_VIS
#include "gcl/Forms.h"
#endif

#define DBG_COUT if (0) cerr
#define DBG2_COUT if (0) cerr

enum HLFlags
{
	hlGotVis	= 0x80
};

// --- HRLink methods ---------------------------------------------------------


HRLink::HRLink() : from(0), to(0), visibility(0.0), flags()
{
}

RefChoice HRLink::RefineOracle()
/* This drives the refinement strategy. */
{
	GCLReal		w, err;
    Vector		temp;
	GCLReal		fromProj, toProj;
	GCLReal 	kAError = gRadControl->kAError;
	GCLReal		kFError = gRadControl->kFError;
	
	DBG2_COUT << "REFINE ORACLE: test link from " << from->id << ':' << 
		from->flags << " to " << to->id << ':' << to->flags << endl;

#ifdef RAD_STEP_VIS
	if (visibility > 0.0 && visibility < 1.0)
		w = gRadControl->visError;
	else
		w = 1.0;
#else
	w = abs(1.0 - 2 * visibility) * (1 - gRadControl->visError) + gRadControl->visError;
#endif
		
	if (gRadControl->useBF)
		err = BFAError();
	else
		err = Error();
	
	DBG2_COUT << "error = " << err << ", w = " << w << endl;

	Expect(err >= 0.0, "negative link error???");
	if (err < 0.0) err = 0.0;
	if (err < kFError * w)
		return(kSubNone);

	if (to == from)
		return(kSubBoth);

	// Find projected area of the two patches
#ifdef RAD_PROJ_AREA_REF
    temp = safe_norm(to->EltCentre() - from->EltCentre());
    toProj = to->EltProjArea(-temp);
    fromProj = from->EltProjArea(temp);
#else
	toProj = to->EltArea();
	fromProj = from->EltArea();
#endif

	DBG2_COUT << "refining : " << fromProj << " -> " << toProj
		<< " : " << from->IsCluster() << ":" << to->IsCluster() << endl;

	//	Decide which patch to subdivide...

	if (gRadControl->favourReceiver * toProj > fromProj)
		if (to->EltArea() < kAError)
			return(kSubNone);
		else
			return(kSubTo);
	else 
		if (from->EltArea() < kAError)
			return(kSubNone);
		else
			return(kSubFrom);
}

Bool HRLink::MakeLink(
		HRElem	*fromElem,
		HRElem	*toElem,
		HRLink	*parentLink,
		Bool	forceVisReuse
	)
/** Create a link between the two elements. 
	Returns true if there is a viable link, false if there is no
	energy transfer and the link can be discarded. */
{
	Assert(fromElem != 0 && toElem != 0, 
		   "(HRLink::MakeLink) null element in to/from");
	
	from = fromElem;
	to = toElem;

	if ((from == to) && from->IsPatch())
		return(false);

	if (to->EltRho() == vl_0)
		return(false);

	if (!CalcTransport())
		return(false);

	if (!gRadControl->visInQuad)
		return(CalcVisibility(parentLink, forceVisReuse));

	return(true);	
}

Bool HRLink::CalcVisibility(HRLink *parentLink, Bool forceVisReuse)
/** Gets called if the fractional visibility needs to be calculated.
	returns true if there is at least some visibility */
{
	if (parentLink && (forceVisReuse || !NeedVisibility(parentLink)))
		visibility = parentLink->visibility;
	else
		visibility = from->EltVis(to);

	// for the moment, we can't be absolutely positive there is
	// no transfer 'cause we're sampling, so always return true.
	return(true);
}

HRLink *HRLink::CreateSubLink(HRElem *fromElem, HRElem *toElem)
{
	HierRad *method = (HierRad*) gRadControl->radObject;

	if (method->linksCreated++ > method->linksCreatedLast + 500)
	{
		if (gRadControl->stop || method->Stage(7))
			return(0);

		method->linksCreatedLast = method->linksCreated;
	}

	HRLink *result = New();
	
	if (result->MakeLink(fromElem, toElem, this))
	{
		DBG2_COUT << "made new link: " << fromElem->id << " " << toElem->id
			 << " vis = " << result->visibility << endl;
		return(result);
	}

	DBG_COUT << "link REDACTED" << endl;
	FreeLink(result);

	return(0);
}

Void HRLink::Print(ostream &s)
{
	s << "[" << from << "->" << to << "] ";
}

Void HRLink::DrawLink(Renderer &r, 
		GCLReal left, GCLReal top, 
		GCLReal right, GCLReal bottom, 
		GCLReal weight
	)
{
	r.C(cCyan);
	PaintRect(r, Coord(left, bottom), Coord(right, top));
}

ostream &operator << (ostream &s, HRLink &rl)
{
	rl.Print(s);
	
	return(s);
}

Bool HRLink::NeedVisibility(HRLink *parentLink)
{
		return(
		(to->EltArea() > gRadControl->forceVisReuseArea) ||
		to->IsCluster() || from->IsCluster() ||
	    (parentLink->visibility > 0.0 && parentLink->visibility < 1.0)
	);
}

Void HRLink::DebugInfo()
{
	cout << "=== link flags/vis: " << Int(flags) << " " << visibility << endl;
	cout << "source radiosity: " << from->B_Coeffs()[0] << endl;
	cout << "transport strength: " << Strength() << endl;
	cout << "power error: " << BFAError() << endl;
}

Void HRLink::FreeLink(HRLink *oldLink)
{
	delete oldLink;
}
