/*
	File:			HierMesh.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

	Notes:			

*/

#include "HierMesh.h"
#include "cl/String.h"
#include "gcl/VecUtil.h"

// --- Internal routines ------------------------------------------------------

static Void PrintCode(ostream &s, Int level, Int treeCode);


/*! \class HierMesh

	\verbatim
	
	Directional set up for a quad element is as follows:

				  3
	
			+-----+-----+
			|     |     |
			|  0  |  3  |
		0	|_____|_____|	2
			|     |     |
			|  1  |  2  |
			|     |     |
			+-----+-----+
				  
				  1
		
	Vertex numbers match with quadrant numbers. So the midpoint of side 0
	(west) will be vertex 1 of quadrant 0, or vertex 0 of quadrant 1.
	The opposite direction of a direction i is (i + 2) % 4.
	
	The setup for triangular quaternary subdivisions is as follows:
	
                  /\
                 /  \
                /  0 \
           0   /______\    2
              /\      /\
             /  \  3 /  \
            /  1 \  /  2 \ 
           /______\/______\

                   1
                   
	A subtlety here is that the centre triangle is a rotated
	version of the original. Moreover, all the neighbouring 
	triangles of any given triangle are of a different rotational
	persuasion, which leads to the surprising result that the opposite
	of any particular direction is itself. (Think about it!)
	Also, child 3's neighbours are all internal to the local subdivision.

	\endverbatim
*/


// --- Constructors -----------------------------------------------------------


HierElem::HierElem() : parent(0), level(0), NbRadElem()
{
	Int i;
	
	treeCode = 0;
	for (i = 0; i < 4; i++)
		child[i] = 0;
}

HierElem::~HierElem()
{
	FreeChildren();
}

static Int tQuadDirToCh1[] = { 0, 1, 2, 3 };
// Tables to find children in a certain direction for quad
static Int tQuadDirToCh2[] = { 1, 2, 3, 0 };
// e.g., the two children in dir 1 (west) are 0 and 2	 

static Int tTriDirToCh1[] = { 0, 1, 2 };
// Tables to find children in a certain direction for triangle
static Int tTriDirToCh2[] = { 1, 2, 0 };
// e.g., the two children in dir 0 (north west) are 0 and 1	
 
Void HierElem::FreeChildren()
{
	if (HasChildren())
		for (Int i = 0; i < 4; i++)
		{
			delete child[i];
			child[i] = 0;
		}
}

Void HierElem::ConnectNeighbours()
{
	Int			i, ch2, ch2o, oppDir;
	HierElem	**nc;
	
	if (IsQuad())
	{
		// Make neighbour connections between children.
		child[0]->nbFace[1] = child[1];		
		child[0]->nbFace[2] = child[3];
		child[1]->nbFace[3] = child[0];
		child[1]->nbFace[2] = child[2];
		child[2]->nbFace[3] = child[3];
		child[2]->nbFace[0] = child[1];
		child[3]->nbFace[0] = child[0];
		child[3]->nbFace[1] = child[2];

		child[0]->nbEdge[1] = 3;		
		child[0]->nbEdge[2] = 0;
		child[1]->nbEdge[3] = 1;
		child[1]->nbEdge[2] = 0;
		child[2]->nbEdge[3] = 1;
		child[2]->nbEdge[0] = 2;
		child[3]->nbEdge[0] = 2;
		child[3]->nbEdge[1] = 3;

		for (i = 0; i < 4; i++)									
		// iterate over the directions...
			if (NbFace(i) && NbFace(i)->HasChildren())	
			// do we have a neighbour with children?
			{
				nc = NbFace(i)->child;
				oppDir = nbEdge[i];
				ch2 = tQuadDirToCh2[i];
				ch2o = tQuadDirToCh2[oppDir];
				
				//	notify neighbour's children that they have new neighbours...
			
				nc[oppDir]->nbFace[oppDir] = child[ch2];
				nc[ch2o]->nbFace[oppDir] = child[i];
				nc[oppDir]->nbEdge[oppDir] = i;
				nc[ch2o]->nbEdge[oppDir] = i;
				
				//	& vice-versa
			
				child[i]->nbFace[i] = nc[ch2o];
				child[ch2]->nbFace[i] = nc[oppDir];
				child[i]->nbEdge[i] = oppDir;
				child[ch2]->nbEdge[i] = oppDir;
			}
	}
	else	// IsTri()
	{
		child[0]->nbFace[1] = child[3];
		child[1]->nbFace[2] = child[3];
		child[2]->nbFace[0] = child[3];
		child[3]->nbFace[0] = child[2];
		child[3]->nbFace[1] = child[0];
		child[3]->nbFace[2] = child[1];
	
		child[0]->nbEdge[1] = 1;
		child[1]->nbEdge[2] = 2;
		child[2]->nbEdge[0] = 0;
		child[3]->nbEdge[0] = 0;
		child[3]->nbEdge[1] = 1;
		child[3]->nbEdge[2] = 2;
	
		// Make neighbour connections between children and neighbours' children
	
		for (i = 0; i < 3; i++)									
		// iterate over the directions...
			if (NbFace(i) && NbFace(i)->HasChildren())	
			// do we have a neighbour with children?
			{
				nc = NbFace(i)->child;
				oppDir = nbEdge[i];
				i = tTriDirToCh1[i];		// our children along this edge
				ch2 = tTriDirToCh2[i];
				ch2o = tTriDirToCh2[oppDir];	// this edge
				
				//	notify neighbour's children that they have new neighbours
				nc[oppDir]->nbFace[oppDir] = child[ch2];
				nc[ch2o]->nbFace[oppDir] = child[i];
				nc[oppDir]->nbEdge[oppDir] = i;
				nc[ch2o]->nbEdge[oppDir] = i;
				
				//	& vice-versa
				child[i]->nbFace[i] = nc[ch2o];
				child[ch2]->nbFace[i] = nc[oppDir];
				child[i]->nbEdge[i] = oppDir;
				child[ch2]->nbEdge[i] = oppDir;
			}
	}
}

Void HierElem::Subdivide()
{
	Int i;
		
	// Hand off to our children if we have them...
	if (HasChildren())
	{
		for (i = 0; i < 4; i++)
			child[i]->Subdivide();
		
		return;
	}

	RadElem		mid; // to store midpoint indices
	Int 		ch1o, ch2o, n, oppDir;
	HierElem	**nc;
		
	// Allocate the children
	for (i = 0; i < 4; i++)
		child[i] = New();

	if (IsQuad())
		// store centrepoint of quad in child 0 for now.
		props->InterpolateIndexes(this, 0, 2, child[0], 2);
		
	// Fill in children's indices, creating new midpoints as necessary.
	
	if (IsTri())
	{
		for (i = 0; i < 3; i++)
		// Iterate over the directions...
			if (NbFace(i) && NbFace(i)->HasChildren())	
			// Do we have a neighbour with children?
			{
				nc = NbFace(i)->child;
				oppDir = nbEdge[i];
				ch1o = tTriDirToCh1[oppDir];
				ch2o = tTriDirToCh2[oppDir];
				
				//	If so, fetch neighbour's midpoint on this edge
				mid.SetIndexes(i, nc[ch1o], ch2o);
			}
			else
			{					
				//	Otherwise, create a new one.
				props->InterpolateIndexes(
					this,
					tTriDirToCh1[i], tTriDirToCh2[i],
					&mid, i);
			}
		
		child[0]->SetAllIndexes(this, 0, &mid, 0, &mid, 2, this, 3);
		child[1]->SetAllIndexes(&mid, 0, this, 1, &mid, 1, this, 3);
		child[2]->SetAllIndexes(&mid, 2, &mid, 1, this, 2, this, 3);
		child[3]->SetAllIndexes(&mid, 1, &mid, 2, &mid, 0, this, 3);

	}
	else	// IsQuad()
	{
		for (i = 0; i < 4; i++)									
		// Iterate over the directions...
			if (NbFace(i) && NbFace(i)->HasChildren())	
			// Do we have a neighbour with children?
			{
				nc = NbFace(i)->child;						
				oppDir = nbEdge[i];
				ch1o = tQuadDirToCh1[oppDir];
				ch2o = tQuadDirToCh2[oppDir];
				
				//	If so, fetch neighbour's midpoint on this edge
				mid.SetIndexes(i, nc[ch1o], ch2o);
			}
			else
			{	
				//	Otherwise, create a new one.
				props->InterpolateIndexes(
					this,
					tQuadDirToCh1[i], tQuadDirToCh2[i],
					&mid, i);
			}
			
		//	Fill in indices.
		child[0]->SetAllIndexes(this, 0, &mid, 0, child[0], 2, &mid, 3);
		child[1]->SetAllIndexes(&mid, 0, this, 1, &mid, 1, child[0], 2);
		child[2]->SetAllIndexes(child[0], 2, &mid, 1, this, 2, &mid, 2);
		child[3]->SetAllIndexes(&mid, 3, child[0], 2, &mid, 2, this, 3);
	}
	
	//	Now, fill in the neighbour fields
	ConnectNeighbours();
	
	//	And fill in other useful fields...
	for (i = 0; i < 4; i++)
	{
		child[i]->SetParent(SELF);
		child[i]->treeCode = (treeCode << 2) | i;
	}
		
	// Subdivide neighbours to maintain an evenly-graded mesh: each quad is 
	// subdivided to within one level of its neighbours.

	if (gRadControl->graded && parent)
	{
		// If we don't have a neighbour at our level, but our parent has one,
		// subdivide it so it's within one level of our (new) children.
	
		for (i = 0; i < Sides(); i++)
			if (!NbFace(i) && parent->NbFace(i))
				parent->NbFace(i)->Subdivide();
	}
}

Void HierElem::Reanimate(RadElem *par)
{
	RadElem::Reanimate(parent);
	
	parent = (HierElem *) par;
	if (parent)
		SetProps(parent->props);
	
	if (HasChildren())
	{
		ConnectNeighbours();
		
		child[0]->Reanimate(this);
		child[1]->Reanimate(this);
		child[2]->Reanimate(this);
		child[3]->Reanimate(this);
	}
}

Void HierElem::SetParent(HierElem &itsParent)
{
	colour = itsParent.colour;
	parent = &itsParent;
	
	level = parent->level + 1;
	area = parent->area / 4;
	SetProps(parent->props);
}

Void HierElem::ColourVertices(Int weights[])
{
	if (HasChildren())
	{
		Int i;
		
		for (i = 0; i < 4; i++)
			child[i]->ColourVertices(weights);
	}
	else
	{
		Int         i, idx, oppDir, ch2o;     
		HierElem    **nc;       
		
		for (i = 0; i < Sides(); i++)
		{
			VtxClr(i) += colour;
			weights[clrIdx[i]] += 1;
		}
		
		if (IsQuad())
		{
			// iterate over directions
			
			for (i = 0; i < 4 ; i++)
				if (NbFace(i) && NbFace(i)->HasChildren())
				// if we have neighbours
				{
					nc = NbFace(i)->child;
					oppDir = nbEdge[i];
					ch2o = tQuadDirToCh2[oppDir];
					
					// find index of neighbouring midcolour.

					idx = nc[oppDir]->clrIdx[ch2o];
					
					// add in 2 x our colour, because if we were subdivided, 
					// two of our children would contribute to this midpoint.
					
					(*props->colours)[idx] += 2 * colour;
					weights[idx] += 2;
				}
		}
		else
		{
			for (i = 0; i < 3 ; i++)
				if (NbFace(i) && NbFace(i)->HasChildren())
				{					
					nc = NbFace(i)->child;
					oppDir = nbEdge[i];
					ch2o = tTriDirToCh2[oppDir];

					// find index of neighbouring midcolour.

					idx = nc[oppDir]->clrIdx[ch2o];
	
					// add in 3 x our colour, because if we were subdivided,
				    // three of our children would contribute to this midpoint.

					(*props->colours)[idx] += 3 * colour;
					weights[idx] += 3;
				}
		}
	}
}

// --- Anchored drawing routines ----------------------------------------------

// Not for the faint of heart!

static Void AE_DrawQuad(Renderer &r, RenderStyle style, 
					 RadElem *e0, Int i0,
					 RadElem *e1, Int i1,
					 RadElem *e2, Int i2,
					 RadElem *e3, Int i3
	)
{
	if (!e0->IsTextured())
		r.Begin(style)
			.C(e0->VtxClr(i0)).P(e0->Vertex(i0))
			.C(e1->VtxClr(i1)).P(e1->Vertex(i1))
			.C(e2->VtxClr(i2)).P(e2->Vertex(i2))
			.C(e3->VtxClr(i3)).P(e3->Vertex(i3))
			.End();
	else
		r.Begin(style)
			.T(e0->TexCoord(i0)).C(e0->VtxClr(i0)).P(e0->Vertex(i0))
			.T(e1->TexCoord(i1)).C(e1->VtxClr(i1)).P(e1->Vertex(i1))
			.T(e2->TexCoord(i2)).C(e2->VtxClr(i2)).P(e2->Vertex(i2))
			.T(e3->TexCoord(i3)).C(e3->VtxClr(i3)).P(e3->Vertex(i3))
			.End();
}

static Void AE_DrawTria(Renderer &r, RenderStyle style, 
					 RadElem *e0, Int i0,
					 RadElem *e1, Int i1,
					 RadElem *e2, Int i2
	)
{
	if (!e0->IsTextured())
		r.Begin(style)
			.C(e0->VtxClr(i0)).P(e0->Vertex(i0))
			.C(e1->VtxClr(i1)).P(e1->Vertex(i1))
			.C(e2->VtxClr(i2)).P(e2->Vertex(i2))
			.End();
	else
		r.Begin(style)
			.T(e0->TexCoord(i0)).C(e0->VtxClr(i0)).P(e0->Vertex(i0))
			.T(e1->TexCoord(i1)).C(e1->VtxClr(i1)).P(e1->Vertex(i1))
			.T(e2->TexCoord(i2)).C(e2->VtxClr(i2)).P(e2->Vertex(i2))
			.End();
}

static Void AE_ProjQuad(Renderer &r, RenderStyle style, 
					 RadElem *e0, Int i0,
					 RadElem *e1, Int i1,
					 RadElem *e2, Int i2,
					 RadElem *e3, Int i3
	)
{
	r.Begin(style); 
	e0->RaiseVertex(r, i0);
	e1->RaiseVertex(r, i1);
	e2->RaiseVertex(r, i2);
	e3->RaiseVertex(r, i3);
	r.End();
}

static Void AE_ProjTria(Renderer &r, RenderStyle style, 
					 RadElem *e0, Int i0,
					 RadElem *e1, Int i1,
					 RadElem *e2, Int i2
	)
{
	r.Begin(style); 
	e0->RaiseVertex(r, i0);
	e1->RaiseVertex(r, i1);
	e2->RaiseVertex(r, i2);
	r.End();
}
 
#define MP(X) mid, X
#define VP(X) this, X
#define AE_QUAD(A, B, C, D)		AE_DrawQuad(r, style, A, B, C, D)
#define AE_TRIA(A, B, C)		AE_DrawTria(r, style, A, B, C)

#ifdef UNFINISHED

typedef Void (*QuadAction)(Void *data,
					 RadElem *e0, Int i0,
					 RadElem *e1, Int i1,
					 RadElem *e2, Int i2,
					 RadElem *e3, Int i3
	);
typedef Void (*TriAction)(Void *data,
					 RadElem *e0, Int i0,
					 RadElem *e1, Int i1,
					 RadElem *e2, Int i2,
					 RadElem *e3, Int i3
	);

Void HierElem::ProcessTri(QuadAction quadAction, TriAction triAction,
		Void *data, Int code, RadElem *mid)
{
	switch (code)	// 8 cases for a tri...
	{
	case 0:
		triAction(data, this, 0, this, 1, this, 2);
		break;
	case 1:
		triAction(data, this, 0, this, 1, mid, 2);
		triAction(data, this, 1, this, 2, mid, 2);
		break;
	case 2:
		triAction(data, this, 0, this, 1, mid, 1);
		triAction(data, mid, 1, this, 2, this, 0);
		break;
	case 3:
		quadAction(data, this, 0, this, 1, mid, 1, mid, 2);
		triAction(data, mid, 1, this, 2, mid, 2);
		break;
	case 4:
		triAction(data, this, 0, mid, 0, this, 2);
		triAction(data, mid, 0, this, 1, this, 2);
		break;
	case 5:
		triAction(data, this, 0, mid, 0, mid, 2);
		quadAction(data, mid, 0, this, 1, this, 2, mid, 2);
		break;				
	case 6:
		triAction(data, mid, 0, this, 1, mid, 1);
		quadAction(data, this, 0, mid, 0, mid, 1, this, 2);
		break;				
	case 7:
		triAction(data, this, 0, mid, 0, mid, 2);
		triAction(data, mid, 0, this, 1, mid, 1);
		triAction(data, mid, 2, mid, 1, this, 2);
		triAction(data, mid, 0, mid, 1, mid, 2);
		break;				
	}
}
#endif

Void HierElem::DrawQuad(Renderer &r, RenderStyle style, Int code, RadElem *mid)
// Draw a quad with anchoring.
{
	switch (code)	// 16 cases for a quad...
	{
	case 0:
		AE_QUAD(VP(1), VP(2), VP(3), VP(0));
		break;
	case 1:
		AE_TRIA(VP(1), MP(3), VP(0));
		AE_TRIA(VP(1), VP(2), MP(3));
		AE_TRIA(VP(2), VP(3), MP(3));
		break;
	case 2:
		AE_TRIA(VP(0), MP(2), VP(3));
		AE_TRIA(VP(0), VP(1), MP(2));
		AE_TRIA(VP(1), VP(2), MP(2));
		break;
	case 3:
		AE_TRIA(VP(1), MP(2), MP(3));
		AE_TRIA(VP(1), VP(2), MP(2));
		AE_TRIA(VP(1), MP(3), VP(0));
		AE_TRIA(MP(2), VP(3), MP(3));
		break;
	case 4:
		AE_TRIA(VP(3), MP(1), VP(2));
		AE_TRIA(VP(3), VP(0), MP(1));
		AE_TRIA(VP(0), VP(1), MP(1));
		break;
	case 5:
		AE_QUAD(VP(1), MP(1), MP(3), VP(0));
		AE_QUAD(MP(1), VP(2), VP(3), MP(3));
		break;				
	case 6:
		AE_TRIA(VP(0), MP(1), MP(2));
		AE_TRIA(VP(0), VP(1), MP(1));
		AE_TRIA(VP(0), MP(2), VP(3));
		AE_TRIA(MP(1), VP(2), MP(2));
		break;				
	case 7:
		AE_QUAD(VP(1), MP(1), MP(3), VP(0));
		AE_TRIA(MP(1), MP(2), MP(3));
		AE_TRIA(MP(1), VP(2), MP(2));
		AE_TRIA(MP(2), VP(3), MP(3));
		break;				
	case 8:
		AE_TRIA(VP(2), MP(0), VP(1));
		AE_TRIA(VP(2), VP(3), MP(0));
		AE_TRIA(VP(3), VP(0), MP(0));
		break;				
	case 9:
		AE_TRIA(VP(2), MP(3), MP(0));
		AE_TRIA(VP(2), VP(3), MP(3));
		AE_TRIA(VP(2), MP(0), VP(1));
		AE_TRIA(MP(3), VP(0), MP(0));
		break;				
	case 10:
		AE_QUAD(VP(2), MP(2), MP(0), VP(1));
		AE_QUAD(MP(2), VP(3), VP(0), MP(0));
		break;				
	case 11:
		AE_QUAD(VP(2), MP(2), MP(0), VP(1));
		AE_TRIA(MP(2), MP(3), MP(0));
		AE_TRIA(MP(2), VP(3), MP(3));
		AE_TRIA(MP(3), VP(0), MP(0));
		break;				
	case 12:
		AE_TRIA(VP(3), MP(0), MP(1));
		AE_TRIA(VP(3), VP(0), MP(0));
		AE_TRIA(VP(3), MP(1), VP(2));
		AE_TRIA(MP(0), VP(1), MP(1));
		break;				
	case 13:
		AE_QUAD(VP(3), MP(3), MP(1), VP(2));
		AE_TRIA(MP(3), MP(0), MP(1));
		AE_TRIA(MP(3), VP(0), MP(0));
		AE_TRIA(MP(0), VP(1), MP(1));
		break;				
	case 14:
		AE_QUAD(VP(0), MP(0), MP(2), VP(3));
		AE_TRIA(MP(0), MP(1), MP(2));
		AE_TRIA(MP(0), VP(1), MP(1));
		AE_TRIA(MP(1), VP(2), MP(2));
		break;				
	case 15:
		AE_TRIA(VP(1), MP(1), MP(0));
		AE_TRIA(MP(0), MP(3), VP(0));
		AE_TRIA(MP(1), VP(2), MP(2));
		AE_TRIA(MP(2), VP(3), MP(3));
		AE_QUAD(MP(0), MP(1), MP(2), MP(3));
		break;				
	}
}

Void HierElem::DrawTri(Renderer &r, RenderStyle style, Int code, RadElem *mid)
{
//	Draw a triangle with anchoring
	switch (code)	// 8 cases for a tri...
	{
	case 0:
		AE_TRIA(VP(0), VP(1), VP(2));
		break;
	case 1:
		AE_TRIA(VP(0), VP(1), MP(2));
		AE_TRIA(VP(1), VP(2), MP(2));
		break;
	case 2:
		AE_TRIA(VP(0), VP(1), MP(1));
		AE_TRIA(MP(1), VP(2), VP(0));
		break;
	case 3:
		AE_QUAD(VP(0), VP(1), MP(1), MP(2));
		AE_TRIA(MP(1), VP(2), MP(2));
		break;
	case 4:
		AE_TRIA(VP(0), MP(0), VP(2));
		AE_TRIA(MP(0), VP(1), VP(2));
		break;
	case 5:
		AE_TRIA(VP(0), MP(0), MP(2));
		AE_QUAD(MP(0), VP(1), VP(2), MP(2));
		break;				
	case 6:
		AE_TRIA(MP(0), VP(1), MP(1));
		AE_QUAD(VP(0), MP(0), MP(1), VP(2));
		break;				
	case 7:
		AE_TRIA(VP(0), MP(0), MP(2));
		AE_TRIA(MP(0), VP(1), MP(1));
		AE_TRIA(MP(2), MP(1), VP(2));
		AE_TRIA(MP(0), MP(1), MP(2));
		break;				
	}
}

#undef AE_QUAD
#undef AE_TRIA
#define AE_QUAD(A, B, C, D)		AE_ProjQuad(r, style, A, B, C, D)
#define AE_TRIA(A, B, C)		AE_ProjTria(r, style, A, B, C)

Void HierElem::ProjQuad(Renderer &r, RenderStyle style, Int code, RadElem *mid)
// Projected version of DrawQuad
{
	switch (code)	// 16 cases for a quad...
	{
	case 0:
		AE_QUAD(VP(1), VP(2), VP(3), VP(0));
		break;
	case 1:
		AE_TRIA(VP(1), MP(3), VP(0));
		AE_TRIA(VP(1), VP(2), MP(3));
		AE_TRIA(VP(2), VP(3), MP(3));
		break;
	case 2:
		AE_TRIA(VP(0), MP(2), VP(3));
		AE_TRIA(VP(0), VP(1), MP(2));
		AE_TRIA(VP(1), VP(2), MP(2));
		break;
	case 3:
		AE_TRIA(VP(1), MP(2), MP(3));
		AE_TRIA(VP(1), VP(2), MP(2));
		AE_TRIA(VP(1), MP(3), VP(0));
		AE_TRIA(MP(2), VP(3), MP(3));
		break;
	case 4:
		AE_TRIA(VP(3), MP(1), VP(2));
		AE_TRIA(VP(3), VP(0), MP(1));
		AE_TRIA(VP(0), VP(1), MP(1));
		break;
	case 5:
		AE_QUAD(VP(1), MP(1), MP(3), VP(0));
		AE_QUAD(MP(1), VP(2), VP(3), MP(3));
		break;				
	case 6:
		AE_TRIA(VP(0), MP(1), MP(2));
		AE_TRIA(VP(0), VP(1), MP(1));
		AE_TRIA(VP(0), MP(2), VP(3));
		AE_TRIA(MP(1), VP(2), MP(2));
		break;				
	case 7:
		AE_QUAD(VP(1), MP(1), MP(3), VP(0));
		AE_TRIA(MP(1), MP(2), MP(3));
		AE_TRIA(MP(1), VP(2), MP(2));
		AE_TRIA(MP(2), VP(3), MP(3));
		break;				
	case 8:
		AE_TRIA(VP(2), MP(0), VP(1));
		AE_TRIA(VP(2), VP(3), MP(0));
		AE_TRIA(VP(3), VP(0), MP(0));
		break;				
	case 9:
		AE_TRIA(VP(2), MP(3), MP(0));
		AE_TRIA(VP(2), VP(3), MP(3));
		AE_TRIA(VP(2), MP(0), VP(1));
		AE_TRIA(MP(3), VP(0), MP(0));
		break;				
	case 10:
		AE_QUAD(VP(2), MP(2), MP(0), VP(1));
		AE_QUAD(MP(2), VP(3), VP(0), MP(0));
		break;				
	case 11:
		AE_QUAD(VP(2), MP(2), MP(0), VP(1));
		AE_TRIA(MP(2), MP(3), MP(0));
		AE_TRIA(MP(2), VP(3), MP(3));
		AE_TRIA(MP(3), VP(0), MP(0));
		break;				
	case 12:
		AE_TRIA(VP(3), MP(0), MP(1));
		AE_TRIA(VP(3), VP(0), MP(0));
		AE_TRIA(VP(3), MP(1), VP(2));
		AE_TRIA(MP(0), VP(1), MP(1));
		break;				
	case 13:
		AE_QUAD(VP(3), MP(3), MP(1), VP(2));
		AE_TRIA(MP(3), MP(0), MP(1));
		AE_TRIA(MP(3), VP(0), MP(0));
		AE_TRIA(MP(0), VP(1), MP(1));
		break;				
	case 14:
		AE_QUAD(VP(0), MP(0), MP(2), VP(3));
		AE_TRIA(MP(0), MP(1), MP(2));
		AE_TRIA(MP(0), VP(1), MP(1));
		AE_TRIA(MP(1), VP(2), MP(2));
		break;				
	case 15:
		AE_TRIA(VP(1), MP(1), MP(0));
		AE_TRIA(MP(0), MP(3), VP(0));
		AE_TRIA(MP(1), VP(2), MP(2));
		AE_TRIA(MP(2), VP(3), MP(3));
		AE_QUAD(MP(0), MP(1), MP(2), MP(3));
		break;				
	}
}

Void HierElem::ProjTri(Renderer &r, RenderStyle style, Int code, RadElem *mid)
// Projected version of DrawTri
{
	switch (code)	// 8 cases for a tri...
	{
	case 0:
		AE_TRIA(VP(0), VP(1), VP(2));
		break;
	case 1:
		AE_TRIA(VP(0), VP(1), MP(2));
		AE_TRIA(VP(1), VP(2), MP(2));
		break;
	case 2:
		AE_TRIA(VP(0), VP(1), MP(1));
		AE_TRIA(MP(1), VP(2), VP(0));
		break;
	case 3:
		AE_QUAD(VP(0), VP(1), MP(1), MP(2));
		AE_TRIA(MP(1), VP(2), MP(2));
		break;
	case 4:
		AE_TRIA(VP(0), MP(0), VP(2));
		AE_TRIA(MP(0), VP(1), VP(2));
		break;
	case 5:
		AE_TRIA(VP(0), MP(0), MP(2));
		AE_QUAD(MP(0), VP(1), VP(2), MP(2));
		break;				
	case 6:
		AE_QUAD(VP(0), MP(0), MP(1), VP(2));
		AE_TRIA(MP(0), VP(1), MP(1));
		break;				
	case 7:
		AE_TRIA(VP(0), MP(0), MP(2));
		AE_TRIA(MP(0), VP(1), MP(1));
		AE_TRIA(MP(2), MP(1), VP(2));
		AE_TRIA(MP(0), MP(1), MP(2));
		break;				
	}
}

#undef AE_QUAD
#undef AE_TRIA

// --- Control method for anchored draw ---------------------------------------


Void HierElem::DrawLeaf(Renderer &r)
{
	if (gRadControl->gouraud && gRadControl->anchor)		// Draw a T-meshed tri/quad
	{
		if (gRadControl->funcView)
		{
			r.Begin(renLineLoop).C(cWhite);
			SendPoints(r);
			r.End();
		}
		
#ifdef RAD_VIS
		if (highlight)
		{
			if (highlight == 1)
				r.C(cRed);
			else if (highlight == 2)
				r.C(cYellow);
			else if (highlight == 3)
				r.C(cGreen);
			
			r.Begin(renPoly);
			SendPoints(r);
			r.End();
			
			if (!gRadControl->funcView)
				return;
		}	
#endif

		RadElem		mid;
		Int			i, code = 0;
		RenderStyle style;
				
		if (gRadControl->wire)
			style = renLineLoop;
		else
			style = renPoly;

		code = CalcMidCode(&mid);

#ifdef RAD_VIS
		mid.normal = normal;
		mid.colour = colour;
#endif		
		if (IsQuad())		// T-mesh a quad...
			if (gRadControl->funcView)
				ProjQuad(r, style, code, &mid);	
			else
				DrawQuad(r, style, code, &mid);
		else	// T-meshing a tri
			if (gRadControl->funcView)
				ProjTri(r, style, code, &mid);
			else
				DrawTri(r, style, code, &mid);
	}
	else
		RadElem::Draw(r);
}

Void HierElem::Draw(Renderer &r)
{
	Int i;

	if (HasChildren() 
#ifdef RAD_VIS
			&& !highlight
#endif
			&& (!gRadControl->choke || gRadControl->choke > level))
		for (i = 0; i < 4; i++)
			child[i]->Draw(r);
	else
		DrawLeaf(r);
}

// --- Tree code utils --------------------------------------------------------


static Void PrintCode(ostream &s, Int level, Int treeCode)
{
	Int i;
	
	for (i = level - 1; i >= 0; i--)
		s << ((treeCode >> (2 * i)) & 3);
}

Void HierElem::PrintID(ostream &s)
{
	s << props->id;
	if (level > 0)
		 s << ":";
	PrintCode(s, level, treeCode);
}

Void HierElem::PrintSelf(ostream &s)
{
	RadElem::Print(s);
}

Void HierElem::PrintRec(ostream &s)
{
	PrintSelf(s);
	
	if (HasChildren())
	{
		s << "\n{ ";
		child[0]->PrintRec(s);
		child[1]->PrintRec(s);
		child[2]->PrintRec(s);
		child[3]->PrintRec(s);
		s << "}\n";
	}
}

StrConst HierElem::Name()
{
	return("hier");
}

Void HierElem::Print(ostream &s)
{
	PrintRec(s);
}

Void HierElem::ParseSelf(istream &s)
{
	RadElem::Parse(s);
}

Void HierElem::Parse(istream &s)
{
	Char c;
	Int i;

	ParseSelf(s);

	ChompWhiteSpace(s);
	
	if (s.peek() == '{')					// children?
	{
		s.get(c);
		
		for (i = 0; i < 4; i++)
		{
			child[i] = New();
			child[i]->Parse(s);
		}	  
		
		ChompWhiteSpace(s);
		if (s.peek() != '}')	
			_Warning("(HierElem::Parse) didn't find '}'");
		else	
			s.get(c);
	}
	else
		for (i = 0; i < 4; i++)
			child[i] = 0;
}

RadElem *HierElem::FindContainer(Coord &coord)
{
	// Find the child the coord belongs to, and recursively descend.
	
	if (!HasChildren())
		return(this);
	else if (IsTri())
	{
		if (coord[0] + coord[1] < 0.5)	// bottom-left
		{
			coord[0] *= 2;
			coord[1] *= 2;
			return(child[1]->FindContainer(coord));
		}
		else if (coord[1] > 0.5)		// top 
		{
			coord[0] *= 2;
			coord[1] = 2 * coord[1] - 1;
			return(child[0]->FindContainer(coord));
		}
		else if (coord[0] > 0.5)		// bottom-right
		{
			coord[0] = 2 * coord[0] - 1;
			coord[1] *= 2;
			return(child[2]->FindContainer(coord));
		}
		else							// middle
		{
			coord[0] = 1 - 2 * coord[0];
			coord[1] = 1 - 2 * coord[1];
			return(child[3]->FindContainer(coord));
		}
	}
	else
	{
		static Int	childMap[4] = { 1, 2, 0, 3 };
		Int			idx = 0;

		coord *= 2.0;

		if (coord[0] >= 1.0)
		{
			idx++;
			coord[0] -= 1.0;
		}

		if (coord[1] >= 1.0)
		{
			idx += 2;
			coord[1] -= 1.0;
		}

		// map into our weird-ass numbering scheme and return
		return(child[childMap[idx]]->FindContainer(coord));
	}
}

GCLReal HierElem::RadError()
{
	Int		i;
	Colour	temp; 
	GCLReal	min, max;
	GCLReal	response;
			
	for (i = 0; i < Sides(); i++)
	{
		temp = VtxClr(i);
		
		// Now we map the colour temp to a perceptual response...
		
		// or we should: simple linear c. + clip for now.
		response = dot(cRGBToLum, temp);
		if (response > 1.0)
			response = 1.0;

		// done.
	   
		if (i == 0)
			min = max = response;
		else if (response > max)
			max = response;
		else if (response < min)
			min = response;
	}

	return((max - min) * area * 10.0);
}


// --- Bilevel radiosity methods ----------------------------------------------


Void HierElem::CreatePatches(PatchList &patches)
{
	Int i;

	level = 0;

	if (area < 1 / sqr(gRadControl->patchSubdivs))
		patches.Append(this);
	else
	{
		if (!HasChildren())
			Subdivide();
		for (i = 0; i < 4; i++)
			child[i]->CreatePatches(patches);
	}
}

Void HierElem::CreateElements(PatchList &elements, IndexList &eltParents,
							  Int parent)
{
	Int i, j;
			
	if (area < 1 / sqr(gRadControl->eltSubdivs))
	{
		elements.Append(this);
		eltParents.Append(parent);
	}
	else	
	{
		if (!HasChildren())
			Subdivide();
		for (i = 0; i < 4; i++)
			child[i]->CreateElements(elements, eltParents, parent);
	}
}

Void HierElem::SetColour(const Colour &c)
{
	colour = c;

	if (HasChildren())
	{
		child[0]->SetColour(c);
		child[1]->SetColour(c);
		child[2]->SetColour(c);
		child[3]->SetColour(c);
	}
}

GCLReal HierElem::MemoryUse()
{
	GCLReal	result = (GCLReal) sizeof(SELF);
	
	if (HasChildren())
	{
		result += child[0]->MemoryUse();
		result += child[1]->MemoryUse();
		result += child[2]->MemoryUse();
		result += child[3]->MemoryUse();
	}

	return(result);
}

Int HierElem::CalcMidCode(RadElem *mid)
{
	Int			i, oppDir, code, ch1o, ch2o;
	HierElem	**nc;
	
	code = 0;
	
	if (IsQuad())		
		for (i = 0; i < 4 ; i++)		// Build midpoint code
		{
			code = code << 1;
			if (NbFace(i) && NbFace(i)->HasChildren())
			{
				nc = NbFace(i)->child;
				oppDir = nbEdge[i];
				ch1o = tQuadDirToCh1[oppDir];
				ch2o = tQuadDirToCh2[oppDir];
				
				mid->SetIndexes(i, nc[ch1o], ch2o);
				code |= 1;
			}
		}
	else	
		for (i = 0; i < 3 ; i++)		// Build midpoint code
		{
			code = code << 1;
			if (NbFace(i) && NbFace(i)->HasChildren())
			{
				nc = NbFace(i)->child;
				oppDir = nbEdge[i];
				ch1o = tTriDirToCh1[oppDir];
				ch2o = tTriDirToCh2[oppDir];

				mid->SetIndexes(i, nc[ch1o], ch2o);
				code |= 1;
			}
		}		

	mid->props = props;

	return(code);
}

#define MAKE_SUB_ELEM(E0, I0, E1, I1, E2, I2, E3, I3)					\
	subElems[numElems] = New();											\
	subElems[numElems]->SetAllIndexes(E0, I0, E1, I1, E2, I2, E3, I3); 	\
	subElems[numElems]->SetProps(props);								\
	numElems++			

#define AE_QUAD(A, B, C, D)		MAKE_SUB_ELEM(A, B, C, D)
#define AE_TRIA(A, B, C)		MAKE_SUB_ELEM(A, B, C, &triTmpl, 3)

Void HierElem::SplitElement(Int &numElems, HierElemPtr subElems[5])
/*!
	Eliminate T-vertices from this triangle by triangulating it.
	This is normally done on the fly in the drawing code,
	but we also need to do it when writing out a conformant
	mesh after calculations.
*/
{

	RadElem		midStore, triTmpl;
	RadElemPtr	mid = &midStore;
	Int			code;
	
	code = CalcMidCode(mid);
	numElems = 0;  
	triTmpl.index[3] = -1;

	// Now, do the necessary triangulation...
	
	if (IsQuad())
		switch (code)	// 16 cases for a quad...
		{
		case 0:
			AE_QUAD(VP(1), VP(2), VP(3), VP(0));
			break;
		case 1:
			AE_TRIA(VP(1), MP(3), VP(0));
			AE_TRIA(VP(1), VP(2), MP(3));
			AE_TRIA(VP(2), VP(3), MP(3));
			break;
		case 2:
			AE_TRIA(VP(0), MP(2), VP(3));
			AE_TRIA(VP(0), VP(1), MP(2));
			AE_TRIA(VP(1), VP(2), MP(2));
			break;
		case 3:
			AE_TRIA(VP(1), MP(2), MP(3));
			AE_TRIA(VP(1), VP(2), MP(2));
			AE_TRIA(VP(1), MP(3), VP(0));
			AE_TRIA(MP(2), VP(3), MP(3));
			break;
		case 4:
			AE_TRIA(VP(3), MP(1), VP(2));
			AE_TRIA(VP(3), VP(0), MP(1));
			AE_TRIA(VP(0), VP(1), MP(1));
			break;
		case 5:
			AE_QUAD(VP(1), MP(1), MP(3), VP(0));
			AE_QUAD(MP(1), VP(2), VP(3), MP(3));
			break;				
		case 6:
			AE_TRIA(VP(0), MP(1), MP(2));
			AE_TRIA(VP(0), VP(1), MP(1));
			AE_TRIA(VP(0), MP(2), VP(3));
			AE_TRIA(MP(1), VP(2), MP(2));
			break;				
		case 7:
			AE_QUAD(VP(1), MP(1), MP(3), VP(0));
			AE_TRIA(MP(1), MP(2), MP(3));
			AE_TRIA(MP(1), VP(2), MP(2));
			AE_TRIA(MP(2), VP(3), MP(3));
			break;				
		case 8:
			AE_TRIA(VP(2), MP(0), VP(1));
			AE_TRIA(VP(2), VP(3), MP(0));
			AE_TRIA(VP(3), VP(0), MP(0));
			break;				
		case 9:
			AE_TRIA(VP(2), MP(3), MP(0));
			AE_TRIA(VP(2), VP(3), MP(3));
			AE_TRIA(VP(2), MP(0), VP(1));
			AE_TRIA(MP(3), VP(0), MP(0));
			break;				
		case 10:
			AE_QUAD(VP(2), MP(2), MP(0), VP(1));
			AE_QUAD(MP(2), VP(3), VP(0), MP(0));
			break;				
		case 11:
			AE_QUAD(VP(2), MP(2), MP(0), VP(1));
			AE_TRIA(MP(2), MP(3), MP(0));
			AE_TRIA(MP(2), VP(3), MP(3));
			AE_TRIA(MP(3), VP(0), MP(0));
			break;				
		case 12:
			AE_TRIA(VP(3), MP(0), MP(1));
			AE_TRIA(VP(3), VP(0), MP(0));
			AE_TRIA(VP(3), MP(1), VP(2));
			AE_TRIA(MP(0), VP(1), MP(1));
			break;				
		case 13:
			AE_QUAD(VP(3), MP(3), MP(1), VP(2));
			AE_TRIA(MP(3), MP(0), MP(1));
			AE_TRIA(MP(3), VP(0), MP(0));
			AE_TRIA(MP(0), VP(1), MP(1));
			break;				
		case 14:
			AE_QUAD(VP(0), MP(0), MP(2), VP(3));
			AE_TRIA(MP(0), MP(1), MP(2));
			AE_TRIA(MP(0), VP(1), MP(1));
			AE_TRIA(MP(1), VP(2), MP(2));
			break;				
		case 15:
			AE_TRIA(VP(1), MP(1), MP(0));
			AE_TRIA(MP(0), MP(3), VP(0));
			AE_TRIA(MP(1), VP(2), MP(2));
			AE_TRIA(MP(2), VP(3), MP(3));
			AE_QUAD(MP(0), MP(1), MP(2), MP(3));
			break;				
		}
	else
		switch (code)	// 8 cases for a tri...
		{
		case 0:
			AE_TRIA(VP(0), VP(1), VP(2));
			break;
		case 1:
			AE_TRIA(VP(0), VP(1), MP(2));
			AE_TRIA(VP(1), VP(2), MP(2));
			break;
		case 2:
			AE_TRIA(VP(0), VP(1), MP(1));
			AE_TRIA(MP(1), VP(2), VP(0));
			break;
		case 3:
			AE_QUAD(VP(0), VP(1), MP(1), MP(2));
			AE_TRIA(MP(1), VP(2), MP(2));
			break;
		case 4:
			AE_TRIA(VP(0), MP(0), VP(2));
			AE_TRIA(MP(0), VP(1), VP(2));
			break;
		case 5:
			AE_TRIA(VP(0), MP(0), MP(2));
			AE_QUAD(MP(0), VP(1), VP(2), MP(2));
			break;				
		case 6:
			AE_TRIA(MP(0), VP(1), MP(1));
			AE_QUAD(VP(0), MP(0), MP(1), VP(2));
			break;				
		case 7:
			AE_TRIA(VP(0), MP(0), MP(2));
			AE_TRIA(MP(0), VP(1), MP(1));
			AE_TRIA(MP(2), MP(1), VP(2));
			AE_TRIA(MP(0), MP(1), MP(2));
			break;				
		}
}

#undef MAKE_SUB_ELEM
#undef AE_QUAD
#undef AE_TRIA
#undef VP
#undef MP

Void HierElem::CorrectLeaves()
/*! If a mesh is balanced, the elements might not correspond to actual
	leaf nodes. We must do a pseudo-push to correct for this before 
	t-meshing. This should only be called on *solution* leaves. */
{
	if (HasChildren())
	{
		child[0]->colour = colour;
		child[1]->colour = colour;
		child[2]->colour = colour;
		child[3]->colour = colour;
		child[0]->CorrectLeaves();
		child[1]->CorrectLeaves();
		child[2]->CorrectLeaves();
		child[3]->CorrectLeaves();
	}
#ifdef OLD
	// what was I thinking? 
	Int		i;
	Int j, n = Sides();
	if (HasChildren())
	{
		// iterate over directions, interpolate midpoint for each.
		if (IsTri())
			for (i = 0; i < 3; i++)
			{
				j = (i + 1) % 3;
				
				child[tTriDirToCh1[i]]->VtxClr(tTriDirToCh2[i]) =
					0.5 * (VtxClr(i) + VtxClr(j));
			}
		else
			for (i = 0; i < 4; i++)
			{
				j = (i + 1) % 4;
				
				child[tQuadDirToCh1[i]]->VtxClr(tQuadDirToCh2[i]) =
					0.5 * (VtxClr(i) + VtxClr(j));
			}
		
		// then, correct children-of-children 
		for (i = 0; i < 4; i++)
			child[i]->CorrectLeaves();
	}
#endif
}

Void HierElem::CollectLeaves(PatchList &leaves)
{
	if (HasChildren())
		for (Int i = 0; i < 4; i++)
			child[i]->CollectLeaves(leaves);
	else
		leaves.Append(this);
}


// --- Hierarchical grid methods ----------------------------------------------


RadElemPtr HierGrid::AddChild(RadElem &quad, Int i, Int j, Int in, Int jn)
{
	HierElem	newElem;
	
	// Called by ::Mesh to add a new quad to the grid.
	
	if (i == 0 && j == 0)
		children.PreAllocate(in * jn);
	
	(RadElem &) newElem = quad;
	children.Append(newElem);

	return(&children.Last());
}

Void HierGrid::ConnectChildren()
//	Fill in the neighbour fields of all our children.
{
	Int i, j;
	Int pos;
	
	//	Join up mesh neighbours
	
	if (IsQuad())
	{
		// traverse the rectangular grid

		for (i = 0; i < rows; i++)
			for (j = 0; j < cols; j++)
			{
				pos = i * cols + j;
			
				if (i > 0) 	// connect with top neighbour
				{
					children[pos].nbFace[3] = &(children[pos - cols]);
					children[pos - cols].nbFace[1] = &(children[pos]);
					children[pos].nbEdge[3] = 1;
					children[pos - cols].nbEdge[1] = 3;
				}
				if (j > 0)	// connect with left neighbour
				{
					children[pos].nbFace[0] = &(children[pos - 1]);
					children[pos - 1].nbFace[2] = &(children[pos]);
					children[pos].nbEdge[0] = 2;
					children[pos - 1].nbEdge[2] = 0;
				}
			}
	}
	else
	{
		Int	orient, left, upper;

		pos = 0;
		
		// traverse the triangular grid
		
		/*
			2	1
		   0+--+		the diagonal is dir 2 for both tris
			|\ |		the upper/lower side is dir 1 for both tris
			| \|		the left/right side is dir 0 for both tris. 
			+--+0		lower tri is orient = 0, upper is orient = 1.
		   1   2
		*/

		for (i = 0; i < rows; i++)
			for (j = 0; j < 2 * i + 1; j++)
			{
				orient = (i + pos) & 1;		// which way is this tri oriented?
						
				if (i > 0 && orient)		// connect with top neighbour
				{
					upper = pos - 2 * i;	// index of upper neighbour
					children[pos].nbFace[1] = &(children[upper]);
					children[upper].nbFace[1] = &(children[pos]);
					children[pos].nbEdge[1] = 1;
					children[upper].nbEdge[1] = 1;
				}
				if (j > 0 || orient)		// connect with left/diag. neighbour
				{
					left = pos - 1;			// index of left neighbour
					orient = orient << 1;	// direction 1 or 2?
					children[pos].nbFace[orient] = &(children[left]);
					children[left].nbFace[orient] = &(children[pos]);
					children[pos].nbEdge[orient] = orient;
					children[left].nbEdge[orient] = orient;
				}
				pos++;
			}
	}
}

Void HierGrid::Draw(Renderer &r)
{
	if (children.NumItems() > 0 
#ifdef RAD_VIS
		&& !highlight
#endif
		)
	{
		Int i;
		
		for (i = 0; i < children.NumItems(); i++)
			children[i].Draw(r);
	}
	else
		RadElem::Draw(r);
}

Void HierGrid::CreatePatches(PatchList &patches)
{
	Int i;
	
	if (gRadControl->patchSubdivs == 0.0)
	{
		gRadControl->patchSubdivs = 1e-32;
		gRadControl->eltSubdivs = 1e-32;
	}
		
	children.Clear();

	Mesh(gRadControl->patchSubdivs);	
	ConnectChildren();
	
	for (i = 0; i < children.NumItems(); i++)
		patches.Append(&(children[i]));
}

RadElem *HierGrid::FindContainer(Coord &coord)
{
	Int i = FindChildIndex(coord);
	
	return(children[i].FindContainer(coord));
}

Void HierGrid::ColourVertices(Int weights[])
{
	Int i;
	
	for (i = 0; i < children.NumItems(); i++)
		children[i].ColourVertices(weights);
}

StrConst HierGrid::Name()
{
	return("hgrid");
}

Void HierGrid::Print(ostream &s)
{
	RadElem::Print(s);
	s << rows << ' ' << cols << ' ' << children << ' ';
}

Void HierGrid::Parse(istream &s)
{
	Int i;
	Char c;

	RadElem::Parse(s);	
	s >> rows >> cols;

	children.SetSize(rows * cols);

	ChompWhiteSpace(s);	// swallow bracket
	s.get(c);
	for (i = 0; i < children.NumItems(); i++)
		children[i].Parse(s);
	ChompWhiteSpace(s);
	s.get(c);

	ConnectChildren();

	for (i = 0; i < children.NumItems(); i++)
		children[i].props = props;
}

Void HierGrid::Reanimate(RadElem *parent)
{
	Int i;

	if (parent)
		SetProps(parent->props);

	for (i = 0; i < children.NumItems(); i++)
		children[i].Reanimate(this);
}

GCLReal HierGrid::MemoryUse()
{
	Int		i;
	GCLReal	result;

	result = sizeof(SELF);
	sGridMem += (Int) result;
	sGridChildMem += children.NumItems() * sizeof(HierElem);	

	for (i = 0; i < children.NumItems(); i++)
		result += children[i].MemoryUse();

	return(result);
}
