/*
	File:			MultiWavelets.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Notes:			
*/
			
#include "MultiWavelets.h"
#include "WaveCoeffs.h"
#ifdef RAD_VIS
#include "RadMethod.h"
#endif

/*
  Sample coordinate systems:

		   			t	 
		  		  0 ^---------+ 3
		   			|	 |	  |
		   			|	 |	  |
		   			|----+----| 
		   			|	 |	  |
		   			|	 |	  |
		   		  1 +---------+ 2 -> s
					
					
					t
					^
				  0	|\
					| \
					|  \
					|   \
				   	+----+> s
                   1      2

	(s, t) E [0, 1] x [0, 1]
	quad: (1, s, t, st)
	tri:  (1 - s - t, s, t)

    Note: the way the current basis functions are set up, B[0] contains
	the constant component for quadrilateral bases, but for triangular
	bases, all components must be averaged to find the constant component.

	NOTE: a future project is to change quads to basis whose functions correspond
	directly to the vertices.

 */

// use smoothness oracle?
#define RAD_USE_SO 
#define DECLARE_T 	GCLMat	T(To()->Coeffs(), From()->Coeffs(), transport)


// --- M2 Classes -------------------------------------------------------------

#define M2Child(x) ((M2Elem *) child[x])

// --- M2Link methods ---------------------------------------------------------


Bool M2Link::CalcTransport()
//	Calculate M2 transport coefficients.
{
	DECLARE_T;
	GCLMat	K;
	GCLReal	*visPtr = 0;
	Bool	usingCCRule;

	if (gRadControl->visInQuad) 
	{
		// calculate the visibility as we go
		visibility = 0.0;
		visPtr = &visibility;
	}
	
	factor = CalcKernel(K, visPtr, usingCCRule);
	if (factor == 0.0)	// no transport?
		return(false);

	// Calculate the transport coefficients by performing quadrature
	// on the kernel samples using the appropriate R and S matrices.

	GCLMat	D(K.Rows(), T.Cols());
	
	if (usingCCRule)
	{
		if (From()->IsQuad()) 
			MatRightTensorMult(K, MCC2_LINE_S, D);
		else
			D = K * MCC2_TRI_S;
	}
	else if (From()->IsQuad()) 
		MatRightTensorMult(K, M2_LINE_S, D);
	else
		D = K * M2_TRI_S;
		
	if (To()->IsQuad()) 
		MatLeftTensorMult(M2_LINE_R, D, T);
	else
		T = M2_TRI_R * D;

	return(true);
}

Bool M2Link::CalcVisibility(HRLink *parent, Bool forceVisReuse)
{
	if (parent && (forceVisReuse || !NeedVisibility(parent)))
		visibility = parent->visibility;
	else
		visibility = From()->Visibility(To());
	
	DECLARE_T;

	T *= visibility;

	return(true);
}


// --- M2 Oracles -------------------------------------------------------------

GCLReal M2Link::OracleError(const GCLMat &K)
// Implements the smoothness oracle of the original
// wavelet paper [Gort93].
// we test the error in K by sampling at the points EU 
{
	GCLMat			samples;
	const GCLMat	*toEU;
	Real			err = 0.0;
	GCLMat			P;
	Int				i, j;
	
#ifdef RAD_VIS
	gRadControl->radObject->display->C(cCyan);
#endif
	if (To()->IsQuad())
	{
		P = M2_QUAD_E * K;
		toEU = &M2_QUAD_EU;
	}
	else
	{
		P = M2_TRI_E * K;
		toEU = &M2_TRI_EU;
	}

	if (From()->IsQuad())
	{
		From()->SampleKernel(M2_QUAD_EU, To(), *toEU, samples, 0);
		
		for (i = 0; i < P.Rows(); i++)
			for (j = 0; j < M2_QUAD_E.Rows(); j++)
				err += abs(samples[i][j] - dot(M2_QUAD_E[j], P[i]));
	}
	else
	{
		From()->SampleKernel(M2_TRI_EU, To(), *toEU, samples, 0);
		
		for (i = 0; i < P.Rows(); i++)
			for (j = 0; j < M2_TRI_E.Rows(); j++)
				err += abs(samples[i][j] - dot(M2_TRI_E[j], P[i]));
	}
		
	return(err / 4.0);
}

GCLReal M2Link::CCOracleError(const GCLMat &K)
// Implements a smoothness oracle for the CC rule.
// We simply take the flatlet panes at each corner
// as our sample points. Because R = I for flatlets,
// each of these samples corresponds precisely to a column
// of P, saving us a dot product for each sample pair.
{
	GCLMat			samples;	
	const GCLMat	*tuv;
	Real			err = 0.0;
	GCLMat			P;
	Int				i, j;
	
#ifdef RAD_VIS
	gRadControl->radObject->display->C(cPurple);
#endif

	if (To()->IsQuad())
	{
		P = M2_QUAD_E * K;
		tuv = &M2_QUAD_EU;
	}
	else
	{
		P = M2_TRI_E * K;
		tuv = &M2_TRI_EU;
	}

	From()->EstSubFormFactor(2, To(), *tuv, samples, 0); // MKN

	for (i = 0; i < P.Rows(); i++)
		for (j = 0; j < P.Cols(); j++)
			err += abs(samples[i][j] - P[i][j]);
	
	return(err / P.Rows()); // MKN
}

GCLReal M2Link::SimpleOracleError(const GCLMat &K)
{
	Int		i;
	GCLReal error;

	// find max element of K, weight by # of source coefficients...
	error = 0.0;
	for (i = 0; i < K.Rows() * K.Cols(); i++)
		if (error < K.Ref()[i])
			error = K.Ref()[i];
	error /= From()->Coeffs();

	return(error);
}

GCLReal M2Link::CalcKernel(GCLMat &K, GCLReal *visPtr, Bool &usingCCRule)
// Sample kernel at all combinations of Quad/Tri quadrature points.
{
	GCLReal			t, error = 0.0;
	const GCLMat	*toU, *fromU;

	if (To()->IsQuad())
		toU = &M2_QUAD_U;
	else
		toU = &M2_TRI_U;

	if (From()->OrientInfo(To()) == -1)
	{
		// Too close to use normal quadrature: swap to the CC
		// rule for stability. Sample to F2 destination.

		From()->EstSubFormFactor(2, To(), *toU, K, visPtr);
		usingCCRule = true;
#ifdef RAD_USE_SO
		error = CCOracleError(K);
#else
		error = SimpleOracleError(K);
#endif
	}
	else
	{
		if (From()->IsQuad())
			fromU = &M2_QUAD_U;
		else
			fromU = &M2_TRI_U; 
	
		From()->SampleKernel(*fromU, To(), *toU, K, visPtr);
		usingCCRule = false;
#ifdef RAD_USE_SO
		error = OracleError(K);
#else
		error = SimpleOracleError(K);
#endif
	}

	return(error); 
}

Void M2Link::Gather()
{
	DECLARE_T;
	// R += T * B
	ClrVecMultAcc(T, From()->B, To()->R);
}

GCLReal M2Link::Error()
{
	return(factor);
}

GCLReal M2Link::BFAError()
{
	if (From()->IsQuad())
		return(To()->area * factor * 
			   dot(kRadRGBToLum, To()->Reflectance() * From()->B[0]));
	else
		return(To()->area * factor * 
			   dot(kRadRGBToLum, To()->Reflectance() * 
				   ClrVecAverage(From()->B, 3)));
}

Void M2Link::Print(ostream &s)
{
	DECLARE_T;
	HRLink::Print(s);
	s << " transport = " << T << " vis = " << visibility;
}

Void M2Link::DrawLink(Renderer &r, GCLReal left, GCLReal top,
					  GCLReal right, GCLReal bottom, GCLReal weight)
{
	DECLARE_T;

	DrawWaveLink(r, left, top, right, bottom, weight * To()->Reflectance(), T);
}


// --- M2Elem methods ---------------------------------------------------------


Colour M2Elem::lastB[4]; // MKN

Void M2Elem::Add()
{	
	for (Int i = 0; i < Coeffs(); i++)
		B[i] += R[i];
}
									
Void M2Elem::Push()
{
	if (IsTri())
	{
		ClrVecMult(M2_TRI_P0, B, M2Child(1)->B);
		ClrVecMult(M2_TRI_P1, B, M2Child(2)->B);
		ClrVecMult(M2_TRI_P2, B, M2Child(0)->B);
		ClrVecMult(M2_TRI_P3, B, M2Child(3)->B);
	}
	else
	{	
	/*
		Conceptually, this is just trans(T) * B * T: T interpolates rows,
		and trans(T) interpolates columns. We break up the multiplication
		because our destination matrix is not stored contigously.
		right: interpolation in s, left: interpolation in t.
	*/
		Colour t[4]; // MKN

		ClrVecRightTensorMult(M2_LINE_P0, B, t);
		ClrVecLeftTensorMult(M2_LINE_P0, t, M2Child(1)->B);
		ClrVecLeftTensorMult(M2_LINE_P1, t, M2Child(0)->B);
		ClrVecRightTensorMult(M2_LINE_P1, B, t);
		ClrVecLeftTensorMult(M2_LINE_P0, t, M2Child(2)->B);
		ClrVecLeftTensorMult(M2_LINE_P1, t, M2Child(3)->B);
	}
}

Void M2Elem::Pull()
{
	if (IsTri())
	{
		// B = sum M2_TRI_L[i] * c[i].B
		ClrVecMult   (M2_TRI_L0, M2Child(1)->B, B);
		ClrVecMultAcc(M2_TRI_L1, M2Child(2)->B, B);
		ClrVecMultAcc(M2_TRI_L2, M2Child(0)->B, B);
		ClrVecMultAcc(M2_TRI_L3, M2Child(3)->B, B);
	}
	else
	{
		Colour t[4]; // MKN
		
	    ClrVecRightTensorMult(M2_LINE_L0, M2Child(1)->B, t);
		ClrVecRightTensorMultAcc(M2_LINE_L1, M2Child(2)->B, t);
		ClrVecLeftTensorMult(M2_LINE_L0, t, B);
		
		ClrVecRightTensorMult(M2_LINE_L0, M2Child(0)->B, t);
		ClrVecRightTensorMultAcc(M2_LINE_L1, M2Child(3)->B, t);
		ClrVecLeftTensorMultAcc(M2_LINE_L1, t, B);
	}
}

Void M2Elem::ClearB()
{
	for (Int i = 0; i < Coeffs(); i++)
	{
		lastB[i] = B[i];
		B[i] = vl_0; 
	}
}

Void M2Elem::CalcLeafRadiosity()
{
	if (IsQuad())
	{
		for (Int i = 0; i < 4; i++)
			B[i] *= Reflectance(); 
		B[0] += Emittance();
	}
	else
		for (Int i = 0; i < 3; i++)
			B[i] = B[i] * Reflectance() + Emittance();
}

GCLReal M2Elem::Error()
{
	return(dot(kRadRGBToLum, B[0] - lastB[0]));
}

Void M2Elem::ClearR()
{
	for (Int i = 0; i < Coeffs(); i++)
		R[i] = vl_0;
}


// --- The numerical code for M2s ---------------------------------------------


Colour M2Elem::SampleLeaf(Coord c)
{
	if (IsQuad())
		return(M2QuadSample(c, B));
	else
		return(M2TriSample(c, B));
}

Void M2Elem::DrawLeaf(Renderer &r)
{
	if (gRadControl->wire)
		RadElem::Draw(r);
	else
	{
		Colour c0, c1, c2, c3;
		
		if (IsQuad())
		{
			c0 = B[0] + B[2];	
			c1 = B[0];
			c2 = B[0] + B[1];
			c3 = B[0] + B[1] + B[2] + B[3];
		}
		else
		{
			c0 = B[2];
			c1 = B[0];
			c2 = B[1];
		}
		
		if (highlight)
		{
			DrawHighlight(r);
			if (!gRadControl->funcView)
				return;
		}
		
		if (gRadControl->funcView)
		{
			r.Begin(renLineLoop).C(cWhite);
			SendPoints(r);
			r.End();
			
			r.Begin(renPoly);
			Project(r, c0, Vertex(0));
			Project(r, c1, Vertex(1));
			Project(r, c2, Vertex(2));
			if (IsQuad())
				Project(r, c3, Vertex(3));
			r.End();
		}
		else
		{
			r.Begin(renPoly)
				.C(c0).P(Vertex(0))
				.C(c1).P(Vertex(1))
				.C(c2).P(Vertex(2));
			if (IsQuad())
			{
				r.End().Begin(renPoly)
					.C(c2).P(Vertex(2))
					.C(c3).P(Vertex(3))
					.C(c0).P(Vertex(0));
			}
			r.End();
		}
	}
}

Void M2Elem::Print(ostream &s)
{
	s << " + m2 "; // MKN
	PrintRec(s);
}

Void M2Elem::PrintSelf(ostream &s)
{
	HRMeshElem::PrintSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s << B[i] << ' ';
}

Void M2Elem::ParseSelf(istream &s)
{
	HRMeshElem::ParseSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s >> B[i];
}

Void M2Elem::SetParent(HierElem &itsParent)
{
	HierElem::SetParent(itsParent);

//	if (itsParent.child[3] == this)
//		((HRElem &) itsParent).Push();
}

Void M2Elem::DistributeColours()
{
	colour = cRed;
	if (HasChildren())
	{
		M2Child(0)->DistributeColours();
		M2Child(1)->DistributeColours();
		M2Child(2)->DistributeColours();
		M2Child(3)->DistributeColours();
	}
}

// --- generic interface ------------------------------------------------------

					// find coeffs s.t. dot(from->B, coeffs) = irradiance
					// at p[], n[]. 
Void M2Elem::EltSampleTransport(
		Int		numSamples,
		Point	p[],
		Vector	n[],
		Matd	&coeffs
	)
// find coeffs s.t. dot(from->B, coeffs[i]) = irradiance
// at p[i], n[i].
{
#ifdef UNFINISHED
	Int		i;

	for (i = 0; i < numSamples; i++)
		coeffs[i][0] = ApproxPatchFactor(p[i], n[i]);

	Coord	*tuv, *fuv;
	Int		i, j, k, fn, tn;

	if (IsQuad())
	{ fx = M2_LINE_x.Ref(); fy = M2_LINE_y.Ref(); fn = M2_LINE_x.Elts(); }
	else
	{ fx = MT2_x.Ref(); fy = MT2_y.Ref(); fn = MT2_x.Elts(); }
	
	SampleKernel(fromU, To(), tuv, tn, K.Ref(), visPtr);
#endif
}

GCLReal M2Elem::EltCalcTransport(HRElem *from, Matd &coeffs)
{
#ifdef UNFINISHED
	if (from == this)
	{
		// no self-reflection for flat poly...
		coeffs = vl_0;
		return(0.0);
	}

	from->EltSampleTransport(1, &centre, &normal, coeffs);
	DBG_COUT << "haar: transport coeffs = " << coeffs << endl;
	return(coeffs[0][0]);
#endif
}

Void M2Elem::DistributeR(const Colour &srcR, const Vector &r)
{
	GCLReal		temp;

	temp = dot(r, normal);
	// coefficient R[0] is the constant function...
	if (temp > 0.0)
		R[0] += temp * srcR;
}


// --- M3 Classes -------------------------------------------------------------

#define M3Child(x) ((M3Elem *) child[x])

// --- M3Link methods ---------------------------------------------------------

Bool M3Link::CalcTransport()
//	Calculate M3 transport coefficients.
{
	DECLARE_T;
	GCLMat	K;
	GCLReal	*visPtr = 0;
	Bool	usingCCRule;

	if (gRadControl->visInQuad) 
	{
		// calculate the visibility as we go
		visibility = 0.0;
		visPtr = &visibility;
	}
	
	factor = CalcKernel(K, visPtr, usingCCRule);
	if (factor == 0.0)	// no transport?
		return(false);

	// Calculate the transport coefficients by performing quadrature
	// on the kernel samples using the appropriate R and S matrices.

	GCLMat	D(K.Rows(), T.Cols());
	
	if (usingCCRule)
	{
		if (From()->IsQuad()) 
			MatRightTensorMult(K, MCC3_LINE_S, D);
		else
			D = K * MCC3_TRI_S;
	}
	else if (From()->IsQuad()) 
		MatRightTensorMult(K, M3_LINE_S, D);
	else
		D = K * M3_TRI_S;
		
	if (To()->IsQuad()) 
		MatLeftTensorMult(M3_LINE_R, D, T);
	else
		T = M3_TRI_R * D;

	return(true);
}

Bool M3Link::CalcVisibility(HRLink *parent, Bool forceVisReuse)
{
	if (parent && (forceVisReuse || !NeedVisibility(parent)))
		visibility = parent->visibility;
	else
		visibility = From()->Visibility(To());
	
	DECLARE_T;

	T *= visibility;

	return(true);
}


// --- M3 Oracles -------------------------------------------------------------

GCLReal M3Link::OracleError(const GCLMat &K)
// Implements the smoothness oracle of the original
// wavelet paper [Gort93].
// we test the error in K by sampling at the points EU.
{
	GCLMat			samples;
	const GCLMat	*toEU;
	Real			err = 0.0;
	GCLMat			P;
	Int				i, j;
	
#ifdef RAD_VIS
	gRadControl->radObject->display->C(cCyan);
#endif
	if (To()->IsQuad())
	{
		P = M3_QUAD_E * K;
		toEU = &M3_QUAD_EU;
	}
	else
	{
		P = M3_TRI_E * K;
		toEU = &M3_TRI_EU;
	}

	if (From()->IsQuad())
	{
		From()->SampleKernel(M3_QUAD_EU, To(), *toEU, samples, 0);
		
		for (i = 0; i < P.Rows(); i++)
			for (j = 0; j < M3_QUAD_E.Rows(); j++)
				err += abs(samples[i][j] - dot(M3_QUAD_E[j], P[i]));
	}
	else
	{
		From()->SampleKernel(M3_TRI_EU, To(), *toEU, samples, 0);
		
		for (i = 0; i < P.Rows(); i++)
			for (j = 0; j < M3_TRI_E.Rows(); j++)
				err += abs(samples[i][j] - dot(M3_TRI_E[j], P[i]));
	}
		
	return(err / P.Rows());
}

GCLReal M3Link::CCOracleError(const GCLMat &K)
// Implements a smoothness oracle for the CC rule.
// We simply take the flatlet panes at each corner
// as our sample points. Because R = I for flatlets,
// each of these samples corresponds precisely to a column
// of P, saving us a dot product for each sample pair.
{
	GCLMat			samples;
	const GCLMat	*tuv;
	Real			err = 0.0;
	GCLMat			P;
	Int				i, j;
	
#ifdef RAD_VIS
	gRadControl->radObject->display->C(cPurple);
#endif

	if (To()->IsQuad())
	{
		P = M3_QUAD_E * K;
		tuv = &M3_QUAD_EU;
	}
	else
	{
		P = M3_TRI_E * K;
		tuv = &M3_TRI_EU;
	}

	From()->EstSubFormFactor(3, To(), *tuv, samples, 0); // MKN

	for (i = 0; i < P.Rows(); i++)
		for (j = 0; j < P.Cols(); j++)
			err += abs(samples[i][j] - P[i][j]);
	
	return(err / P.Rows());
}

GCLReal M3Link::SimpleOracleError(const GCLMat &K)
{
	Int		i;
	GCLReal error;

	// find max element of K, weight by # of source coefficients...
	error = 0.0;
	for (i = 0; i < K.Rows() * K.Cols(); i++)
		if (error < K.Ref()[i])
			error = K.Ref()[i];
	error /= From()->Coeffs();

	return(error);
}

GCLReal M3Link::CalcKernel(GCLMat &K, GCLReal *visPtr, Bool &usingCCRule)
// Sample kernel at all combinations of Quad/Tri quadrature points.
{
	GCLReal			t, error = 0.0;
	const GCLMat	*tuv, *fuv;

	if (To()->IsQuad())
		tuv = &M3_QUAD_U;
	else
		tuv = &M3_TRI_U; 

	if (From()->OrientInfo(To()) == -1)
	{
		// Too close to use normal quadrature: swap to the CC
		// rule for stability.

		From()->EstSubFormFactor(3, To(), *tuv, K, visPtr);
		usingCCRule = true;
#ifdef RAD_USE_SO
		error = CCOracleError(K);
#else
		error = SimpleOracleError(K);
#endif
	}
	else
	{
		if (From()->IsQuad())
			fuv = &M3_QUAD_U;
		else
			fuv = &M3_TRI_U;
	
		From()->SampleKernel(*fuv, To(), *tuv, K, visPtr);
		usingCCRule = false;
#ifdef RAD_USE_SO
		error = OracleError(K);
#else
		error = SimpleOracleError(K);
#endif
	}

	return(error); 
}

Void M3Link::Gather()
{
	DECLARE_T;
	// R += T * B
	ClrVecMultAcc(T, From()->B, To()->R);
}

GCLReal M3Link::Error()
{
	return(factor);
}

GCLReal M3Link::BFAError()
{
	if (From()->IsQuad())
		return(To()->area * factor * 
			   dot(kRadRGBToLum, To()->Reflectance() * From()->B[0]));
	else
		return(To()->area * factor * 
			   dot(kRadRGBToLum, To()->Reflectance() * 
				   ClrVecAverage(From()->B, 6)));
}

Void M3Link::Print(ostream &s)
{
	DECLARE_T;
	HRLink::Print(s);
	s << " transport = " << T << " vis = " << visibility;
}

Void M3Link::DrawLink(Renderer &r, GCLReal left, GCLReal top,
					  GCLReal right, GCLReal bottom, GCLReal weight)
{
	DECLARE_T;

	DrawWaveLink(r, left, top, right, bottom, weight * To()->Reflectance(), T);
}


// --- M3Elem methods ---------------------------------------------------------


Colour M3Elem::lastB[9]; // MKN

Void M3Elem::Add()
{	
	for (Int i = 0; i < Coeffs(); i++)
	{
		// 3rd-order bases can actually go negative, esp. round corners
		// if we don't nip this in the bud by clamping here, solution
		// will spiral out of control.
		ClipColourZero(R[i]);
		B[i] += R[i];
	}
}
									
Void M3Elem::Push()
{
	if (IsTri())
	{
		ClrVecMult(M3_TRI_P0, B, M3Child(1)->B);
		ClrVecMult(M3_TRI_P1, B, M3Child(2)->B);
		ClrVecMult(M3_TRI_P2, B, M3Child(0)->B);
		ClrVecMult(M3_TRI_P3, B, M3Child(3)->B);
	}
	else
	{	
	/*
		Conceptually, this is just trans(T) * B * T: T interpolates rows,
		and trans(T) interpolates columns. We break up the multiplication
		because our destination matrix is not stored contigously.
		right: interpolation in s, left: interpolation in t.
	*/
		Colour t[9]; // MKN

		ClrVecRightTensorMult(M3_LINE_P0, B, t);
		ClrVecLeftTensorMult(M3_LINE_P0, t, M3Child(1)->B);
		ClrVecLeftTensorMult(M3_LINE_P1, t, M3Child(0)->B);
		ClrVecRightTensorMult(M3_LINE_P1, B, t);
		ClrVecLeftTensorMult(M3_LINE_P0, t, M3Child(2)->B);
		ClrVecLeftTensorMult(M3_LINE_P1, t, M3Child(3)->B);
	}
}

Void M3Elem::Pull()
{
	if (IsTri())
	{
		// B = sum M3_TRI_L[i] * c[i].B
		ClrVecMult   (M3_TRI_L0, M3Child(1)->B, B);
		ClrVecMultAcc(M3_TRI_L1, M3Child(2)->B, B);
		ClrVecMultAcc(M3_TRI_L2, M3Child(0)->B, B);
		ClrVecMultAcc(M3_TRI_L3, M3Child(3)->B, B);
	}
	else
	{
		Colour t[9]; // MKN
		
	    ClrVecRightTensorMult(M3_LINE_L0, M3Child(1)->B, t);
		ClrVecRightTensorMultAcc(M3_LINE_L1, M3Child(2)->B, t);
		ClrVecLeftTensorMult(M3_LINE_L0, t, B);
		
		ClrVecRightTensorMult(M3_LINE_L0, M3Child(0)->B, t);
		ClrVecRightTensorMultAcc(M3_LINE_L1, M3Child(3)->B, t);
		ClrVecLeftTensorMultAcc(M3_LINE_L1, t, B);
	}
}

Void M3Elem::ClearB()
{
	for (Int i = 0; i < Coeffs(); i++)
	{
		lastB[i] = B[i];
		B[i] = vl_0; 
	}
}

Void M3Elem::CalcLeafRadiosity()
{
	if (IsQuad())
	{
		for (Int i = 0; i < 9; i++)
			B[i] *= Reflectance(); 
		B[0] += Emittance();
	}
	else
		for (Int i = 0; i < 6; i++)
			B[i] = B[i] * Reflectance() + Emittance();
}

GCLReal M3Elem::Error()
{
	return(dot(kRadRGBToLum, B[0] - lastB[0]));
}

Void M3Elem::ClearR()
{
	for (Int i = 0; i < Coeffs(); i++)
		R[i] = vl_0;
}


// --- The numerical code for M3s ---------------------------------------------


Colour M3Elem::SampleLeaf(Coord c)
{
	if (IsQuad())
		return(M3QuadSample(c, B));
	else
		return(M3TriSample(c, B));
}

Void M3Elem::DrawLeaf(Renderer &r)
{
	if (gRadControl->wire)
		RadElem::Draw(r);
	else
	{
		Bool project = gRadControl->funcView;
		
		if (project)
		{
			r.Begin(renLineLoop).C(cWhite);
			SendPoints(r);
			r.End();
		}
		if (highlight)
		{
			DrawHighlight(r);
			if (!project)
				return;
		}
	
		DrawSampledLeaf(r, 4);
	}
}

Void M3Elem::Print(ostream &s)
{
	s << " + m3 "; // MKN
	PrintRec(s);
}

Void M3Elem::PrintSelf(ostream &s)
{
	HRMeshElem::PrintSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s << B[i] << ' ';
}

Void M3Elem::ParseSelf(istream &s)
{
	HRMeshElem::ParseSelf(s);
	for (Int i = 0; i < Coeffs(); i++)
		s >> B[i];
}

Void M3Elem::SetParent(HierElem &itsParent)
{
	HierElem::SetParent(itsParent);

//	if (itsParent.child[3] == this)
//		((HRElem &) itsParent).Push();
}

Void M3Elem::DistributeColours()
{
	colour = cRed;
	if (HasChildren())
	{
		M3Child(0)->DistributeColours();
		M3Child(1)->DistributeColours();
		M3Child(2)->DistributeColours();
		M3Child(3)->DistributeColours();
	}
}

// --- generic interface ------------------------------------------------------

					// find coeffs s.t. dot(from->B, coeffs) = irradiance
					// at p[], n[]. If n = 0, calculate isotropic point
					// irradiance.
Void M3Elem::EltSampleTransport(
		Int		numSamples,
		Point	p[],
		Vector	n[],
		Matd	&coeffs
	)
// find coeffs s.t. dot(from->B, coeffs[i]) = irradiance
// at p[i], n[i].
{
#ifdef UNFINISHED
	Int		i;

	for (i = 0; i < numSamples; i++)
		coeffs[i][0] = ApproxPatchFactor(p[i], n[i]);

	Coord	*tuv, *fuv;
	Int		i, j, k, fn, tn;

	if (IsQuad())
	{ fx = M3_LINE_x.Ref(); fy = M3_LINE_y.Ref(); fn = M3_LINE_x.Elts(); }
	else
	{ fx = MT2_x.Ref(); fy = MT2_y.Ref(); fn = MT2_x.Elts(); }
	
	K.SetSize(tn, fn);
	SampleKernel(fuv, fn, To(), tuv, tn, K.Ref(), visPtr);
#endif
}

GCLReal M3Elem::EltCalcTransport(HRElem *from, Matd &coeffs)
{
#ifdef UNFINISHED
	if (from == this)
	{
		// no self-reflection for flat poly...
		coeffs = vl_0;
		return(0.0);
	}

	from->EltSampleTransport(1, &centre, &normal, coeffs);
	DBG_COUT << "haar: transport coeffs = " << coeffs << endl;
	return(coeffs[0][0]);
#endif
}

Void M3Elem::DistributeR(const Colour &srcR, const Vector &r)
{
	GCLReal		temp;

	temp = dot(r, normal);
	// coefficient R[0] is the constant function...
	if (temp > 0.0)
		R[0] += temp * srcR;
}
