/*
	File:			RadMethod.cc

	Function:		See header file

	Author(s):		Andrew Willmott

	Copyright:		(c) 1997-2000, Andrew Willmott

	Notes:			

*/

#include "RadMethod.h"

#include "fstream.h"
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "cl/String.h"
#include "gcl/VertexJoin.h"
#include "gcl/MRModel.h"
#include "gcl/SceneLang.h"

#include "MeshJoin.h"

#ifdef RAD_VIS
#include "gcl/XGraphicsSystem.h"
#include "RadScene.h"
#endif

#ifdef DEBUG
#define DBG_COUT cerr
//#define RAD_HIST
#else
#define DBG_COUT if (0) cerr
#endif

/// time-limited updates?
#define RAD_TLD
/// leave time TLD_DELTA between updates, in wallclock time, in seconds
#define RAD_TLD_DELTA 1.0

// --- Base radiosity method class --------------------------------------------

RadAttributes::RadAttributes() :
	points(0),
	colours(0),
	normals(0),
	texCoords(0)
{
}

RadMethod::RadMethod() : 
#ifdef RAD_VIS
	doUpdate(false),
	display(0), 
	gsP(0),
#endif
	RadAttributes(),
	grid(0),
	object(0),
	props(0),
	scene(0)
{
	gRadControl->radObject = this;
}

RadMethod::~RadMethod()
{
	delete grid;
	delete scene;
}

scScenePtr RadMethod::GetScene()
{
#ifdef RAD_VIS
	if (!scene)
	{
		scRadMesh 	*rm = new scRadMesh;
		rm->SetMethod(this);	
		scene = slBeginObject("rad mesh scene");
		slObject(rm);
		slEndObject();
	}
#endif
	return(scene);
}


class RadElemBuilder : public Decimator
{
public:
				RadElemBuilder(RadMethod *rm) : method(rm) {};

	Void		HandlePoly(Int numVertices, Int vertices[], Int changed)
				{ method->AddQuadTri(numVertices, vertices, changed, this); };

	RadMethod	*method;
};

Void RadMethod::SetScene(scScenePtr theScene)
{
	// We convert the scene into a quad-tri mesh

	Int		i;
	UInt32	decFlags;
	
	DBG_COUT << "setting scene" << endl;
	
	delete props;
	
	baseElems.FreeAll();	
	baseElems.Clear();	
	
#ifdef RAD_VIS
	Field(out1) << "Creating radiosity elements" << endl;
#else
	GCLReal saveComplexity = scMRModel::GetComplexity();
	scMRModel::SetComplexity(gRadControl->meshComplexity);
#endif

	sceneName = theScene->Label();
	decFlags = DecTris | DecQuads;
#ifdef RAD_MR
	if (gRadControl->fcr && gRadControl->method == kHierarchical)
		decFlags |= DecIgnoreMRM;
#endif
	theScene->Decimate(qtInfo, decFlags);

	DBG_COUT << "scene has " << qtInfo.numProps << " materials, "
		<< qtInfo.numPolys << " base elems." << endl;
	DBG_COUT << "qtinfo [tqp props]: " << qtInfo.numTris << ' ' << qtInfo.numQuads <<
		' ' << qtInfo.numPolys << ' ' << qtInfo.numProps << endl;
	
	if (!colours)
		colours = new ColourList;
//	if (!normals)
//		normals = new VectorList;
	props = new RadProps[qtInfo.numProps];
	numProps = 0;
	gRadControl->numPolys = qtInfo.numPolys;

	// build element list
	RadElemBuilder	elemBuilder(this);
	theScene->Decimate(elemBuilder, decFlags | DecUseMaster);
	// set up the mesh!
	SetupMesh();

	// set up raytracing
#ifndef RAD_VIS
	scMRModel::SetComplexity(gRadControl->rtComplexity);
#endif
	CreateRTGrid(theScene);
#ifndef RAD_VIS
	scMRModel::SetComplexity(saveComplexity);
#endif
	
	// Setup elements and method-specific stuff
	ResetOptions();
}

Void RadMethod::ResetOptions()
{
	gRadControl->finalPass = false;
	CreatePatches();
	ColourMeshInitial();
	ColourVertices();

#ifdef RAD_VIS
	gRadControl->pvData = 0;
	Field(out1) << patches.NumItems() << " patches." << show;
#endif
}

Void RadMethod::AddQuadTri(
		Int			numVertices,
		Int			vertices[],
		Int			changed,
		Decimator	*state
	)
{
	if (numVertices == DEC_End)
	{
		// we're cooked: grab the master point list
		points = state->pointsAcc;
		return;
	}
	else if (numVertices < 0) return;

	Int			i;
	RadElem		*elem = NewMesh();
	RadProps	*elemProps;
	
	elem->index[3] = -1;
	Assert(numVertices <= 4, "too many vertices!");
	
	for (i = 0; i < numVertices; i++)
	{
		elem->index[i] = vertices[i];
		elem->clrIdx[i] = vertices[i];
	}
			
	if (changed)
	{ 
		SetupProps(props + numProps, state);
		numProps++;
	}
	
	elemProps = props + (numProps - 1);

	if (elemProps->texture && elemProps->texCoords)
	{
		IndexList	*coordIndexes = 0;

		// XXX
		//DEC_GET(CoordIndexes);

		if (coordIndexes)
			for (i = 0; i < numVertices; i++)
				elem->texIdx[i] = coordIndexes->Item(i);	
		else
			for (i = 0; i < numVertices; i++)
				elem->texIdx[i] = i;	
	}
			
	elem->SetProps(elemProps);
	baseElems.Append(elem);
}

// --- Setup routines ---------------------------------------------------------

Void RadMethod::SetupMesh()
{
	Int		i;
	
	// Given 'baseElems', sets up other necessary stuff.	

	// Find bounds of our scene...
	FindBounds(min, max);

	// fix the mesh if necessary. (Make sure that vertices are shared.)	
	if (gRadControl->fixMesh)
		FixMesh();

	// Connect up the mesh: find adjacency info & use it to determine
	// placement of shared colours & normals.
	if (gRadControl->connectMesh)
		ConnectMesh();
	else
	{
		colours->SetSize(points->NumItems());		
		if (normals)
			normals->SetSize(points->NumItems());
	}
	
	// Now, reset all the prop references...	
	for (i = 0; i < numProps; i++)
	{
		props[i].points = points;
		props[i].colours = colours;
		props[i].normals = normals;
	}

	// should we be collapsing normIdx/clrIdx?
	for (i = 0; i < baseElems.NumItems(); i++)
	{
		baseElems[i]->normIdx[0] = baseElems[i]->clrIdx[0];
		baseElems[i]->normIdx[1] = baseElems[i]->clrIdx[1];
		baseElems[i]->normIdx[2] = baseElems[i]->clrIdx[2];
		baseElems[i]->normIdx[3] = baseElems[i]->clrIdx[3];
	}
	AssignNormals();

	DBG_COUT << "mesh has " << points->NumItems() << " vertices, " 
		 << colours->NumItems() << " colours";
	if (normals)
		DBG_COUT << ", " << normals->NumItems() << " vertex normals" << endl;
	DBG_COUT << endl;
	
}

Void RadMethod::FixMesh()
{
	VertexJoin	joiner;
	Int			i, m;

	DBG_COUT << "sharing vertices..." << endl;

	joiner.Init(min, max, points->NumItems());
	
	for (i = 0; i < baseElems.NumItems(); i++)
		for (m = 0; m < baseElems[i]->Sides(); m++)
	{
		baseElems[i]->index[m] = joiner.AddVertex(baseElems[i]->Vertex(m));
		baseElems[i]->clrIdx[m] = baseElems[i]->index[m];
	}
	
	DBG_COUT << "old points: " << points->NumItems() << endl;
	delete points;
	points = joiner.GetFinalVertexList();
	DBG_COUT << "new points: " << points->NumItems() << endl;

	DBG_COUT << "done." << endl;
	
#ifdef CL_CHECKING
	// mesh consistency check...
	
	Int			j, k, a, b;
	NbRadElem	*be, *nb;
	
	DBG_COUT << "checking mesh..." << endl;
	for (i = 0; i < baseElems.NumItems(); i++)
	{
		be = (NbRadElem*) baseElems[i];
		for (j = 0; j < 3; j++)
			if (be->nbFace[j])
			{
				nb = be->nbFace[j];
				k = (j + 1) % 3;
				a = be->nbEdge[j];
				b = (a + 1) % 3;
				
				Assert(baseElems[i]->index[j] == nb->index[b], "1st vertices");
				Assert(baseElems[i]->index[k] == nb->index[a], "2nd vertices");
			}
	}
	DBG_COUT << "done." << endl;
#endif	
}

Void RadMethod::ConnectMesh()
{
	MeshJoin	meshJoiner;
	Int			i;
	
	DBG_COUT << "connecting mesh..." << endl;

	meshJoiner.Init(points->NumItems());
	meshJoiner.StartGroup();
	for (i = 0; i < baseElems.NumItems(); i++)
		meshJoiner.AddTriangle((NbRadElem*) baseElems[i]);
	meshJoiner.EndGroup();
	i = meshJoiner.Finished();
	colours->SetSize(i);
	if (normals)
		normals->SetSize(i);

	DBG_COUT << "done." << endl;
}

Void RadMethod::FindBounds(Point &min, Point &max)
{
	Int i;
	
	min.MakeBlock(HUGE_VAL);
	max.MakeBlock(-HUGE_VAL);
	for (i = 0; i < points->NumItems(); i++)
		UpdateBounds((*points)[i], min, max);
}

Void RadMethod::ColourMeshInitial()
{
	// just sets colours to emittance + reflectance...

	Int i;
	
	for (i = 0; i < patches.NumItems(); i++)
		patches[i]->SetColour(patches[i]->Reflectance()
			+ patches[i]->Emittance());
}

Void RadMethod::CreatePatches()
{
	Int i;
	
	patches.Clear();
	stats.Init();	

	DBG_COUT << "creating patches" << endl;

	// Subdivide base poly to initial mesh, accumulate into polys and
	// patches lists.	
				
	for (i = 0; i < baseElems.NumItems(); i++)
		baseElems[i]->CreatePatches(patches);

	for (i = 0; i < patches.NumItems(); i++)
		stats.Update(patches[i]);

#ifdef DEBUG
	stats.Report(cerr);
#endif
}

// --- Ray tracing support ----------------------------------------------------


class RTElemBuilder : public Decimator
{
public:
				RTElemBuilder(RadMethod *rm) : method(rm) {};

	Void		HandlePoly(Int numVertices, Int vertices[], Int changed);

	RadMethod	*method;
};

Void RTElemBuilder::HandlePoly(
		Int			numVertices,
		Int			vertices[],
		Int			changed
	)
{
	RT_Tri		*tri;
	
	if (numVertices == DEC_End)
	{
		// we're cooked: attach the master point list to the RT object
		method->object->numPoints = pointsAcc->NumItems();
		method->object->points = pointsAcc->Detach();
		return;
	}
	else if (numVertices < 0) return;
			
	
   	tri = method->object->tris + method->rtID;
	tri->Init(vertices[0], vertices[1],
			 vertices[2], method->object, method->rtID);
#ifdef RAD_VIS
	CheckRange(decNum, 0, method->baseElems.NumItems(),
	   "mismatch between radiosity and ray-trace triangles");
	if (method->rtElemPtrs.NumItems() > 0)
		method->rtElemPtrs[method->rtID] = method->baseElems[decNum];
#endif

	tri->flags |= TRI_2SIDED_V;		// visibility checking is two sided
	method->rtID++;

	if (numVertices > 3)
  	// quad! add second triangle
	{
		tri = method->object->tris + method->rtID;
		tri->Init(vertices[2], vertices[3],
				  vertices[0], method->object, method->rtID);
#ifdef RAD_VIS
		if (method->rtElemPtrs.NumItems() > 0)
			method->rtElemPtrs[method->rtID] = method->baseElems[decNum];
#endif
		method->rtID++;
	}
}

Void RadMethod::CreateRTObjects(scScenePtr scene)
{
	UInt32	decFlags = DecTris | DecQuads;

	if (object)
	{
		object->Free(); 
		delete object;
		object = 0;
	}

#ifdef RAD_MR
	if (gRadControl->fcr && gRadControl->method == kHierarchical)
		decFlags |= DecIgnoreMRM;
#endif

	// add all the triangles...
	DBG_COUT << "adding base triangles..." << endl;

	if (gRadControl->rtComplexity != gRadControl->meshComplexity)
	{
		// not guaranteed correspondence between rt & radiosity triangles,
		// so recalc.
		scene->Decimate(qtInfo, decFlags);
		// can't have mapping between rt elems and rad elems
#ifdef RAD_VIS
		rtElemPtrs.Clear();
#endif
	}
#ifdef RAD_VIS
	else
		rtElemPtrs.SetSize(qtInfo.numPolys + qtInfo.numQuads);
#endif

	rtID = 0;
	object = new RT_Object;
	object->Init(qtInfo.numPolys + qtInfo.numQuads);
	
	RTElemBuilder	elemBuilder(this);
	scene->Decimate(elemBuilder, decFlags | DecUseMaster);
	object->Setup();
	DBG_COUT << rtID << " triangles added." << endl;

	grid->AddObject(object);
}

Void RadMethod::CreateRTGrid(scScenePtr scene)
{	
	if (grid)
	{
		grid->Free();
		delete grid;
		grid = 0;
	}

	if (!gRadControl->gridOn)
		return;

#ifdef RAD_VIS
	RM_OUT1("Creating ray-trace grid");
#endif
	DBG_COUT << "Creating grid..." << endl;

	// get the grid ready.
	// set up the grid
	grid = new RT_Grid();
	grid->MasterInit();

	CreateRTObjects(scene);

	grid->sEpsilon = 1e-6;
	grid->PrepareGrid();
#ifdef DEBUG
	grid->DumpMemoryUsage();
#endif

	DBG_COUT << "done." << endl;
#ifdef RAD_VIS
	RM_OUT1("done.");
#endif
}

Bool RadMethod::IntersectsWithRay(const Point &start, const Point &end)
{
	if (grid)
		return(grid->IntersectionExists(start, end));
	else
		return(false);
}

RadElem *RadMethod::ClosestIntersection(const Point &start, const Point &dir, Point &p)
{
	if (grid)
	{
		if (grid->ClosestIntersection(start, dir))
		{
			p = grid->hitT * dir + start;

#ifdef RAD_VIS
			if (rtElemPtrs.NumItems() > 0)
				return(rtElemPtrs[grid->hitPrim->id]);
			else
#endif
				return(0);
		}
		else
			return(0);
	}
	else
		return(0);
}

Bool RadMethod::Render()
{
	gRadControl->rays = 0;
	totTime = 0;
	lastTime = 0;
	dumpID = 0;

	if (gRadControl->kFErrorRel >= 0.0)
	{
		cerr << "setting absolute threshold to " << gRadControl->kFErrorRel;
		cerr << " * " << dot(kRadRGBToLum, stats.totalPower) << endl;

		gRadControl->kFError = gRadControl->kFErrorRel * 
			dot(kRadRGBToLum, stats.totalPower);
	}
	
	timer.StartTimer();	// Initialise, then suspend timer.
	timer.StopTimer();
	
	return(true);
}

Void RadMethod::RemoveDirect()
{
	// default is not to do it.
}

// --- Measurement ------------------------------------------------------------

Int RadMethod::Stage(Int stage)
{
	return(0);
}

Bool RadMethod::Idle()
{
#ifdef RAD_VIS
	if (gsP)
	{
		gsP->Spin();
		
		return(gRadControl->stop);
	}
	else
#endif
		return(false);
}

Bool RadMethod::Pause()
/*! Called when we reach a potential pause point. */
{
#ifdef RAD_VIS
	if (gsP)
	{
		do	
			gsP->Spin();
		while (gRadControl->pause && !gRadControl->step);
		
		gRadControl->step = 0;
		
		return(gRadControl->stop);
	}
	else
#endif
		return(false);
}

Void RadMethod::DumpScene()
{
	ofstream	outFile;
	String		filename;

	if (gRadControl->dumpScenes)
	{
		ColourVertices();

		if (dumpID != 0)	// don't bother with first one.
		{
			filename.Printf("%s.%02d.sl", gRadControl->outFile, dumpID);

			outFile.open(filename);
			if (outFile)
				GetScene()->HierPrint(outFile, 0);
			outFile.close();
		}
	}
	dumpID++;
}

Bool RadMethod::CheckTime()
{
	timer.StopTimer();
	totTime = timer.GetTimer();

	// if we're out of time, dump stats, and terminate.
	if (totTime > gRadControl->limitTime)
	{
		gRadControl->finalPass = true;
		DumpStats();
		return(true);
	}

	// if another time-slice has elapsed, dump stats
	if (totTime - lastTime >= gRadControl->sliceTime)
	{
		while (totTime - lastTime >= gRadControl->sliceTime)
			lastTime += gRadControl->sliceTime;

		DumpStats();
	}

	return(false);
}

#ifdef RAD_VIS

Bool RadMethod::Update()
{
#ifndef RAD_TLD
	// always update
	return(true); 
#else
	
	if (utimer.GetTimer() > nextUpdate)
	{
		nextUpdate = utimer.GetTimer() + RAD_TLD_DELTA;
		return(true);
	}
	return(false);
#endif
}

Void RadMethod::StartUpdate()
{
	utimer.StartTimer();
	nextUpdate = 0.0;
}

Void RadMethod::UpdateCont()
{
	nextUpdate = utimer.GetTimer() + RAD_TLD_DELTA;
}

#endif


// --- Visualisation stuff ----------------------------------------------------

Void RadMethod::Draw(Renderer &r)
{
	Int			i;

	for (i = 0; i < baseElems.NumItems(); i++)
	{
		if (gRadControl->texture && baseElems[i]->props->texture)
			r.SetTexture(baseElems[i]->props->texture->image);
		baseElems[i]->Draw(r);
	}

	if (gRadControl->texture)
		r.SetTexture(0);

	if (gRadControl->outlineVisGrid)
		grid->Draw(r, 0);
}

Void RadMethod::DrawMatrix(Renderer &r)
{
	Expect(false, "Radiosity method has no draw matrix method.");
}

#ifdef RAD_VIS

Void RadMethod::RenderMatrix()
{
	if (matDisplay)
		matDisplay->Redraw();
}

#endif

Void RadMethod::WriteSLFile(StrConst filename)
// Write final mesh as a .sl file
{
	Int				i;
	ofstream		s;
	PatchList		*leaves = GetElements();

	s.open(filename);
	if (!s)
		return;
	
	s << "object \"rad_output\"" << endl;
	s << "camera" << endl;
	s << "points " << *points << endl;
	s << "colours " << *colours << endl;
	if (normals)
		s << "normals " << *normals << endl;
	if (texCoords)
		s << "coords " << *texCoords << endl;
		
	for (i = 0; i < leaves->NumItems(); i++)
	{
		RadElem *elt = (*leaves)[i];
		s << "indexes [" << elt->index[0] << ' '
			<< elt->index[1] << ' '<< elt->index[2];
		if (elt->IsQuad())
			s << ' ' << elt->index[3];
		s << ']' << endl;
		s << "surfaces [" << elt->clrIdx[0] << ' '
			<< elt->clrIdx[1] << ' '<< elt->clrIdx[2];
		if (elt->IsQuad())
			s << ' ' << elt->clrIdx[3];
		s << ']' << endl;

		// XXX
		if (elt->props->texCoords)
		{
			s << "coords [";
			for (Int j = 0; j < elt->Sides(); j++)
				s << elt->TexCoord(j);
			s << ']' << endl;
		}
		s << "poly" << endl;
	}	
	s << "end" << endl;
	s.close();
}

Void RadMethod::WriteObjFile(StrConst filename)
// Write final mesh as a wavefront obj file.
{
	Int				i, j;
	FILE			*file;
	PatchList		*leaves = GetElements();
	IndexList		texOffsets(numProps);
	Int				texOffset;
	StrConst		lastLabel = 0;
	
	file = fopen(filename, "w");

	if (!file)
	{
		_Warning("(WriteObjFile) couldn't open output file for writing");
		return;
	}

	fprintf(file, "# rad output\n");
	fprintf(file, "# faces: %d, points: %d, colours: %d\n",
		leaves->NumItems(), points->NumItems(), colours->NumItems());
	fprintf(file, "\n");
	// dump points, colours, and texture coords if we have 'em.

	for (i = 0; i < points->NumItems(); i++)
	{
		GCLReal *v = (*points)[i].Ref();

		fprintf(file, "v %g %g %g\n",
			(Double) v[0], (Double) v[1], (Double) v[2]);
	}
	
	for (i = 0; i < colours->NumItems(); i++)
	{
		ClrReal *v = (*colours)[i].Ref();

		fprintf(file, "vc %g %g %g\n",
			 (Double) v[0], (Double) v[1], (Double) v[2]);
	}

	// XXX this could lead to repeated tex coord lists
	texOffsets.ClearTo(0);
	texOffset = 0;
	for (i = 0; i < numProps; i++)
		if (props[i].texCoords)
		{
			texOffsets[i] = texOffset;
			texOffset += props[i].texCoords->NumItems();

			for (j = 0; j < props[i].texCoords->NumItems(); j++)
			{
				Coord &vc = props[i].texCoords->Item(j);

				fprintf(file, "vt %f %f\n",
					 (Double) vc[0], (Double) vc[1]);
			}
		}

	// now dump the elements

	for (i = 0; i < leaves->NumItems(); i++)
	{
		RadElem *elt = (*leaves)[i];

		if (elt->props->label && (!lastLabel || (elt->props->label != lastLabel)))
		{
			fprintf(file, "usemtl %s\n", elt->props->label);
			lastLabel = elt->props->label;
		}
		
		if (elt->props->texCoords)
		{
			texOffset = texOffsets[elt->props - props] + 1;

			fprintf(file, "f %d/%d/%d %d/%d/%d %d/%d/%d", 
					elt->index[0] + 1,
					elt->texIdx[0] + texOffset,
					elt->clrIdx[0] + 1,
					elt->index[1] + 1,
					elt->texIdx[1] + texOffset,
					elt->clrIdx[1] + 1,
					elt->index[2] + 1,
					elt->texIdx[2] + texOffset,
					elt->clrIdx[2] + 1
				);
			if (elt->IsQuad())
				fprintf(file, " %d/%d/%d", 
						elt->index[3] + 1,
						elt->texIdx[3] + texOffset,
						elt->clrIdx[3] + 1
						);
		}
		else
		{
			fprintf(file, "f %d//%d %d//%d %d//%d", 
					elt->index[0] + 1,
					elt->clrIdx[0] + 1,
					elt->index[1] + 1,
					elt->clrIdx[1] + 1,
					elt->index[2] + 1,
					elt->clrIdx[2] + 1
				);
			if (elt->IsQuad())
				fprintf(file, " %d//%d", 
						elt->index[3] + 1,
						elt->clrIdx[3] + 1
						);
		}
		fprintf(file, "\n");
	}

	fclose(file);
}

Void RadMethod::WriteMRBFile(StrConst filename)
// Write final mesh as a MRB obj file.
// for now we just write geometry
{
	Int				i, j, totFaces;
	FILE			*file;
	PatchList		*leaves = GetElements();
	MRB_Header		header;
	FaceIdxArray	faces;
		
	totFaces = leaves->NumItems();
	for (i = 0; i < leaves->NumItems(); i++)
		if (leaves->Item(i)->IsQuad())
			totFaces++;

	faces.SetSize(totFaces);
	
	for (i = 0, j = 0; i < leaves->NumItems(); i++)
	{
		RadElem *elt = leaves->Item(i);

		faces[j].v[0] = elt->index[0];
		faces[j].v[1] = elt->index[1];
		faces[j].v[2] = elt->index[2];
		j++;
		
		if (elt->IsQuad())
		{
			faces[j].v[0] = elt->index[2];
			faces[j].v[1] = elt->index[3];
			faces[j].v[2] = elt->index[0];
			j++;
		}
	}

	file = fopen(filename, "w");

	if (!file)
	{
		_Warning("(WriteMRBFile) couldn't open output file for writing");
		return;
	}

	header.version = 1;
#ifdef GCL_FLOAT
	head.flags.Set(MRB_Float);
#endif

	do 
	{
		if (sizeof(MRB_Header) != fwrite(&header, 1, sizeof(MRB_Header), file)) break;
		if (points->FWrite(file)) break;
		if (faces.FWrite(file)) break;
	}
	while (false);

	if (ferror(file))
		perror("MRB write");
	fclose(file);
}


// --- Utility routines -------------------------------------------------------


#ifdef __linux__

#include <unistd.h>

struct MemStat
{
    long		size; 		/* total # of pages of memory */
    long		resident; 	/* number of resident set (non-swapped) pages (4k) */
    long		share; 		/* number of pages of shared (mmap'd) memory */
    long		trs; 		/* text resident set size */
    long		lrs; 		/* shared-lib resident set size */
    long		drs; 		/* data resident set size */
    long		dt; 		/* dirty pages */
};

Int FindLinuxMem(MemStat &stat)
{
	FILE		*procFile;
	Int			pageKB = getpagesize() / 1024;
    Int			num;

	procFile = fopen(String().Printf("/proc/%d/statm", getpid()), "r");

	if (procFile)
	{
	    num = fscanf(procFile, "%ld %ld %ld %ld %ld %ld %ld",
				   &stat.size, &stat.resident, &stat.share,
				   &stat.trs, &stat.lrs, &stat.drs, &stat.dt
			   );

		fclose(procFile);

		stat.size *= pageKB;
		stat.resident *= pageKB;
		stat.share *= pageKB;
		stat.trs *= pageKB;
		stat.lrs *= pageKB;
		stat.drs *= pageKB;
		stat.dt *= pageKB;
	}
	else
		_Warning("Error reading /proc/*/statm");
}

#endif

Int RadMethod::TotalMemoryUse()
{
#ifdef __linux__
	MemStat		stat;

	FindLinuxMem(stat);

	return(stat.size);
#else
	rusage	myRusage;

	getrusage(RUSAGE_SELF, &myRusage);

	return(myRusage.ru_maxrss);
#endif
}

Void RadMethod::SetupProps(RadProps *props, Decimator *state)
{
	SLContext	*context = state->context;
	Colour		*reflectancePtr = SC_GET(Colour);
	Colour		*emittancePtr = SC_GET(Emittance);
	CoordList	*texCoords = SC_GET(Coords);
	Texture		*texture = SC_GET(Texture);
	String		*label = SC_GET(Label);
	
	//	Fill in the properties structure...
	
	props->id = state->decNum;

	if (label)
	{
		props->label = *label;
		DBG_COUT << "material " << *label << endl;
	}
	else
		props->label = 0;

	if (reflectancePtr)
		props->reflectance = *reflectancePtr;	
	else
		props->reflectance = cOrange;

	if (emittancePtr)
		props->emittance = *emittancePtr;
	else
		props->emittance = cBlack;
	
	props->points = state->pointsAcc;
	props->colours = colours;

	if (texCoords && texture)
	{
		props->texCoords = texCoords;
		props->texture = texture;
	}
}

Void RadMethod::Reanimate()
{	
	Int		i;

	numProps = 1;
	props = new RadProps[1];
	props[0].points = points;
	props[0].colours = colours;
	props[0].id = 0;
	props[0].reflectance = cOrange;
	props[0].emittance = cOrange;

	for (i = 0; i < baseElems.NumItems(); i++)
	{
		baseElems[i]->SetProps(props);
		baseElems[i]->Reanimate(0);
	}

	ColourVertices();
}

Void RadMethod::AssignNormals()
{
	if (!normals)
		return;

	Int 	i, j, n = normals->NumItems();
	Int		*weights;
		

	DBG_COUT << "creating vertex normals..." << endl;
	
	weights = new Int[n];

	for (i = 0; i < n; i++)
	{
		weights[i] = 0;
		normals->Item(i) = vl_0;
	}
	
	for (i = 0; i < baseElems.NumItems(); i++)
		for (j = 0; j < baseElems[i]->Sides(); j++)
		{
			baseElems[i]->Normal(j) += baseElems[i]->normal;
			weights[baseElems[i]->normIdx[j]] += 1;
		}	

	for (i = 0; i < n; i++)
	{
		if (weights[i])
			normals->Item(i) /= GCLReal(weights[i]);
		else
			normals->Item(i) = vl_0;
	}

	delete[] weights;
}

Void RadMethod::ColourVertices()
/*! This function should be called before attempting to display the solution
	mesh or attempting to save it. */
{
	Int 	i, n = colours->NumItems();
	Int		*weights;
	
	DBG_COUT << "colouring vertices..." << endl;
	
	weights = new Int[n];
	for (i = 0; i < n; i++)
	{
		weights[i] = 0;
		colours->Item(i) = vl_0;
	}
	
	for (i = 0; i < baseElems.NumItems(); i++)
		baseElems[i]->ColourVertices(weights);

	for (i = 0; i < n; i++)
		if (weights[i])
			colours->Item(i) /= ClrReal(weights[i]);
		else
			colours->Item(i) = cRed; // mark so we can pick up error

#if defined(RAD_HIST)
	Int hist[32];

	for (i = 0; i < 32; i++)
		hist[i] = 0;
	for (i = 0; i < n; i++)
		if (weights[i] < 32)
			hist[weights[i]]++;
	printf("smoothed mesh: %d %d %d %d\n", 
		hist[0], hist[1], hist[2], hist[3]);
#endif

	delete[] weights;
}

PatchList *RadMethod::GetElements()
{
	ColourVertices();
	return(&patches);
}



// --- I/O --------------------------------------------------------------------

static char *kMethTable[] = 
{
	"null",
	"mat",
	"cg",
	"prog",
	"hprog",
	"hier",
	"ana"
};

Void RadMethod::Print(ostream &s) const
{
	s << kMethTable[gRadControl->method] << endl;
	PrintCmds(s);
}

Bool RadMethod::ParseCmd(StrConst str, istream &s)
{
	return(false);
}

Void RadMethod::PrintCmds(ostream &s) const
{
	Int		i;

	s << "points " << *points << endl;
	s << "colours " << *colours << endl;

	if (baseElems.NumItems() > 0)
	{
		s << " base_elems " << baseElems.NumItems() << endl;
			for (i = 0; i < baseElems.NumItems(); i++)
			{
				baseElems[i]->Print(s);
				s << endl;
			}
	}
}

Void RadMethod::Parse(istream &s)
{
	String	str;
	Bool	done = false;

	while (!done)
	{
		str.ReadWord(s);

#ifdef DEBUG
		DBG_COUT << "parsing mesh > " << str << endl;
#endif

		if (str == "end")
			done = true;
		else if (str == "base_elems")
		{
			Int		numBEs, i;

			s >> numBEs;
			baseElems.SetSize(numBEs);
			for (i = 0; i < numBEs; i++)
			{
				// hack XXX
				ChompWhiteSpace(s);
				if (s.peek() == '+')
				{
					Char c;
					s.get(c);
					str.ReadWord(s);
					if (str == "haar")
						gRadControl->basis = kHaar;
					else
						cerr << "ignoring unknown basis type "
							 << str << endl;
				}		

				baseElems[i] = NewMesh();
				baseElems[i]->Parse(s);
			}
			//	str.ReadWord(s);
		}
		else if (str == "points")
		{
			points = new PointList;
			s >> *points;
		}
		else if (str == "colours")
		{
			colours = new ColourList;
			s >> *colours;
		}
		else if (str == "normals")
		{
			normals = new VectorList;
			s >> *normals;
		}
		else if (ParseCmd(str, s))
			;
		else 
		{
			cerr << "unknown mesh keyword: " << str << endl;
			do
			{
				str.ReadLine(s);
				str.ReadWord(s);
			}
			while (str != "end");
			break;
		}
	}

	Reanimate();
}


// --- Creation ---------------------------------------------------------------


#include "MatRad.h"
#include "ProgRad.h"
#include "HProgRad.h"
#include "HierRad.h"
#include "AnaRad.h"

RadMethod *RadMethod::NewRadMethod()
{
	RadMethod *result;
	
	switch (gRadControl->method)
	{
	case kNoMethod:
		_Error("No method selected");
	case kMatrix:
		result = (RadMethod*) MatRad::New();
		break;
	case kProgressive:
		result = (RadMethod*) ProgRad::New();
		break;
	case kProgSubstruct:
		result = (RadMethod*) HProgRad::New();
		break;
	case kHierarchical:
		result = (RadMethod*) HierRad::New();
		break;
	case kAnalytical:
		result = (RadMethod*) AnaRad::New();
		break;
	default:
		cerr << ">>> " << gRadControl->method << endl;
		_Error("Unhandled method");
	}

	Assert(result != 0, "method not supported");
	
	return(result);
}
