/*
	File:			SampleGen.cc

	Function:		Program to calculate sample positions for... well, 
					all sorts of things these days.

	Author(s):		Andrew Willmott

	Copyright:		(c) 1999-2000, Andrew Willmott

	Notes:			Added cube samples as well, 6/97
					Integrated all sample rules into this file + unified
					Samples.h, Samples.cc. Started experimenting with Halton
					points etc. 1/99
					
					Some explanation is in order: this program
					generates samples over the unit square, triangle, 
					cube, circle, and sphere for stratified sampling.

					It also generates 16 sample patterns for use in
					patch-patch visibility testing according to magic
					square rules. For both the square and triangle, two
					different cube rules are alternated so that the
					patterns generated between two linked patches have
					suitably random orientations.
*/

#include "gcl/Geometry.h"
#include "gcl/VecUtil.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iomanip.h>

// --- by-hand stuff ----------------------------------------------------------

// Magic squares!

static Int square1[4][4] = {
	15,  8,  5,  2,
	 4,  3, 14,  9,
	10, 13,  0,  7,
	 1,  6, 11, 12
	};

static Int square2[4][4] = {
	10, 7, 12, 1,
	13, 0, 11, 6,
	3, 14, 5, 8,
	4, 9, 2, 15
	};

Void CheckSquare(Int square[4][4])
// Is it really magic?
{
	Int i, j;
	Int csum, rsum;
	
	for (i = 0; i < 4; i++)
	{
		rsum = 0;
		csum = 0;
		for (j = 0; j < 4; j++)
		{
			rsum += square[i][j];
			csum += square[j][i];
		}
		cout << rsum << ", " << csum << endl;
	}
}

Void GenSquareCoords16(Int numArrays, GCLReal jitterAmt, Char *name)
// range of jitterAmt: 0-1
{
	Int i, j, k, m;
	GCLReal x, y;
	
	cout << "GCLReal " << name << "[" << numArrays * 2
			<< "][16][2] =" << endl <<"{" << endl;
	
	for (m = 0; m < numArrays; m++)
	{
		for (k = 0; k < 16; k++)
			for (i = 0; i < 4; i++)
				for (j = 0; j < 4; j++)
					if (square1[i][j] == k)
					{
						x = 1 / 8.0 + i / 4.0 + (drand48() - 0.5)
							* jitterAmt * 1.0 / 4.0;
						y = 1 / 8.0 + j / 4.0 + (drand48() - 0.5)
							* jitterAmt * 1.0 / 4.0;
						
						if (k % 4 == 0)
							cout << "    ";
						
						cout << setprecision(3) << x << ", "
							 << setprecision(3) << y;
						
						if (k % 4 != 3)
							cout << ", ";	
						else
							cout << "," << endl;
					}

		for (k = 0; k < 16; k++)
			for (i = 0; i < 4; i++)
				for (j = 0; j < 4; j++)
					if (square2[i][j] == k)
					{
						x = 1 / 8.0 + i / 4.0 + (drand48() - 0.5) 
							* jitterAmt * 1.0 / 4.0;
						y = 1 / 8.0 + j / 4.0 + (drand48() - 0.5) 
							* jitterAmt * 1.0 / 4.0;
						
						if (k % 4 == 0)
							cout << "    ";
						
						cout << setprecision(3) << x << ", "
							 << setprecision(3) << y;
						
						if (k % 4 != 3)
							cout << ", ";	
						else if (k / 4 == 3 && m == numArrays - 1)
							cout << endl;
						else
							cout << "," << endl;
					}
	}
	
	cout << "};" << endl << endl;
}

Void GenTriCoords16(Int numArrays, GCLReal jitterAmt, Char *name)
{
	Int i, j, k, m;
	GCLReal x, y, s, t;
	
	cout << "GCLReal " << name << "[" << numArrays * 2
			<< "][16][2] =" << endl <<"{" << endl;
	
	for (m = 0; m < numArrays; m++)
	{
		for (k = 0; k < 16; k++)
			for (i = 0; i < 4; i++)
				for (j = 0; j < 4; j++)
					if (square1[i][j] == k)
					{
						x = 1 / 8.0 + i / 4.0 + (drand48() - 0.5) 
							* jitterAmt * 1.0 / 4.0;
						y = 1 / 8.0 + j / 4.0 + (drand48() - 0.5) 
							* jitterAmt * 1.0 / 4.0;
						
						y = sqrt(y);
						s = x * y;
						t = 1 - y;
						
						if (k % 4 == 0)
							cout << "    ";
						
						cout << setprecision(3) << s << ", " 
							 << setprecision(3) << t;
						
						if (k % 4 != 3)
							cout << ", ";	
						else
							cout << "," << endl;
					}

		for (k = 0; k < 16; k++)
			for (i = 0; i < 4; i++)
				for (j = 0; j < 4; j++)
					if (square2[i][j] == k)
					{
						x = 1 / 8.0 + i / 4.0 + (drand48() - 0.5) 
							* jitterAmt * 1.0 / 4.0;
						y = 1 / 8.0 + j / 4.0 + (drand48() - 0.5) 
							* jitterAmt * 1.0 / 4.0;
						
						y = sqrt(y);
						s = x * y;
						t = 1 - y;
						
						if (k % 4 == 0)
							cout << "    ";
						
						cout << setprecision(3) << s << ", "
							 << setprecision(3) << t;
						
						if (k % 4 != 3)
							cout << ", ";	
						else if (k / 4 == 3 && m == numArrays - 1)
							cout << endl;
						else
							cout << "," << endl;
					}
	}

	cout << "};" << endl << endl;
}

Void GenCubeCoords16(Int numArrays, GCLReal jitterAmt, Char *name)
{
	Int 		k, b, m;
	GCLReal		x, y, z;
	
	cout << "GCLReal " << name << "[" << numArrays
			<< "][16][3] =" << endl <<"{" << endl;
	
	for (m = 0; m < numArrays; m++)
		for (k = 0; k < 16; k++)
		{			
			b = Int(drand48() * 4); if (b == 4) b = 0;
			x = 1 / 8.0 + b / 4.0 + (drand48() - 0.5) * jitterAmt * 1.0 / 4.0;
			b = Int(drand48() * 4); if (b == 4) b = 0;
			y = 1 / 8.0 + b / 4.0 + (drand48() - 0.5) * jitterAmt * 1.0 / 4.0;
			b = Int(drand48() * 4); if (b == 4) b = 0;
			z = 1 / 8.0 + b / 4.0 + (drand48() - 0.5) * jitterAmt * 1.0 / 4.0;
						
			if (k % 2 == 0)
				cout << "    ";
			
			cout << setprecision(3) << x << ", " << setprecision(3) << y 
				<< ", " << setprecision(3) << z;
			
			if (k % 2 == 0)
				cout << ", ";	
			else if (k == 15 && m == numArrays - 1)
				cout << endl;
			else
				cout << "," << endl;
		}

	cout << "};" << endl << endl;
}

// --- Generate circle samples ------------------------------------------------

/*
	Creates jittered circle sampling pattern which can be used incrementally.

	The pattern is four concentric rings of samples. The inner
	ring contains 3 samples, and thus the subsequent rings
	contain 6, 9 and 12 samples, for a total of 30 samples.
	The spacing chosen allows us to use three evenly spaced
	samples on the outer ring
	
	We want conflicting things from our pattern; the first n
	samples for any n should be reasonably evenly distributed
	over the unit circle, and the first three samples should be
	evenly spaced along the edge of this circle (this is for
	sampling optimisations used by the ray tracer).
	
	NOTE: Halton point set is possible replacement for this hack.
 */
 
typedef CoordList SampleList;


Void CreateSamples(SampleList &samples, GCLReal jitterAmt)
{
	Int		i, j, totSamples;
	GCLReal	angle, angleInc, angleOffset;
	Coord	newSample;
	GCLReal	jitterAngle, jitterRadius;
		
	// Create 4 sample rings, & offset them rotationally from each other.
	// 30 samples in all: 12 + 9 + 6 + 3

	for (i = 1; i <= 4; i++)
	{
		totSamples = 3 * i;
		angleInc = 2 * vl_pi / totSamples;
		angleOffset = 0.5 * vl_pi * i;
		angle = angleOffset;
				
		for (j = 0; j < totSamples; j++)
		{
			if (i < 4 || j % 4 != 0)
			{
				jitterAngle = angleInc * jitterAmt * (drand48() - 0.5);
				jitterRadius = jitterAmt * (drand48() - 0.5);
			}
			else
			{
				// Don't jitter outer three samples.
				
				jitterAngle = 0;
				jitterRadius = 0;	
			}
			
			newSample = Coord(sin(angle + jitterAngle), 
							  cos(angle + jitterAngle));
			newSample *= 0.25 * (i - 0.5 + jitterRadius);
			samples.Append(newSample);	
			angle += angleInc;
		}
	}
}

Void JitterRotateSamples(SampleList &samples, Int j, Int n, GCLReal jitterAmt)
{
	Mat2d T;
	GCLReal rotateAngle;
	Int i;
	
	rotateAngle = (j + 2 * jitterAmt * (drand48() - 0.5)) * 2 * vl_pi / n;
	
	T = Rot2d(rotateAngle);
	
	for (i = 0; i < samples.NumItems(); i++)
		samples[i] = T * samples[i];	
}

Int gReorderArray[30] =
{
	// Pen and paper produce these!
	
	//	pattern: x2, x4, x3, x4, x3, x1, x4, x2, x3, x4
	
	18, 22, 26, 	// first three
	0, 1, 2,		// for balance
	
	3,      9, 21, 12,    24, 6, 15, 27, 	// 0 3 9 18
	4, 19, 10,     13,    25, 7, 16, 28, 	// 1 4 10 19
	5, 20, 11, 23, 14,        8, 17, 29		// 2 5 11 20
};

Void ReorderSamples(SampleList &samples)
{
	Int			i;
	SampleList	newSamples(samples.NumItems());

	// reorders circle samples so that subset [1..n] of them still gives
	// good coverage of the circle.
	
	for (i = 0; i < 30; i++)
		newSamples[i] = samples[gReorderArray[i]];
	
	samples = newSamples;
}

Void PrintSamples(SampleList &samples)
{
	Int i;
	
	for (i = 0; i < samples.NumItems(); i++)
		cout << "    " << samples[i][0] << ", " 
			 << samples[i][1] << ", " << endl;
}

Void GenCircleCoords30(Int numArrays, GCLReal jitterAmt, Char *name)
{
	SampleList	samples;
	Int			i;
		
	CreateSamples(samples, 0);

	cout << "GCLReal " << name << " [" << numArrays << "]["
		 << samples.NumItems() << "][2] =" << endl;
		 
	cout << "{" << endl;

	for (i = 0; i < numArrays; i++)
	{
		samples.Clear();
		CreateSamples(samples, jitterAmt);
		ReorderSamples(samples);
		JitterRotateSamples(samples, i, numArrays, jitterAmt);

		PrintSamples(samples);
		cout << endl;
	}
	
	cout << "};" << endl << endl;	
}


// --- Quasi-Monte Carlo sampling ---------------------------------------------

// Mapping routines

Point SquareToSphere(Coord c)
// convert c to a unit length screw
// this is not a particularly good mapping -- distortion at the poles.
// need better!
{
    GCLReal 	phirad, w;
	Point		result;
	
	c = 2.0 * c - vl_1;
	phirad =  2.0 * c[1] * vl_pi;
	w = sqrt(1.0 - sqr(c[0]));
	result = Point(w * cos(phirad), w * sin(phirad), c[0]);

	return(result);
}

Coord SquareToTriangle(Coord c)
{
	GCLReal t = sqrt(c[1]);
	Coord	result;	

	result[0] = t * c[0];
	result[1] = 1 - t;

	return(result);
}

Coord SquareToCircle(Coord c)
// From Chiu & Shirley
{
	GCLReal		phi, r;
	Coord		a = 2.0 * c - vl_1;	// a is now on [-1,1] ^ 2 

	if (a[0] > -a[1])    	// region 1 or 2
	{
		if (a[0] > a[1])   	// region 1, also |a| > |b| 

		{
			r = a[0];
			phi = (vl_pi / 4.0) * (a[1] / a[0]);
		}
		else				// region 2, also |b| > |a| 
		{
			r = a[1];
			phi = (vl_pi / 4.0) * (2.0 - (a[0] / a[1]));
		}
	}
	else         			// region 3 or 4
	{
		if (a[0] < a[1])	// region 3, also |a| >= |b|, a != 0 
        { 
            r = -a[0];
            phi = (vl_pi / 4.0) * (4.0 + (a[1] / a[0]));
		}
		else          		// region 4, |b| >= |a|, but a==0 and b==0 could occur.
		{
            r = -a[1];
            if (a[1] != 0)
                phi = (vl_pi / 4.0) * (6.0 - (a[0] / a[1]));
            else
                phi = 0.0;
		}
	}

	return(r * Coord(cos(phi), sin(phi)));
}

// Sample generation routines 

GCLReal HaltonSequence(Int n, Int b)
// return term i of the base b Halton sequence
{
	GCLReal		p, ip, result;
	Int			k, a;
	
	result = 0;
	ip = 1.0 / b;         

	for (p = ip, k = n ; k; p *= ip, k = (Int) (k / b))
		if (a = k % b)
			result += a * p;

	return(result);
}

Void PrintCoords(Coord *coords, Int n, Char *name, Char *brackets)
{
	Int		i;

	cout << "GCLReal " << name << brackets << "[2] =" << endl << "{" << endl;

	for (i = 0; i < n; i++)
	{
		if ((i & 0x3) == 0)
			cout << "    ";
			
		cout << coords[i][0] << ", " << coords[i][1];

		if (i == n - 1)
			cout << endl;
		else
		{
			cout << ", ";
			if ((i & 0x3) == 3)
				cout << endl;
		}
	}
	
	cout << "};" << endl << endl;
}

Void PrintPoints(Point *points, Int n, Char *name, Char *brackets)
{
	Int		i;

	cout << "GCLReal " << name << brackets << "[3] =" << endl << "{" << endl;

	for (i = 0; i < n; i++)
	{
		if ((i & 0x1) == 0)
			cout << "    ";
			
		cout << points[i][0] << ", " << points[i][1] << ", " << points[i][2];

		if (i == n - 1)
			cout << endl;
		else
		{
			cout << ", ";
			if ((i & 0x1) == 1)
				cout << endl;
		}
	}
	
	cout << "};" << endl << endl;
}

Void PrintCoordsArray(Coord *coords, Int n, Char *name)
{
	static Char str[16];

	sprintf(str, "[%d]", n);
	PrintCoords(coords, n, name, str);
}

Void PrintPointsArray(Point *points, Int n, Char *name)
{
	static Char str[16];

	sprintf(str, "[%d]", n);
	PrintPoints(points, n, name, str);
}

Void GenSquareCoords(Int m, Char *name)
{
	Coord		*coords;
	Int			i;

	coords = new Coord[m];

	for (i = 0; i < m; i++)
		coords[i] = Coord(HaltonSequence(i + 1, 2), HaltonSequence(i + 1, 3));

	PrintCoordsArray(coords, m, name);

	delete[] coords;
}

Void GenCubePoints(Int m, Char *name)
{
	Point		*points;
	Int			i;
	
	points = new Point[m];

	for (i = 0; i < m; i++)
		points[i] = Point(HaltonSequence(i + 1, 2), HaltonSequence(i + 1, 3),
			HaltonSequence(i + 1, 5));

	PrintPointsArray(points, m, name);	

	delete[] points;
}

Void GenTriCoords(Int m, Char *name)
{
	Coord		*coords;
	Int			i;

	coords = new Coord[m];

	for (i = 0; i < m; i++)
		coords[i] = SquareToTriangle(Coord(HaltonSequence(i + 1, 2), HaltonSequence(i + 1, 3)));
	
	PrintCoordsArray(coords, m, name);	

	delete[] coords;
}

Void GenCircleCoords(Int m, Char *name)
{
	Coord		*coords;
	Int			i;

	coords = new Coord[m];

	for (i = 0; i < m; i++)
	{
		coords[i] = SquareToCircle(Coord(HaltonSequence(i + 1, 2), HaltonSequence(i + 1, 3)));
	}
	
	PrintCoordsArray(coords, m, name);	

	delete[] coords;
}

Void GenSpherePoints(Int m, Char *name)
{
	Point		*points;
	Int			i;
	
	points = new Point[m];

	for (i = 0; i < m; i++)
		points[i] = SquareToSphere(Coord(HaltonSequence(i + 1, 2), HaltonSequence(i + 1, 3)));

	PrintPointsArray(points, m, name);	

	delete[] points;
}


// --- main -------------------------------------------------------------------


static Char *gHeaderMsg[] = 
{
	"/*",
	"	File:           Samples.h",
	"",
	"	Function:       Gives samples over various unit entities.",
	"*/",
	"",
	"#include \"gcl/Geometry.h\"",
	"",
	0
};

Int main(Int argc, Char *argv[])
{
	GCLReal	jitterAmt = 0.8;
	Int		numArrays = 16;
	Int		numGenSamples = 1024;
	
	if (argc != 2 || argv[1][1] != 'h')
	{
		cout << "// This file was automatically generated by " << argv[0] << endl
			 << endl;
			 
		cout << "#include \"Samples.h\"" << endl << endl;
		
		cerr << "gen square16..." << endl;
		GenSquareCoords16(numArrays / 2, jitterAmt, "tSquareSamples16");
		cerr << "gen triangle16..." << endl;
		GenTriCoords16(numArrays / 2, jitterAmt, "tTriangleSamples16");

		cerr << "gen square..." << endl;
		GenSquareCoords(numGenSamples, "tSquareSamples");
		cerr << "gen triangle..." << endl;
		GenTriCoords(numGenSamples, "tTriangleSamples");
		cerr << "gen circle..." << endl;
		GenCircleCoords(numGenSamples, "tCircleSamples");
		cerr << "gen cube..." << endl;
		GenCubePoints(numGenSamples, "tCubeSamples");
		cerr << "gen sphere..." << endl;
		GenSpherePoints(numGenSamples, "tSphereSamples");
		cerr << "done." << endl;
	}
	else
	{
		Char	**p = gHeaderMsg;

		while (*p)
			printf("%s\n", *p++);

		printf("const Int kNumSampArrays = %d;\n", numArrays);
		printf("const Int kMaxStratSamples = %d;\n", numGenSamples);
		printf("\n");

		printf("extern GCLReal tSquareSamples16[%d][%d][2];\n", numArrays, 16);
		printf("extern GCLReal tTriangleSamples16[%d][%d][2];\n", numArrays, 16);

		printf("extern GCLReal tSquareSamples[%d][2];\n", numGenSamples);
		printf("extern GCLReal tTriangleSamples[%d][2];\n", numGenSamples);
		printf("extern GCLReal tCircleSamples[%d][2];\n", numGenSamples);
		printf("extern GCLReal tCubeSamples[%d][3];\n", numGenSamples);
		printf("extern GCLReal tSphereSamples[%d][3];\n", numGenSamples);
	}		

	return(0);
}
