/*
	File:		VecLink.cc
	
	Purpose:	Implements vector-transfer for wavelet radiosity
	
	Author:		Andrew Willmott
	
	Notes:		
*/

#include "gcl/Draw.h"
#include "gcl/VecUtil.h"
#include "VecLink.h"
#include "RadMethod.h"
#include "Cluster.h"
#include "FCElem.h"

#ifdef RAD_VIS
#include "gcl/ScenePane.h"
#endif

#define DBG_COUT if (1) cerr
#define DBG2_COUT if (0) cerr

enum VLFlags
{
	vlTo 		= 0x10,
	vlFrom		= 0x20,
	vlInactive	= 0x40,
	vlGotVis	= 0x80
};

// use upper bound in error calcs
#define VR_USE_UB

// use avg of upper + lower for transport
#define VR_AVG_TRANS

// use bounds to decide which elt to subdivide
//#define VR_SMART_SUB

// decrease error directly by visErr when there's any partial shadow,
// rather than using Gibson's method. (More robust.)
//#define RAD_STEP_VIS

// use the projected area rather than total area for deciding which
// element to subdivide.
//#define RAD_PROJ_AREA_REF

const Int kNumTransSamples = 16;


// --- VecLink methods ----------------------------------------------------

RefChoice VecLink::RefineOracle()
{
	// force refinement of mixed clusters, a la Gibson
	// (a mixed cluster contains a light source.)
	
	if (from->flags.AreSet(hrMixed))
	{
		if (from == to)
			return(kSubBoth);
		else
			return(kSubFrom);
	}

#ifdef UNFINISHED
	if (flags.IsSet(vlInactive))
	{
		CalcTransport();
	}
#endif

	GCLReal		w, err;
	GCLReal 	kAError = gRadControl->kAError;
	GCLReal		kFError = gRadControl->kFError;
	
	DBG2_COUT << "REFINE ORACLE: test link from " << from->id << ':' << 
		from->flags << " to " << to->id << ':' << to->flags << endl;

#ifdef RAD_STEP_VIS
	if (visibility > 0.0 && visibility < 1.0)
		w = gRadControl->visError;
	else
		w = 1.0;
#else
	w = abs(1.0 - 2 * visibility) * (1.0 - gRadControl->visError) 
		+ gRadControl->visError;
#endif

	err = BFAError();

	DBG2_COUT << "error = " << err << ", w = " << w << endl;

	Expect(err >= 0.0, "negative link error???");
	if (err < 0.0) err = 0.0;

	if (err < kFError * w)
		return(kSubNone);

	if (to == from)
		return(kSubBoth);

	DBG2_COUT << from->EltArea() << ' ' << to->EltArea() << ' ' << kAError << endl;
	DBG2_COUT << flags << endl;
	//	Decide which patch to subdivide...

	if (flags.AreSet(vlFrom | vlTo))
		if (to->EltArea() < kAError && from->EltArea() < kAError)
			return(kSubNone);
		else
			return(kSubBoth);

	if (flags.IsSet(vlFrom))
		if (from->EltArea() < kAError)
			return(kSubNone);
		else
			return(kSubFrom);

	if (flags.IsSet(vlTo))
		if (to->EltArea() < kAError)
			return(kSubNone);
		else
			return(kSubTo);

    Vector		temp;
	GCLReal		fromArea, toArea;

	temp = safe_norm(from->EltCentre() - to->EltCentre());
#ifdef RAD_PROJ_AREA_REF
    toArea = to->EltMaxProjArea(temp);
    fromArea = from->EltMaxProjArea(-temp);
#else
	toArea = to->EltArea();
	fromArea = from->EltArea();
#endif

	DBG2_COUT << "refining : " << fromArea << " -> " << toArea
		<< " : " << from->IsCluster() << ":" << to->IsCluster() << endl;

	//	Decide which patch to subdivide...

	if (gRadControl->favourReceiver * toArea > fromArea)
		if (to->EltArea() < kAError)
			return(kSubNone);
		else
			return(kSubTo);
	else 
		if (from->EltArea() < kAError)
			return(kSubNone);
		else
			return(kSubFrom);
}


typedef Coord Bounds;

Void UpdateBounds(GCLReal val, Bounds &bounds)
{
	if (val < bounds[0])
		bounds[0] = val;
	if (val > bounds[1])
		bounds[1] = val;
}

Void InitBounds(Bounds &bounds)
{
	bounds[0] = vl_inf;
	bounds[1] = -vl_inf;
}

inline GCLReal BoundsErr(Bounds &bounds)
{ return(bounds[1] - bounds[0]); }

inline GCLReal BoundsAvg(Bounds &bounds)
{ return((bounds[0] + bounds[1]) * 0.5); }

Bool VecLink::CalcTransport()
{	
	Assert(vl_is_finite(sqrlen(from->B_Coeffs()[0])), "infinite source coeffs");
	Assert(vl_is_finite(sqrlen(to->B_Coeffs()[0])), "infinite dest coeffs");

#ifdef UNFINISHED
	// lazy evaluation...
	error = to->EltArea();
	if (!flags.IsSet(vlLazy) && BFAError() <= 0.01)
	// don't bother -- not enough energy to transfer yet.
	{
		flags.Set(vlLazy);
		return(true);
	}	
#endif

	if (from == to)
	{
		// XXX find internal ff & use that.
		m = vl_0;
		if (from->IsFaceClus())
		{
			FCElem *fc = (FCElem*) from;
//			error = (fc->totArea - fc->maxVisArea);
//			error = from->EltArea();
			error = fc->maxVisArea - fc->projArea;
		}
		else
			error = from->EltArea();

		return(true);
	}

// We have two things to do here: find a good estimate for the transport 
// vector m(mS), and calculate the error in the transfer.

	Point		p1[kNumTransSamples];
	Point		p2[kNumTransSamples];
	Point		c1, c2;
	Int			i;
	Vector		r, mEst;
	GCLReal		rLen, a, m2Len;
	GCLReal		l1, l2;
	GCLReal		u1, u2;
	GCLReal		s1, s2;
	GCLReal		err1, err2;
	Bounds		pa1, pa2, factor;
	
	from->EltGetSamples(kNumTransSamples, p1);
	to->EltGetSamples(kNumTransSamples, p2);

#ifdef VR_SMART_SUB
	c1 = from->EltCentre();
	c2 = to->EltCentre();
#ifdef VR_USE_UB
	s1 = from->EltMaxProjArea(safe_norm(c2 - c1));
	s2 = to->EltMaxProjArea(safe_norm(c1 - c2));
#else
	s1 = from->EltProjArea(safe_norm(c2 - c1));
	s2 = to->EltProjArea(safe_norm(c1 - c2));
#endif
	DBG2_COUT << "s1 = " << s1 << endl;
	DBG2_COUT << "s2 = " << s2 << endl;
	InitBounds(pa1);
	InitBounds(pa2);
#endif

	m = vl_0;
	a = from->EltArea();
	
	InitBounds(factor);
	
	for (i = 0; i < kNumTransSamples; i++)
	{
		r = p1[i] - p2[i];

		rLen = len(r);
		if (rLen > 0.0)
		{
			r /= rLen;
			m2Len = vl_pi * sqr(rLen) + a;

			l1 = from->EltProjArea(-r);
			l2 = to->EltProjArea(r);
			UpdateBounds(l1 * l2 / m2Len, factor);

			m += r * l1 / m2Len;

#ifdef VR_USE_UB
			u1 = from->EltMaxProjArea(-r);
			u2 = to->EltMaxProjArea(r);
			UpdateBounds(u1 * u2 / m2Len, factor);

#ifdef VR_AVG_TRANS
			m += 0.5 * (u1 - l1)* r / m2Len;
#endif

			Assert(u1 >= 0.95 * l1, "Bad from bounds!");
			Assert(u2 >= 0.95 * l2, "Bad to bounds!");
#endif
		}

#ifdef VR_SMART_SUB
		r = c2 - p1[i];
		rLen = len(r);
		if (rLen > 0.0)
		{
			r *= s2 / (rLen * (vl_pi * sqr(rLen) + a));
			
			l1 = from->EltProjArea(r);
			UpdateBounds(l1, pa1);

#ifdef VR_USE_UB
			u1 = from->EltMaxProjArea(r);
			UpdateBounds(u1, pa1);
#endif

			DBG2_COUT << "1: " << r << ' ' << l1 << ' ' << u1 << ' ' << pa1 << endl;
		}

		r = c1 - p2[i];
		rLen = len(r);
		if (rLen > 0.0)
		{
			r *= s1 / (rLen * (vl_pi * sqr(rLen) + a));
			
			l2 = to->EltProjArea(r);
			UpdateBounds(l2, pa2);

#ifdef VR_USE_UB
			u2 = to->EltMaxProjArea(r);
			UpdateBounds(u2, pa2);
#endif

			DBG2_COUT << "2: " << r << ' ' << l2 << ' ' << u2 << ' ' << pa2 << endl;
		}
#endif
	}

	DBG2_COUT << pa1 << pa2 << factor << endl;
	m /= kNumTransSamples;
#ifdef VR_MAX_ERR
	error = factor[1];
#else
	error = BoundsErr(factor);
#endif

#ifdef VR_SMART_SUB
	err1 = BoundsErr(pa1) * ;
	err2 = BoundsErr(pa2);

	flags.UnSet(vlTo | vlFrom);
	if (err1 < err2 + a * 0.001)
		flags.Set(vlTo);
	if (err2 < err1 + a * 0.001)
		flags.Set(vlFrom);
#endif

		
#ifdef RAD_VIS
	if (gRadControl->showLinks)
	{
		Int s1 = from->eltHighlight, s2 = to->eltHighlight;
		
		from->SetHighlight(3);
		to->SetHighlight(2);

		RM_DISPLAY_START;
		gRadControl->radObject->display->Begin(renLines)
			.C(cRed)
			.P(from->EltCentre()).P(to->EltCentre())
			.C(cYellow);
		for (i = 0; i < kNumTransSamples; i++)
			gRadControl->radObject->display->P(p1[i]).P(p2[i]);
		gRadControl->radObject->display->End();
		RM_DISPLAY_END;

#ifdef VR_SMART_SUB
		RM_OUT1("Link Error: " << error << ' ' 
			<< err1 << ' ' << err2 << ' '
			<< pa1 << pa2
			);
#else
		RM_OUT1("Link Error: " << error << ' ' << factor);
#endif
		RM_OUT2("Link Radiosity: " << from->B_Coeffs()[0] << " -> " << to->B_Coeffs()[0]);
		RM_OUT3(BFAError());
		if (RM_PAUSE) return(0);

		from->SetHighlight(s1);
		to->SetHighlight(s2);
	}
#endif

	// if there's some error, keep the link around even if it's zero
	return(len(m) > 0.0 || error > 0.0);
}

Bool VecLink::CalcVisibility(HRLink *parent, Bool reuse)
{
	if (flags.IsSet(vlInactive) || BFAError() >= gRadControl->kFError)
	{
		visibility = 1.0;
		return(true);
	}

	visibility = from->EltVis(to);
	m *= visibility;
	flags.Set(vlGotVis);

	return(true);
}

Void VecLink::Gather()
{		
	Int		i;
	Colour	E;

	if (flags.IsSet(vlInactive))
		return;
	if (!flags.IsSet(vlGotVis))
	{
		visibility = from->EltVis(to);
		m *= visibility;
	}
	Assert(vl_is_finite(m[0]) && vl_is_finite(m[1]) && vl_is_finite(m[2]),
		"Gather:BAD m!!!");

	E = from->B_Coeffs()[0];
	Assert(vl_is_finite(E[0]) && vl_is_finite(E[1]) && vl_is_finite(E[2]),
		"Gather: BAD E!!!");

	to->AddIrradiance(E, m);		// add in irradiance vector Em
}

GCLReal VecLink::Error()
{
	return(error);
}

GCLReal VecLink::BFAError()
{
	GCLReal result;
	// error is in power, so just B*error

	result = error * dot(kRadRGBToLum, to->EltRho() * from->B_Coeffs()[0]);

	return(result);
}

Void VecLink::DebugInfo()
{
	HRLink::DebugInfo();
//	cout << "m = " << m << endl;
}

Vector VecLink::PointIrradiance(const Point &p, const Vector &clipNormal, Colour &clr)
{	
	const Int kNumVisSamples = 16;

	Point		p1[kNumVisSamples];
	Int			i;
	Vector		r, mEst, m;
	GCLReal		rLen, a, m2Len;
	GCLReal		factor, l1, u1;
	
	clr = from->B_Coeffs()[0];
	from->EltSetVisPoints(to, p1);

	m = vl_0;
	a = from->EltArea();
	
	for (i = 0; i < kNumVisSamples; i++)
	{
		r = p1[i] - p;
		if (dot(r, clipNormal) <= 0.0)
			continue;
		
		rLen = len(r);
		if (rLen > 0.0)
		{
			r /= rLen;
			m2Len = vl_pi * sqr(rLen) + a;

			l1 = from->EltProjArea(-r);
			u1 = from->EltMaxProjArea(-r);

			factor = 0.5 * (l1 + u1) / m2Len;

			// trace ray if there's something there & we had partial vis.
			if (gRadControl->bestVisPass && factor > 0.0 && visibility < 1.0)
			{
				gRadControl->rays++;
				if (gRadControl->radObject->
					IntersectsWithRay(p1[i], p))
						continue;
			}

			m += r * factor;
		}
	}

	m /= kNumVisSamples;
	if (!gRadControl->bestVisPass)
		m *= visibility;

	return(m);
}
