/*
	File:		WaveSupport.h

	Function:	Support routines for flatlets & multiwavelets

	Author:		Andrew Willmott

    Copyright:  (c) 1996-2000, Andrew Willmott
*/

#ifndef __WaveSupport__
#define __WaveSupport__

#include "gcl/Renderer.h"

// define a vector of colours
typedef Colour ClrVec[];

// Matrix and Vector operations

GCLReal	ClrVecCalcError(ClrVec B1, ClrVec B2,  Int n);
Colour	ClrVecAverage(ClrVec B, Int n);
Void	ClrVecScale(ClrVec c, Int n, GCLReal s);

Void	ClrVecMult(const GCLMat &m, ClrVec src, ClrVec dst);
Void	ClrVecMultAcc(const GCLMat &m, ClrVec src, ClrVec dst);

// Tensor multiplicaton routines for transforming vectors of colours
// by scalar tensor matrices.
//
// (A tensor-product multiplication is O(n ^ 3/2) where the equivalent full 
// multiplication (using Mult) would be O(n^2), which is why we go to
// this extra trouble to perform it.)

Void	ClrVecTensorMult(const GCLMat &m, ClrVec src, ClrVec dst);
// [dst] = m * [src] * trans(m)
// [src] is the matrix S[i][j] = src[i.n + j], where n is the # of
// rows (and columns) in m.
Void	ClrVecLeftTensorMult(const GCLMat &m, ClrVec src, ClrVec dst);
// dst = m * [src]
Void	ClrVecRightTensorMult(const GCLMat &m, ClrVec src, ClrVec dst);
// [dst] = [src] * trans(m)
Void	ClrVecLeftTensorMultAcc(const GCLMat &m, ClrVec src, ClrVec dst);
// [dst] += m * [src]
Void	ClrVecRightTensorMultAcc(const GCLMat &m, ClrVec src, ClrVec dst);
// [dst] += [src] * trans(m)

// Tensor multiplication routines for transforming ordinary vectors of
// scalars.

Void MatMakeTensorMatrix(const GCLMat &t, GCLMat &m);
// m = T, where T is the tensor-product of matrix t with itself. O(n^4)
Void MatLeftTensorMult(const GCLMat &t, const GCLMat &k, GCLMat &d);
// d = T * k, where T is the tensor-product of matrix t with itself.
// O(mn^3).
Void MatRightTensorMult(const GCLMat &k, const GCLMat &t, GCLMat &d);
// d = k * T, where T is the tensor-product of matrix t with itself.
// O(mn^3).

// Utility routines for basis reconstruction

Colour M2QuadSample(Coord &c, ClrVec B);
Colour M2TriSample(Coord &c, ClrVec B);
Colour M3QuadSample(Coord &c, ClrVec B);
Colour M3TriSample(Coord &c, ClrVec B);

Void DrawWaveLink(Renderer &r, GCLReal left, GCLReal top, 
	GCLReal right, GCLReal bottom, Colour weight, const GCLMat &m);


#endif
